Type: Package
Package: gert
Title: Simple Git Client for R
Version: 2.4.0
Authors@R: c(
    person("Jeroen", "Ooms", role = c("aut", "cre"), email = "jeroenooms@gmail.com",
      comment = c(ORCID = "0000-0002-4035-0289")),
    person("Maëlle", "Salmon", role = "aut", comment = c(ORCID = "0000-0002-2815-0399")),
    person("Jennifer", "Bryan", role = "ctb", email = "jenny@posit.co",
      comment = c(ORCID = "0000-0002-6983-2759")))
Description: Simple git client for R based on 'libgit2' <https://libgit2.org> with
    support for SSH and HTTPS remotes. All functions in 'gert' use basic R data 
    types (such as vectors and data-frames) for their arguments and return values.
    User credentials are shared with command line 'git' through the git-credential
    store and ssh keys stored on disk or ssh-agent.
License: MIT + file LICENSE
URL: https://docs.ropensci.org/gert/,
        https://ropensci.r-universe.dev/gert
BugReports: https://github.com/r-lib/gert/issues
Imports: askpass, credentials (>= 1.2.1), openssl (>= 2.0.3),
        rstudioapi (>= 0.11), sys, zip (>= 2.1.0)
Suggests: spelling, knitr, rmarkdown, testthat (>= 3.0.0), roxygen2
VignetteBuilder: knitr
Encoding: UTF-8
SystemRequirements: libgit2 (>= 1.0): libgit2-devel (rpm) or
        libgit2-dev (deb)
Language: en-US
Config/roxygen2/version: 8.0.0
Config/testthat/edition: 3
NeedsCompilation: yes
Packaged: 2026-07-22 09:03:39 UTC; jeroen
Author: Jeroen Ooms [aut, cre] (ORCID: <https://orcid.org/0000-0002-4035-0289>),
  Maëlle Salmon [aut] (ORCID: <https://orcid.org/0000-0002-2815-0399>),
  Jennifer Bryan [ctb] (ORCID: <https://orcid.org/0000-0002-6983-2759>)
Maintainer: Jeroen Ooms <jeroenooms@gmail.com>
Repository: CRAN
Date/Publication: 2026-07-22 09:40:02 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-07-22 12:10:20 UTC; unix
Archs: gert.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(format,rd_section_gitcommands)
S3method(print,gert_signature)
S3method(print,git_repo_ptr)
S3method(roxygen2::roxy_tag_parse,roxy_tag_git)
S3method(roxygen2::roxy_tag_rd,roxy_tag_git)
export(git_add)
export(git_ahead_behind)
export(git_archive_zip)
export(git_branch)
export(git_branch_checkout)
export(git_branch_create)
export(git_branch_delete)
export(git_branch_exists)
export(git_branch_fast_forward)
export(git_branch_list)
export(git_branch_move)
export(git_branch_set_upstream)
export(git_branch_switch)
export(git_checkout_pull_request)
export(git_cherry_pick)
export(git_clone)
export(git_commit)
export(git_commit_all)
export(git_commit_descendant_of)
export(git_commit_id)
export(git_commit_info)
export(git_commit_stats)
export(git_config)
export(git_config_get)
export(git_config_global)
export(git_config_global_get)
export(git_config_global_set)
export(git_config_global_unset)
export(git_config_local_get)
export(git_config_set)
export(git_config_unset)
export(git_conflicts)
export(git_diff)
export(git_diff_patch)
export(git_fetch)
export(git_fetch_pull_requests)
export(git_find)
export(git_ignore_path_is_ignored)
export(git_info)
export(git_init)
export(git_log)
export(git_ls)
export(git_merge)
export(git_merge_abort)
export(git_merge_analysis)
export(git_merge_find_base)
export(git_merge_stage_only)
export(git_open)
export(git_pull)
export(git_push)
export(git_rebase_commit)
export(git_rebase_list)
export(git_remote_add)
export(git_remote_info)
export(git_remote_list)
export(git_remote_ls)
export(git_remote_refspecs)
export(git_remote_remove)
export(git_remote_set_pushurl)
export(git_remote_set_url)
export(git_reset_hard)
export(git_reset_mixed)
export(git_reset_soft)
export(git_restore)
export(git_revert)
export(git_rm)
export(git_signature)
export(git_signature_default)
export(git_signature_parse)
export(git_stash_drop)
export(git_stash_list)
export(git_stash_pop)
export(git_stash_save)
export(git_stat_files)
export(git_status)
export(git_submodule_add)
export(git_submodule_fetch)
export(git_submodule_info)
export(git_submodule_init)
export(git_submodule_list)
export(git_submodule_set_to)
export(git_tag_create)
export(git_tag_delete)
export(git_tag_list)
export(git_tag_push)
export(git_worktree_add)
export(git_worktree_exists)
export(git_worktree_is_locked)
export(git_worktree_is_prunable)
export(git_worktree_is_valid)
export(git_worktree_list)
export(git_worktree_lock)
export(git_worktree_path)
export(git_worktree_prune)
export(git_worktree_remove)
export(git_worktree_unlock)
export(libgit2_config)
export(user_is_configured)
importFrom(askpass,askpass)
importFrom(credentials,git_credential_ask)
importFrom(credentials,git_credential_forget)
importFrom(credentials,ssh_key_info)
importFrom(credentials,ssh_read_key)
importFrom(openssl,write_openssh_pem)
importFrom(openssl,write_pem)
importFrom(openssl,write_pkcs1)
importFrom(openssl,write_ssh)
useDynLib(gert,R_git_ahead_behind)
useDynLib(gert,R_git_branch_current)
useDynLib(gert,R_git_branch_exists)
useDynLib(gert,R_git_branch_list)
useDynLib(gert,R_git_branch_move)
useDynLib(gert,R_git_branch_set_target)
useDynLib(gert,R_git_branch_set_upstream)
useDynLib(gert,R_git_checkout_branch)
useDynLib(gert,R_git_checkout_unborn)
useDynLib(gert,R_git_cherry_pick)
useDynLib(gert,R_git_commit_create)
useDynLib(gert,R_git_commit_descendant)
useDynLib(gert,R_git_commit_id)
useDynLib(gert,R_git_commit_info)
useDynLib(gert,R_git_commit_log)
useDynLib(gert,R_git_commit_stats)
useDynLib(gert,R_git_config_list)
useDynLib(gert,R_git_config_set)
useDynLib(gert,R_git_config_unset)
useDynLib(gert,R_git_conflict_list)
useDynLib(gert,R_git_create_branch)
useDynLib(gert,R_git_delete_branch)
useDynLib(gert,R_git_diff_list)
useDynLib(gert,R_git_ignore_path_is_ignored)
useDynLib(gert,R_git_merge_analysis)
useDynLib(gert,R_git_merge_cleanup)
useDynLib(gert,R_git_merge_find_base)
useDynLib(gert,R_git_merge_parent_heads)
useDynLib(gert,R_git_merge_stage)
useDynLib(gert,R_git_rebase)
useDynLib(gert,R_git_remote_add)
useDynLib(gert,R_git_remote_add_fetch)
useDynLib(gert,R_git_remote_fetch)
useDynLib(gert,R_git_remote_info)
useDynLib(gert,R_git_remote_list)
useDynLib(gert,R_git_remote_ls)
useDynLib(gert,R_git_remote_push)
useDynLib(gert,R_git_remote_refspecs)
useDynLib(gert,R_git_remote_remove)
useDynLib(gert,R_git_remote_set_url)
useDynLib(gert,R_git_repository_add)
useDynLib(gert,R_git_repository_clone)
useDynLib(gert,R_git_repository_find)
useDynLib(gert,R_git_repository_info)
useDynLib(gert,R_git_repository_init)
useDynLib(gert,R_git_repository_ls)
useDynLib(gert,R_git_repository_open)
useDynLib(gert,R_git_repository_path)
useDynLib(gert,R_git_repository_rm)
useDynLib(gert,R_git_reset)
useDynLib(gert,R_git_restore)
useDynLib(gert,R_git_revert)
useDynLib(gert,R_git_signature_create)
useDynLib(gert,R_git_signature_default)
useDynLib(gert,R_git_signature_parse)
useDynLib(gert,R_git_stash_drop)
useDynLib(gert,R_git_stash_list)
useDynLib(gert,R_git_stash_pop)
useDynLib(gert,R_git_stash_save)
useDynLib(gert,R_git_stat_files)
useDynLib(gert,R_git_status_list)
useDynLib(gert,R_git_submodule_info)
useDynLib(gert,R_git_submodule_init)
useDynLib(gert,R_git_submodule_list)
useDynLib(gert,R_git_submodule_save)
useDynLib(gert,R_git_submodule_set_to)
useDynLib(gert,R_git_submodule_setup)
useDynLib(gert,R_git_submodule_update)
useDynLib(gert,R_git_tag_create)
useDynLib(gert,R_git_tag_delete)
useDynLib(gert,R_git_tag_list)
useDynLib(gert,R_git_worktree_add)
useDynLib(gert,R_git_worktree_exists)
useDynLib(gert,R_git_worktree_is_locked)
useDynLib(gert,R_git_worktree_is_prunable)
useDynLib(gert,R_git_worktree_is_valid)
useDynLib(gert,R_git_worktree_list)
useDynLib(gert,R_git_worktree_lock)
useDynLib(gert,R_git_worktree_path)
useDynLib(gert,R_git_worktree_prune)
useDynLib(gert,R_git_worktree_unlock)
useDynLib(gert,R_libgit2_config)
useDynLib(gert,R_set_cert_locations)
useDynLib(gert,R_static_libgit2)

#' @importFrom openssl write_ssh write_pem write_pkcs1
#' @importFrom credentials ssh_key_info git_credential_forget ssh_read_key
#' @importFrom askpass askpass
make_key_cb <- function(ssh_key = NULL, host = NULL, password = askpass){
  function(){
    if(is.null(ssh_key)){
      ssh_key <- try(ssh_key_info(host = host, auto_keygen = FALSE)$key)
      if(inherits(ssh_key, "try-error"))
        return(NULL)
    }
    key <- tryCatch(ssh_read_key(ssh_key, password = password), error = function(e){
      stop(sprintf("Unable to load key: %s", ssh_key), call. = FALSE)
    })
    # NB: need to add ID for buggy libssh2: https://github.com/libgit2/libgit2/issues/5162
    writeLines(paste(write_ssh(key$pubkey), "git@localhost"), tmp_pub <- tempfile())

    # NB: pkcs1 is the only format that works on all libssh2 configurations
    tmp_key <- if(inherits(key, c("rsa", "dsa", "ecdsa"))){
      write_pkcs1(key, tempfile())
    } else {
      write_pem(key, tempfile())
    }
    if(.Platform$OS.type == "unix"){
      Sys.chmod(tmp_pub, '0644')
      Sys.chmod(tmp_key, '0400')
    }
    c(tmp_pub, tmp_key, "")
  }
}

#' @importFrom credentials git_credential_ask
make_cred_cb <- function(password = askpass, verbose = TRUE){
  if(!is.character(password) && !is.function(password)){
    stop("Password parameter must be string or callback function")
  }
  function(url, username, retries){
    # Case of hardcoded (string) password
    if(is.character(password)){
      if(!length(username) || is.na(username)){
        stop("To use a hardcoded password, include your username in the URL like 'https://jerry@github.com")
      }
      return(c(username, password))
    }

    # Look for GITHUB_PAT variable
    if(retries < 2){
      github_pat <- Sys.getenv('GITHUB_PAT')
      if(nchar(github_pat) > 0 && grepl('^https?://([^/]*@)?github.com', url)){
        return(c("git", github_pat))
      }
    }

    # Retrieve a password from the credential manager
    if(retries > 2){
      try(git_credential_forget(url))
    }
    cred <- try(git_credential_ask(url, verbose = verbose), silent = !verbose)
    if(inherits(cred, 'git_credential')){
      return(c(cred$username, cred$password))
    }

    # If that doesn't work try to manually prompt
    if(!length(username) || is.na(username)){
      password(sprintf("Please enter username for %s", url))
    }
    pwd <- password(sprintf("Please enter a PAT or password for %s", url))
    as.character(c(username, pwd))
  }
}

remote_to_host <- function(repo, remote){
  rms <- git_remote_list(repo = repo)
  url <- rms[rms$name == remote, ]$url
  if(length(url)){
    url_to_host(url)
  }
}

url_to_host <- function(url){
  credentials:::parse_url(url, allow_ssh = TRUE)[['host']]
}

#' Create or discover a local Git repository
#'
#' Use `git_init()` to create a new repository or `git_find()` to discover an
#' existing local repository. `git_info()` shows basic information about a
#' repository, such as the SHA and branch of the current HEAD.
#'
#' For `git_init()` the `path` parameter sets the directory of the git repository
#' to create. If this directory already exists, it must be empty. If it does
#' not exist, it is created, along with any intermediate directories that don't
#' yet exist. For `git_find()` the `path` arguments specifies the directory at
#' which to start the search for a git repository. If it is not a git repository
#' itself, then its parent directory is consulted, then the parent's parent, and
#' so on.
#'
#' @export
#' @rdname git_repo
#' @name git_repo
#' @family git
#' @useDynLib gert R_git_repository_init
#' @inheritParams git_open
#' @param path the location of the git repository, see details.
#' @param bare if true, a Git repository without a working directory is created
#' @return The path to the Git repository.
#' @examples
#' # directory does not yet exist
#' r <- tempfile(pattern = "gert")
#' git_init(r)
#' git_find(r)
#'
#' # create a child directory, then a grandchild, then search
#' r_grandchild_dir <- file.path(r, "aaa", "bbb")
#' dir.create(r_grandchild_dir, recursive = TRUE)
#' git_find(r_grandchild_dir)
#'
#' # cleanup
#' unlink(r, recursive = TRUE)
#'
#' # directory exists but is empty
#' r <- tempfile(pattern = "gert")
#' dir.create(r)
#' git_init(r)
#' git_find(r)
#'
#' # cleanup
#' unlink(r, recursive = TRUE)
git_init <- function(path = '.', bare = FALSE){
  path <- normalizePath(path.expand(path), mustWork = FALSE)
  repo <- .Call(R_git_repository_init, path, as.logical(bare))
  git_repo_path(repo)
}

#' @export
#' @rdname git_repo
#' @useDynLib gert R_git_repository_find
git_find <- function(path = '.'){
  path <- normalizePath(path.expand(path), mustWork = FALSE)
  out <- .Call(R_git_repository_find, path)
  dirname(out)
}

#' @export
#' @rdname git_repo
#' @useDynLib gert R_git_repository_info
git_info <- function(repo = '.'){
  repo <- git_open(repo)
  .Call(R_git_repository_info, repo)
}

test_that("can check that files are ignored", {
  repo <- git_init(tempfile("gert-tests-open"))
  writeLines(c(
    "*.so",
    "!foo.so",
    "some-file",
    "directory/"),
    file.path(repo, ".gitignore"))
  expect_true(git_ignore_path_is_ignored("x.so", repo))
  expect_false(git_ignore_path_is_ignored("foo.so", repo))
  expect_true(git_ignore_path_is_ignored("some-file", repo))
  expect_false(git_ignore_path_is_ignored("some-file.txt", repo))
  expect_true(git_ignore_path_is_ignored("directory/a", repo))
  expect_equal(git_ignore_path_is_ignored(c("x.so", "foo.so"), repo),
               c(TRUE, FALSE))
  expect_equal(git_ignore_path_is_ignored(character(), repo), logical())
})

test_that("can check that files are ignored", {
  repo <- git_init(tempfile("gert-tests-open"))
  writeLines(
    c(
      "*.so",
      "!foo.so",
      "some-file",
      "directory/"
    ),
    file.path(repo, ".gitignore")
  )
  expect_true(git_ignore_path_is_ignored("x.so", repo))
  expect_false(git_ignore_path_is_ignored("foo.so", repo))
  expect_true(git_ignore_path_is_ignored("some-file", repo))
  expect_false(git_ignore_path_is_ignored("some-file.txt", repo))
  expect_true(git_ignore_path_is_ignored("directory/a", repo))
  expect_equal(
    git_ignore_path_is_ignored(c("x.so", "foo.so"), repo),
    c(TRUE, FALSE)
  )
  expect_equal(git_ignore_path_is_ignored(character(), repo), logical())
})

test_that("can opt out of recursive search", {
  repo <- git_init(tempfile("gert-tests-open"))
  on.exit(unlink(repo, recursive = TRUE))

  grandchild <- file.path(repo, "aaa", "bbb")
  dir.create(grandchild, recursive = TRUE)
  expect_error(git_open(I(grandchild)), class = "GIT_ENOTFOUND")
})

#' Get or set Git configuration
#'
#' @description
#' Get or set Git options, as `git config` does on the command line. **Global**
#' settings affect all of a user's Git operations (`git config --global`),
#' whereas **local** settings are scoped to a specific repository (`git config
#' --local`). When both exist, local options always win. Four functions address
#' the four possible combinations of getting vs setting and global vs. local.
#'
#' ```{r echo = FALSE, results = "asis"}
#' dat <- data.frame(
#'   local = c("`git_config()`", "`git_config_set()`"),
#'   global = c("`git_config_global()`", "`git_config_global_set()`"),
#'   row.names = c("get", "set")
#' )
#' knitr::kable(dat, col.names = paste0("**", colnames(dat), "**"))
#' ```
#' @return
#' * `git_config()`: a `data.frame` of the Git options "in force" in the context
#'   of `repo`, one row per option. The `level` column reveals whether the
#'   option is determined from global or local config.
#' * `git_config_global()`: a `data.frame`, as for `git_config()`, except only
#'   for global Git options.
#' * `git_config_set()`, `git_config_global_set()`: The previous value of
#'   `name` in local or global config, respectively. If this option was
#'   previously unset, returns `NULL`. Returns invisibly.
#'
#' @note All entries in the `name` column are automatically normalised to 
#'   lowercase (see
#'   <https://libgit2.org/libgit2/#HEAD/type/git_config_entry> for details).
#'
#' @examples
#' # Set and inspect a local, custom Git option
#' r <- file.path(tempdir(), "gert-demo")
#' git_init(r)
#'
#' previous <- git_config_set("aaa.bbb", "ccc", repo = r)
#' previous
#' cfg <- git_config(repo = r)
#' subset(cfg, level == "local")
#' cfg$value[cfg$name == "aaa.bbb"]
#'
#' previous <- git_config_set("aaa.bbb", NULL, repo = r)
#' previous
#' cfg <- git_config(repo = r)
#' subset(cfg, level == "local")
#' cfg$value[cfg$name == "aaa.bbb"]
#'
#' unlink(r, recursive = TRUE)
#'
#' \dontrun{
#' # Set global Git options
#' git_config_global_set("user.name", "Your Name")
#' git_config_global_set("user.email", "your@email.com")
#' git_config_global()
#' }
#' @export
#' @family git
#' @inheritParams git_open
#' @useDynLib gert R_git_config_list
git_config <- function(repo = '.'){
  repo <- git_open(repo)
  .Call(R_git_config_list, repo)
}

#' @export
#' @rdname git_config
git_config_global <- function(){
  .Call(R_git_config_list, NULL)
}

#' @export
#' @rdname git_config
#' @useDynLib gert R_git_config_set
#' @param name Name of the option to set
#' @param value Value to set. Must be a string, logical, number or `NULL` (to
#'   unset).
git_config_set <- function(name, value, repo = '.'){
  repo <- git_open(repo)
  name <- as.character(name)
  orig_cfg <- git_config(repo = repo)
  out <- orig_cfg$value[orig_cfg$name == name & orig_cfg$level == "local"]
  .Call(R_git_config_set, repo, name, value)
  if (length(out) > 0) {
    invisible(out)
  } else {
    invisible(NULL)
  }
}

#' @export
#' @rdname git_config
git_config_global_set <- function(name, value){
  orig_cfg <- git_config_global()
  out <- orig_cfg$value[orig_cfg$name == name]
  .Call(R_git_config_set, NULL, name, value)
  if (length(out) > 0) {
    invisible(out)
  } else {
    invisible(NULL)
  }
}

#' Show libgit2 version and capabilities
#'
#' `libgit2_config()` reveals which version of libgit2 gert is using and which
#' features are supported, such whether you are able to use ssh remotes.
#'
#' @export
#' @useDynLib gert R_libgit2_config
#' @examples
#' libgit2_config()
libgit2_config <- function(){
  res <- .Call(R_libgit2_config)
  res$version <- as.numeric_version(res$version)
  res
}

# helpers used in tests
configure_local_user <- function(repo = ".") {
  git_config_set('user.name', "Jerry Johnson", repo = repo)
  git_config_set('user.email', "jerry@gmail.com", repo = repo)
}

local_author <- function(repo = ".") {
  user <- git_signature_parse(git_signature_default(repo))
  sprintf("%s <%s>", user$name, user$email)
}

# helpers used in pkgdown setup
configure_global_user <- function() {
  git_config_global_set('user.name', "Jerry Johnson")
  git_config_global_set('user.email', "jerry@gmail.com")
}

global_user_is_configured <- function() {
  cfg <- git_config_global()
  user_name_exists <- any(cfg$name == "user.name")
  user_email_exists <- any(cfg$name == "user.email")
  user_name_exists && user_email_exists
}

# helpers used in examples

#' Test if a Git user is configured
#'
#' This function exists mostly to guard examples that rely on having a user
#' configured, in order to make commits. `user_is_configured()` makes no
#' distinction between local or global user config.
#'
#' @param repo An optional `repo`, in the sense of [git_open()].
#'
#' @return `TRUE` if `user.name` and `user.email` are set locally or globally,
#'   `FALSE` otherwise.
#'
#' @export
#' @examples
#' user_is_configured()
user_is_configured <- function(repo = ".") {
  cfg <- tryCatch(
    git_config(repo),
    error = function(e) git_config_global()
  )
  user_name_exists <- any(cfg$name == "user.name")
  user_email_exists <- any(cfg$name == "user.email")
  user_name_exists && user_email_exists
}

test_that("local, custom config roundtrip", {
  repo <- git_init(tempfile("gert-tests-config"))
  on.exit(unlink(repo, recursive = TRUE))

  orig <- git_config_set("aaa.bbb", "ccc", repo)
  expect_null(orig)
  cfg <- git_config(repo)
  expect_equal(cfg$value[cfg$name == "aaa.bbb"], "ccc")

  orig <- git_config_set("aaa.bbb", NULL, repo)
  expect_equal(orig, "ccc")
  cfg <- git_config(repo)
  expect_equal(cfg$value[cfg$name == "aaa.bbb"], character())
})

#' Merging tools
#'
#' Use `git_merge` to merge a branch into the current head. Based on how the branches
#' have diverged, the function will select a fast-forward or merge-commit strategy.
#'
#' By default `git_merge` automatically commits the merge commit upon success.
#' However if the merge fails with merge-conflicts, or if `commit` is set to
#' `FALSE`, the changes are staged and the repository is put in merging state,
#' and you have to manually run `git_commit` or `git_merge_abort` to proceed.
#'
#' Other functions are more low-level tools that are used by `git_merge`.
#' `git_merge_find_base` looks up the commit where two branches have diverged
#' (i.e. the youngest common ancestor). The `git_merge_analysis` is used to
#' test if a merge can simply be fast forwarded or not.
#'
#' The `git_merge_stage_only` function applies and stages changes, without
#' committing or fast-forwarding.
#'
#' @export
#' @family git
#' @rdname git_merge
#' @name git_merge
#' @inheritParams git_open
#' @param ref branch or commit that you want to merge
#' @param commit automatically create a merge commit if the merge succeeds without
#' conflicts. Set this to `FALSE` if you want to customize your commit message/author.
#' @param squash omits the second parent from the commit, which make the merge a regular
#' single-parent commit.
git_merge <- function(ref, commit = TRUE, squash = FALSE, repo = '.'){
  state <- git_merge_analysis(ref = ref, repo = repo)
  if(state == "up_to_date"){
    inform("Already up to date, nothing to merge")
  } else if(state == "fastforward"){
    inform("Performing fast-forward merge, no commit needed")
    git_branch_fast_forward(ref = ref, repo = repo)
  } else if(state == "normal"){
    merged_without_conflict <- git_merge_stage_only(ref = ref, squash = squash, repo = repo)
    if(!nrow(git_status(repo = repo))){
      inform("Merge did not result in any changes")
    } else if(isTRUE(merged_without_conflict)){
      if(isTRUE(commit)){
        commit_message <- sprintf("Merged %s into %s", ref, git_info(repo = repo)$shorthand)
        git_commit(commit_message, repo = repo)
        inform(commit_message)
      } else {
        inform("Merge was not committed due to merge conflict(s). Please fix and run git_commit() or git_merge_abort()")
      }
    } else {
      inform("Merge has resulted in merge conflict(s).")
    }
  } else {
    stop(sprintf("State is '%s', not sure what to do", state))
  }
}

#' @export
#' @rdname git_merge
#' @useDynLib gert R_git_merge_stage
git_merge_stage_only <- function(ref, squash = FALSE, repo = '.'){
  repo <- git_open(repo)
  success <- .Call(R_git_merge_stage, repo, ref)
  if(isTRUE(squash)) # This turns it in a regular commit
    git_merge_cleanup(repo = repo)
  return(success)
}

#' @export
#' @rdname git_merge
#' @useDynLib gert R_git_merge_find_base
#' @param target the branch where you want to merge into. Defaults to current `HEAD`.
git_merge_find_base <- function(ref, target = "HEAD", repo = '.'){
  repo <- git_open(repo)
  .Call(R_git_merge_find_base, repo, ref, target)
}

#' @export
#' @rdname git_merge
#' @useDynLib gert R_git_merge_analysis
git_merge_analysis <- function(ref, repo = '.'){
  repo <- git_open(repo)
  .Call(R_git_merge_analysis, repo, ref)
}

#' @useDynLib gert R_git_merge_cleanup
git_merge_cleanup <- function(repo = '.'){
  repo <- git_open(repo)
  .Call(R_git_merge_cleanup, repo)
}

#' @export
#' @rdname git_merge
git_merge_abort <- function(repo = '.'){
  if(length(git_merge_parent_heads(repo = repo))){
    git_reset_hard(repo = repo)
    git_merge_cleanup(repo = repo)
  } else {
    inform("There is no merge to abort")
  }
}

#' @useDynLib gert R_git_merge_parent_heads
git_merge_parent_heads <- function(repo = '.'){
  repo <- git_open(repo)
  .Call(R_git_merge_parent_heads, repo)
}

test_that("merge analysis works", {
  repo <- git_init(tempfile("gert-tests-merge"))
  on.exit(unlink(repo, recursive = TRUE))
  oldwd <- getwd()
  on.exit(setwd(oldwd), add = TRUE)
  setwd(repo)
  configure_local_user()
  for(i in 1:5){
    writeLines(paste('Blabla', i), "test.txt")
    git_add("test.txt")
    git_commit(paste("This is commit number:", i))
  }
  master_log <- git_log()
  first_commit <- tail(master_log$commit, 1)
  git_branch_create('new', first_commit)
  git_branch_checkout('new')
  expect_equal(git_log()$commit, tail(master_log,1)$commit)
  expect_equal(git_merge_analysis('master'), 'fastforward')

  # Expect no merge commit (ffwd)
  git_merge("master")
  expect_equal(master_log, git_log())

  # Create a branch that diverges
  git_branch_create('new2', first_commit)
  git_branch_checkout('new2')
  writeLines('Something else', 'test2.txt')
  git_add('test2.txt')
  git_commit("Some other commit")
  expect_equal(git_merge_find_base('master'), first_commit)

  # Require merge commit
  expect_equal(git_merge_analysis('master'), 'normal')
  git_merge('master')
  newlog <- git_log()
  expect_length(newlog$commit, 3)
  expect_equal(newlog$merge, c(TRUE, FALSE, FALSE))
  expect_null(git_merge_parent_heads())

  # Expect no changes
  git_merge('master')
  expect_equal(git_log(), newlog)
})

test_that("merge analysis works", {
  repo <- git_init(tempfile("gert-tests-merge"))
  on.exit(unlink(repo, recursive = TRUE))
  oldwd <- getwd()
  on.exit(setwd(oldwd), add = TRUE)
  setwd(repo)
  configure_local_user()
  for (i in 1:5) {
    writeLines(paste('Blabla', i), "test.txt")
    git_add("test.txt")
    git_commit(paste("This is commit number:", i))
  }
  main <- git_branch()
  master_log <- git_log()
  first_commit <- tail(master_log$commit, 1)
  git_branch_create('new', first_commit)
  git_branch_checkout('new')
  expect_equal(git_log()$commit, tail(master_log, 1)$commit)
  expect_equal(git_merge_analysis(main), 'fastforward')

  # Expect no merge commit (ffwd)
  git_merge(main)
  expect_equal(master_log, git_log())

  # Create a branch that diverges
  git_branch_create('new2', first_commit)
  git_branch_checkout('new2')
  writeLines('Something else', 'test2.txt')
  git_add('test2.txt')
  git_commit("Some other commit")
  expect_equal(git_merge_find_base(main), first_commit)

  # Require merge commit
  expect_equal(git_merge_analysis(main), 'normal')
  git_merge(main)
  newlog <- git_log()
  expect_length(newlog$commit, 3)
  expect_equal(newlog$merge, c(TRUE, FALSE, FALSE))
  expect_null(git_merge_parent_heads())

  # Expect no changes
  git_merge(main)
  expect_equal(git_log(), newlog)
})

#' @importFrom openssl write_ssh write_pem write_pkcs1 write_openssh_pem
#' @importFrom credentials ssh_key_info git_credential_forget ssh_read_key
#' @importFrom askpass askpass
make_key_cb <- function(ssh_key = NULL, host = NULL, password = askpass){
  function(){
    if(is.null(ssh_key)){
      ssh_key <- try(ssh_key_info(host = host, auto_keygen = FALSE)$key)
      if(inherits(ssh_key, "try-error"))
        return(NULL)
    }
    key <- tryCatch(ssh_read_key(ssh_key, password = password), error = function(e){
      stop(sprintf("Unable to load key: %s", ssh_key), call. = FALSE)
    })
    # NB: need to add ID for buggy libssh2: https://github.com/libgit2/libgit2/issues/5162
    writeLines(paste(write_ssh(key$pubkey), "git@localhost"), tmp_pub <- tempfile())

    # NB: pkcs1 is the only format that works on all libssh2 configurations
    tmp_key <- if(inherits(key, c("rsa", "dsa", "ecdsa"))){
      write_pkcs1(key, tempfile())
    } else {
      write_openssh_pem(key, tempfile())
    }
    if(.Platform$OS.type == "unix"){
      Sys.chmod(tmp_pub, '0644')
      Sys.chmod(tmp_key, '0400')
    }
    c(tmp_pub, tmp_key, "")
  }
}

#' @importFrom credentials git_credential_ask
make_cred_cb <- function(password = askpass, verbose = TRUE){
  if(!is.character(password) && !is.function(password)){
    stop("Password parameter must be string or callback function")
  }
  function(url, username, retries){
    # Case of hardcoded (string) password
    if(is.character(password)){
      if(!length(username) || is.na(username)){
        stop("To use a hardcoded password, include your username in the URL like 'https://jerry@github.com")
      }
      return(c(username, password))
    }

    # Look for GITHUB_PAT variable
    if(retries < 2){
      github_pat <- Sys.getenv('GITHUB_PAT')
      if(nchar(github_pat) > 0 && grepl('^https?://([^/]*@)?github.com', url)){
        return(c("git", github_pat))
      }
    }

    # Retrieve a password from the credential manager
    if(retries > 2){
      try(git_credential_forget(url))
    }
    cred <- try(git_credential_ask(url, verbose = verbose), silent = !verbose)
    if(inherits(cred, 'git_credential')){
      return(c(cred$username, cred$password))
    }

    # If that doesn't work try to manually prompt
    if(!length(username) || is.na(username)){
      password(sprintf("Please enter username for %s", url))
    }
    pwd <- password(sprintf("Please enter a PAT or password for %s", url))
    as.character(c(username, pwd))
  }
}

remote_to_host <- function(repo, remote){
  rms <- git_remote_list(repo = repo)
  url <- rms[rms$name == remote, ]$url
  if(length(url)){
    url_to_host(url)
  }
}

url_to_host <- function(url){
  credentials:::parse_url(url, allow_ssh = TRUE)[['host']]
}

#' Git Worktrees
#'
#' @description
#' Worktrees represent an alternative location to checkout a branch into. Rather
#' than checking out a branch in your main working tree (which changes the
#' branch you are currently on and forces you to stash any existing work), you
#' can instead check that branch out into a separate _linked worktree_ with its
#' own working tree. Practically, a worktree is just a separate folder that a
#' branch is checked out into, with some extra git metadata that links it back
#' to the main working tree.
#'
#' `git_worktree_list()` returns a data frame of information about the worktrees
#' linked to the main working tree.
#'
#' `git_worktree_exists()` lets you check whether or not a worktree by the name
#' of `name` exists for this `repo`.
#'
#' `git_worktree_path()` returns the file path to the worktree.
#'
#' `git_worktree_add()` creates a new worktree called `name` in the folder
#' pointed to by `path`, and checks `branch` out into it.
#'
#' `git_worktree_remove()` removes a worktree. It does so by deleting the folder
#' provided as the `path` to `git_worktree_add()`, and then cleaning up some git
#' metadata in the main working tree that linked the main working tree to the
#' removed worktree. The `branch` checked out by the worktree is not deleted.
#' Note that this is just a wrapper around `git_worktree_prune()` that sets some
#' desirable defaults for aggressive removal.
#'
#' `git_worktree_prune()` is more cautious than `git_worktree_remove()`. It
#' refuses to prune _valid_ or _locked_ worktrees by default, and also refuses
#' the delete the working tree of the worktree by default (i.e. the folder at
#' `path`). It is automatically run by git itself on periodic intervals to prune
#' outdated worktrees. For interactive usage, you typically want
#' `git_worktree_remove()` instead. `git_worktree_is_prunable()` lets you check
#' if a worktree is prunable with the given options.
#'
#' `git_worktree_lock()`, `git_worktree_unlock()`, and
#' `git_worktree_is_locked()` help you manage whether or not a worktree is
#' _locked_. When a worktree is locked, it is not automatically cleaned up by
#' `git_worktree_prune()` (and git itself) on periodic intervals, even when it
#' looks _invalid_. This is typically only useful when your worktree is on a
#' hard drive that isn't always connected (which can make it look _invalid_ when
#' disconnected, typically making it a candidate for automatic pruning).
#'
#' `git_worktree_is_valid()` checks whether a worktree is valid or not. A
#' _valid_ worktree requires both the git data structures inside the main
#' working tree and this worktree to be present.
#'
#' @name git_worktree
#' @family git
#' @inheritParams git_open
#'
#' @param name The name of the worktree.
#'
#' @examples
#' repo <- git_init(tempfile("gert-examples-repo"))
#'
#' writeLines("hello", file.path(repo, 'hello.txt'))
#' git_add('hello.txt', repo = repo)
#' git_commit("First commit", author = "jeroen <jeroen@blabla.nl>", repo = repo)
#'
#' # Create a branch that is going to be used for the worktree,
#' # but don't check it out!
#' git_branch_create(branch = "branch", checkout = FALSE, repo = repo)
#'
#' path <- tempfile("gert-examples-worktree")
#'
#' # Add a worktree for this branch
#' git_worktree_add(
#'   name = "worktree",
#'   path = path,
#'   branch = "branch",
#'   repo = repo
#' )
#'
#' # Worktree info
#' git_worktree_list(repo = repo)
#'
#' # Note how the files are checked out here
#' dir(path, all.files = TRUE)
#'
#' # And the branch that we are on at `path` is `"branch"`
#' git_branch(repo = path)
#'
#' # Cleanup worktree, and the folder at `path`
#' git_worktree_remove("worktree", repo = repo)
#'
#' # Cleanup repo
#' unlink(repo, recursive = TRUE)
NULL

#' @export
#' @rdname git_worktree
#' @useDynLib gert R_git_worktree_list
git_worktree_list <- function(repo = '.') {
  repo <- git_open(repo)
  .Call(R_git_worktree_list, repo)
}

#' @export
#' @rdname git_worktree
#' @useDynLib gert R_git_worktree_exists
git_worktree_exists <- function(name, repo = '.') {
  repo <- git_open(repo)
  name <- as.character(name)
  .Call(R_git_worktree_exists, repo, name)
}

#' @export
#' @rdname git_worktree
#' @useDynLib gert R_git_worktree_path
git_worktree_path <- function(name, repo = ".") {
  repo <- git_open(repo)
  name <- as.character(name)
  .Call(R_git_worktree_path, repo, name)
}

#' @export
#' @rdname git_worktree
#' @inheritParams git_branch
#' @param path The path to checkout `branch` into. Importantly, the path up to
#'   the folder name must exist, but the folder name itself must not exist yet
#'   and will be created.
#' @param branch The branch to checkout into `path`.
#' @param lock Whether or not to lock the worktree on creation.
#' @useDynLib gert R_git_worktree_add
git_worktree_add <- function(
  name,
  path,
  branch,
  lock = FALSE,
  local = TRUE,
  repo = "."
) {
  repo <- git_open(repo)
  name <- as.character(name)
  path <- normalizePath(path.expand(path), mustWork = FALSE)
  stopifnot(length(path) == 1L)
  # Avoid footguns where worktree metadata can get created, but branch can't be
  # checked out into the worktree, resulting in broken worktree state requiring
  # manual touching of `.git/` files
  if (dir.exists(path)) {
    stop(sprintf("Path at '%s' must not exist.", path))
  }
  if (!dir.exists(dirname(path))) {
    stop(sprintf("Path at '%s' must exist.", dirname(path)))
  }
  branch <- as.character(branch)
  lock <- as.logical(lock)
  if (!is.null(local)) {
    local <- as.logical(local)
  }
  .Call(R_git_worktree_add, repo, name, path, branch, lock, local)
  invisible()
}

#' @export
#' @rdname git_worktree
git_worktree_remove <- function(name, repo = ".") {
  git_worktree_prune(
    name = name,
    repo = repo,
    prune_valid = TRUE,
    prune_locked = TRUE,
    prune_working_tree = TRUE
  )
}

#' @export
#' @rdname git_worktree
#' @param prune_valid Whether or not to forcibly prune a _valid_ worktree.
#' @param prune_locked Whether or not to forcibly prune a _locked_ worktree.
#' @param prune_working_tree Whether or not to also remove the folder that the
#'   worktree was using, i.e. the `path` supplied to `git_worktree_add()`.
#' @useDynLib gert R_git_worktree_prune
git_worktree_prune <- function(
  name,
  prune_valid = FALSE,
  prune_locked = FALSE,
  prune_working_tree = FALSE,
  repo = "."
) {
  repo <- git_open(repo)
  name <- as.character(name)
  prune_valid <- as.logical(prune_valid)
  prune_locked <- as.logical(prune_locked)
  prune_working_tree <- as.logical(prune_working_tree)
  .Call(
    R_git_worktree_prune,
    repo,
    name,
    prune_valid,
    prune_locked,
    prune_working_tree
  )
  invisible()
}

#' @export
#' @rdname git_worktree
#' @useDynLib gert R_git_worktree_is_prunable
git_worktree_is_prunable <- function(
  name,
  prune_valid = FALSE,
  prune_locked = FALSE,
  repo = "."
) {
  repo <- git_open(repo)
  name <- as.character(name)
  prune_valid <- as.logical(prune_valid)
  prune_locked <- as.logical(prune_locked)
  .Call(
    R_git_worktree_is_prunable,
    repo,
    name,
    prune_valid,
    prune_locked
  )
}

#' @export
#' @rdname git_worktree
#' @useDynLib gert R_git_worktree_lock
git_worktree_lock <- function(name, repo = ".") {
  repo <- git_open(repo)
  name <- as.character(name)
  .Call(R_git_worktree_lock, repo, name)
  invisible()
}

#' @export
#' @rdname git_worktree
#' @useDynLib gert R_git_worktree_unlock
git_worktree_unlock <- function(name, repo = ".") {
  repo <- git_open(repo)
  name <- as.character(name)
  .Call(R_git_worktree_unlock, repo, name)
  invisible()
}

#' @export
#' @rdname git_worktree
#' @useDynLib gert R_git_worktree_is_locked
git_worktree_is_locked <- function(name, repo = ".") {
  repo <- git_open(repo)
  name <- as.character(name)
  .Call(R_git_worktree_is_locked, repo, name)
}

#' @export
#' @rdname git_worktree
#' @useDynLib gert R_git_worktree_is_valid
git_worktree_is_valid <- function(name, repo = ".") {
  repo <- git_open(repo)
  name <- as.character(name)
  .Call(R_git_worktree_is_valid, repo, name)
}

test_that("`git_worktree_list()`, `git_worktree_exists()`, and `git_worktree_path()` work", {
  repo <- git_init(tempfile("gert-tests-repo"))
  on.exit(unlink(repo, recursive = TRUE), add = TRUE, after = FALSE)

  # Need `HEAD`
  writeLines("hello", file.path(repo, 'hello.txt'))
  git_add('hello.txt', repo = repo)
  git_commit("First commit", author = "jeroen <jeroen@blabla.nl>", repo = repo)

  git_branch_create(branch = "branch", checkout = FALSE, repo = repo)

  git_worktree_add(
    name = "worktree",
    path = tempfile("gert-tests-worktree"),
    branch = "branch",
    repo = repo
  )
  on.exit(
    git_worktree_remove("worktree", repo = repo),
    add = TRUE,
    after = FALSE
  )

  # `git_worktree_list()`
  worktrees <- git_worktree_list(repo = repo)
  expect_identical(worktrees$name, "worktree")
  # Exact path may be normalized by libgit2
  expect_type(worktrees$path, "character")
  expect_identical(worktrees$valid, TRUE)
  expect_identical(worktrees$locked, FALSE)

  # `git_worktree_exists()`
  expect_false(git_worktree_exists("fake", repo = repo))
  expect_true(git_worktree_exists("worktree", repo = repo))

  # `git_worktree_path()`
  # Exact path may be normalized by libgit2
  path <- git_worktree_path("worktree", repo = repo)
  expect_type(path, "character")
  expect_length(path, 1)
})

test_that("`git_worktree_lock()`, `git_worktree_unlock()`, and `git_worktree_is_locked()` work", {
  repo <- git_init(tempfile("gert-tests-repo"))
  on.exit(unlink(repo, recursive = TRUE), add = TRUE, after = FALSE)

  # Need `HEAD`
  writeLines("hello", file.path(repo, 'hello.txt'))
  git_add('hello.txt', repo = repo)
  git_commit("First commit", author = "jeroen <jeroen@blabla.nl>", repo = repo)

  git_branch_create(branch = "branch1", checkout = FALSE, repo = repo)
  git_branch_create(branch = "branch2", checkout = FALSE, repo = repo)

  # Unlocked on creation
  git_worktree_add(
    name = "worktree1",
    path = tempfile("gert-tests-worktree1"),
    branch = "branch1",
    lock = FALSE,
    repo = repo
  )
  on.exit(
    git_worktree_remove("worktree1", repo = repo),
    add = TRUE,
    after = FALSE
  )

  # Locked on creation
  git_worktree_add(
    name = "worktree2",
    path = tempfile("gert-tests-worktree2"),
    branch = "branch2",
    lock = TRUE,
    repo = repo
  )
  on.exit(
    git_worktree_remove("worktree2", repo = repo),
    add = TRUE,
    after = FALSE
  )

  expect_false(git_worktree_is_locked("worktree1", repo = repo))
  git_worktree_lock("worktree1", repo = repo)
  expect_true(git_worktree_is_locked("worktree1", repo = repo))
  git_worktree_unlock("worktree1", repo = repo)
  expect_false(git_worktree_is_locked("worktree1", repo = repo))

  expect_true(git_worktree_is_locked("worktree2", repo = repo))
  git_worktree_unlock("worktree2", repo = repo)
  expect_false(git_worktree_is_locked("worktree2", repo = repo))
  git_worktree_lock("worktree2", repo = repo)
  expect_true(git_worktree_is_locked("worktree2", repo = repo))
})

test_that("`git_worktree_is_valid()` works", {
  skip_if_not(libgit2_config()$version > "1.1.0", "Skipping Debian 11")

  repo <- git_init(tempfile("gert-tests-repo"))
  on.exit(unlink(repo, recursive = TRUE), add = TRUE, after = FALSE)

  # Need `HEAD`
  writeLines("hello", file.path(repo, 'hello.txt'))
  git_add('hello.txt', repo = repo)
  git_commit("First commit", author = "jeroen <jeroen@blabla.nl>", repo = repo)

  git_branch_create(branch = "branch", checkout = FALSE, repo = repo)

  path <- tempfile("gert-tests-worktree")

  git_worktree_add(
    name = "worktree",
    path = path,
    branch = "branch",
    repo = repo
  )

  expect_true(git_worktree_is_valid("worktree", repo = repo))

  # Remove the worktree's folder
  expect_true(unlink(path, recursive = TRUE) == 0L)
  expect_false(git_worktree_is_valid("worktree", repo = repo))

  # Still exists even though it is invalid
  expect_true(git_worktree_exists("worktree", repo = repo))

  # Now prune it
  expect_true(git_worktree_is_prunable("worktree", repo = repo))
  git_worktree_prune("worktree", repo = repo)
  expect_false(git_worktree_exists("worktree", repo = repo))
})

test_that("`git_worktree_add()` requires that `branch` exist", {
  repo <- git_init(tempfile("gert-tests-repo"))
  on.exit(unlink(repo, recursive = TRUE), add = TRUE, after = FALSE)

  # Need `HEAD`
  writeLines("hello", file.path(repo, 'hello.txt'))
  git_add('hello.txt', repo = repo)
  git_commit("First commit", author = "jeroen <jeroen@blabla.nl>", repo = repo)

  expect_error(
    git_worktree_add(
      name = "worktree",
      path = tempfile("gert-tests-worktree"),
      branch = "branch",
      repo = repo
    ),
    "cannot locate local branch 'branch'"
  )
})

test_that("`git_worktree_add()` can't check out already checked out branch", {
  repo <- git_init(tempfile("gert-tests-repo"))
  on.exit(unlink(repo, recursive = TRUE), add = TRUE, after = FALSE)

  # Need `HEAD`
  writeLines("hello", file.path(repo, 'hello.txt'))
  git_add('hello.txt', repo = repo)
  git_commit("First commit", author = "jeroen <jeroen@blabla.nl>", repo = repo)

  # Check out `branch`
  git_branch_create(branch = "branch", checkout = TRUE, repo = repo)

  expect_error(
    git_worktree_add(
      name = "worktree",
      path = tempfile("gert-tests-worktree"),
      branch = "branch",
      repo = repo
    ),
    "already checked out"
  )
})

test_that("`git_worktree_add()` errors early if `path` already exists", {
  # We error at R level to avoid libgit2 putting us in a corrupt worktree state

  repo <- git_init(tempfile("gert-tests-repo"))
  on.exit(unlink(repo, recursive = TRUE), add = TRUE, after = FALSE)

  # `path` exists, but should not!
  path <- tempfile("gert-tests-worktree")
  dir.create(path)

  expect_error(
    git_worktree_add(
      name = "worktree",
      path = path,
      branch = "branch",
      repo = repo
    ),
    "must not exist"
  )
})

test_that("`git_worktree_add()` errors early if `dirname(path)` does not exist", {
  # We error at R level to avoid libgit2 putting us in a corrupt worktree state

  repo <- git_init(tempfile("gert-tests-repo"))
  on.exit(unlink(repo, recursive = TRUE), add = TRUE, after = FALSE)

  # `dir` does not exist, but needs to
  dir <- tempfile("gert-tests-worktree-directory")
  path <- tempfile("gert-tests-worktree", tmpdir = dir)

  expect_error(
    git_worktree_add(
      name = "worktree",
      path = path,
      branch = "branch",
      repo = repo
    ),
    "must exist"
  )
})

test_that("`git_worktree_remove()` removes worktree and `path` folder", {
  repo <- git_init(tempfile("gert-tests-repo"))
  on.exit(unlink(repo, recursive = TRUE), add = TRUE, after = FALSE)

  # Need `HEAD`
  writeLines("hello", file.path(repo, 'hello.txt'))
  git_add('hello.txt', repo = repo)
  git_commit("First commit", author = "jeroen <jeroen@blabla.nl>", repo = repo)

  git_branch_create(branch = "branch", checkout = FALSE, repo = repo)

  path <- tempfile("gert-tests-worktree")

  git_worktree_add(
    name = "worktree",
    path = path,
    branch = "branch",
    repo = repo
  )

  expect_true(dir.exists(path))
  expect_true(git_worktree_exists("worktree", repo = repo))

  git_worktree_remove("worktree", repo = repo)

  expect_false(dir.exists(path))
  expect_false(git_worktree_exists("worktree", repo = repo))
})

test_that("`git_worktree_prune()` refuses to remove `valid` worktrees", {
  repo <- git_init(tempfile("gert-tests-repo"))
  on.exit(unlink(repo, recursive = TRUE), add = TRUE, after = FALSE)

  # Need `HEAD`
  writeLines("hello", file.path(repo, 'hello.txt'))
  git_add('hello.txt', repo = repo)
  git_commit("First commit", author = "jeroen <jeroen@blabla.nl>", repo = repo)

  git_branch_create(branch = "branch", checkout = FALSE, repo = repo)

  git_worktree_add(
    name = "worktree",
    path = tempfile("gert-tests-worktree"),
    branch = "branch",
    repo = repo
  )

  # Refuses to prune by default
  expect_false(git_worktree_is_prunable("worktree", repo = repo))
  expect_error(
    git_worktree_prune("worktree", repo = repo),
    "not pruning valid working tree"
  )

  # Can be forced to prune
  expect_no_error(git_worktree_prune(
    "worktree",
    repo = repo,
    prune_valid = TRUE
  ))
})

test_that("`git_worktree_prune()` refuses to remove `locked` worktrees", {
  repo <- git_init(tempfile("gert-tests-repo"))
  on.exit(unlink(repo, recursive = TRUE), add = TRUE, after = FALSE)

  # Need `HEAD`
  writeLines("hello", file.path(repo, 'hello.txt'))
  git_add('hello.txt', repo = repo)
  git_commit("First commit", author = "jeroen <jeroen@blabla.nl>", repo = repo)

  git_branch_create(branch = "branch", checkout = FALSE, repo = repo)

  git_worktree_add(
    name = "worktree",
    path = tempfile("gert-tests-worktree"),
    branch = "branch",
    lock = TRUE,
    repo = repo
  )

  # Refuses to prune by default
  expect_false(git_worktree_is_prunable("worktree", repo = repo))
  expect_error(
    git_worktree_prune("worktree", repo = repo),
    "not pruning locked working tree"
  )

  # Refuses to prune valid worktree by default as well
  expect_error(
    git_worktree_prune("worktree", repo = repo, prune_locked = TRUE),
    "not pruning valid working tree"
  )

  # Can be forced to prune with both options
  expect_no_error(git_worktree_prune(
    "worktree",
    repo = repo,
    prune_valid = TRUE,
    prune_locked = TRUE
  ))
})

check_ref_in_history <- function(ref, repo) {
  sha <- tryCatch(git_commit_id(ref, repo = repo), error = function(e) {
    stop(sprintf("Cannot resolve '%s' to a commit", ref))
  })

  head_sha <- git_commit_id("HEAD", repo = repo)
  sha_descends_from_head <- git_commit_descendant_of(
    ancestor = sha,
    ref = "HEAD",
    repo = repo
  )
  if (sha != head_sha && !sha_descends_from_head) {
    stop(sprintf("'%s' is not in the current branch history", ref))
  }

  sha
}

#' Git Diff
#'
#' View changes in a commit or in the current working directory.
#'
#' @export
#' @rdname git_diff
#' @name git_diff
#' @inheritParams git_open
#' @family git
#' @param ref a reference such as `"HEAD"`, or a commit id, or `NULL`
#' to the diff the working directory against the repository index.
#' @useDynLib gert R_git_diff_list
git_diff <- function(ref = NULL, repo = '.'){
  repo <- git_open(repo)
  ref <- as.character(ref)
  .Call(R_git_diff_list, repo, ref)
}

#' @export
#' @rdname git_diff
git_diff_patch <- function(ref = NULL, repo = '.'){
  git_diff(ref = ref, repo = repo)$patch
}

#' Git Worktrees
#'
#' @description
#' Worktrees represent an alternative location to checkout a branch into. Rather
#' than checking out a branch in your main working tree (which changes the
#' branch you are currently on and forces you to stash any existing work), you
#' can instead check that branch out into a separate _linked worktree_ with its
#' own working tree. Practically, a worktree is just a separate folder that a
#' branch is checked out into, with some extra git metadata that links it back
#' to the main working tree.
#'
#' `git_worktree_list()` returns a data frame of information about the worktrees
#' linked to the main working tree.
#'
#' `git_worktree_exists()` lets you check whether or not a worktree by the name
#' of `name` exists for this `repo`.
#'
#' `git_worktree_path()` returns the file path to the worktree.
#'
#' `git_worktree_add()` creates a new worktree called `name` in the folder
#' pointed to by `path`, and checks `branch` out into it.
#'
#' `git_worktree_remove()` removes a worktree. It does so by deleting the folder
#' provided as the `path` to `git_worktree_add()`, and then cleaning up some git
#' metadata in the main working tree that linked the main working tree to the
#' removed worktree. The `branch` checked out by the worktree is not deleted.
#' Note that this is just a wrapper around `git_worktree_prune()` that sets some
#' desirable defaults for aggressive removal.
#'
#' `git_worktree_prune()` is more cautious than `git_worktree_remove()`. It
#' refuses to prune _valid_ or _locked_ worktrees by default, and also refuses
#' the delete the working tree of the worktree by default (i.e. the folder at
#' `path`). It is automatically run by git itself on periodic intervals to prune
#' outdated worktrees. For interactive usage, you typically want
#' `git_worktree_remove()` instead. `git_worktree_is_prunable()` lets you check
#' if a worktree is prunable with the given options.
#'
#' `git_worktree_lock()`, `git_worktree_unlock()`, and
#' `git_worktree_is_locked()` help you manage whether or not a worktree is
#' _locked_. When a worktree is locked, it is not automatically cleaned up by
#' `git_worktree_prune()` (and git itself) on periodic intervals, even when it
#' looks _invalid_. This is typically only useful when your worktree is on a
#' hard drive that isn't always connected (which can make it look _invalid_ when
#' disconnected, typically making it a candidate for automatic pruning).
#'
#' `git_worktree_is_valid()` checks whether a worktree is valid or not. A
#' _valid_ worktree requires both the git data structures inside the main
#' working tree and this worktree to be present.
#'
#' @name git_worktree
#' @family git
#' @inheritParams git_open
#' @git worktree
#'
#' @param name The name of the worktree.
#'
#' @examples
#' repo <- git_init(tempfile("gert-examples-repo"))
#'
#' writeLines("hello", file.path(repo, 'hello.txt'))
#' git_add('hello.txt', repo = repo)
#' git_commit("First commit", author = "jeroen <jeroen@blabla.nl>", repo = repo)
#'
#' # Create a branch that is going to be used for the worktree,
#' # but don't check it out!
#' git_branch_create(branch = "branch", checkout = FALSE, repo = repo)
#'
#' path <- tempfile("gert-examples-worktree")
#'
#' # Add a worktree for this branch
#' git_worktree_add(
#'   name = "worktree",
#'   path = path,
#'   branch = "branch",
#'   repo = repo
#' )
#'
#' # Worktree info
#' git_worktree_list(repo = repo)
#'
#' # Note how the files are checked out here
#' dir(path, all.files = TRUE)
#'
#' # And the branch that we are on at `path` is `"branch"`
#' git_branch(repo = path)
#'
#' # Cleanup worktree, and the folder at `path`
#' git_worktree_remove("worktree", repo = repo)
#'
#' # Cleanup repo
#' unlink(repo, recursive = TRUE)
NULL

#' @export
#' @rdname git_worktree
#' @useDynLib gert R_git_worktree_list
git_worktree_list <- function(repo = '.') {
  repo <- git_open(repo)
  .Call(R_git_worktree_list, repo)
}

#' @export
#' @rdname git_worktree
#' @useDynLib gert R_git_worktree_exists
git_worktree_exists <- function(name, repo = '.') {
  repo <- git_open(repo)
  name <- as.character(name)
  .Call(R_git_worktree_exists, repo, name)
}

#' @export
#' @rdname git_worktree
#' @useDynLib gert R_git_worktree_path
git_worktree_path <- function(name, repo = ".") {
  repo <- git_open(repo)
  name <- as.character(name)
  .Call(R_git_worktree_path, repo, name)
}

#' @export
#' @rdname git_worktree
#' @inheritParams git_branch
#' @param path The path to checkout `branch` into. Importantly, the path up to
#'   the folder name must exist, but the folder name itself must not exist yet
#'   and will be created.
#' @param branch The branch to checkout into `path`.
#' @param lock Whether or not to lock the worktree on creation.
#' @useDynLib gert R_git_worktree_add
git_worktree_add <- function(
  name,
  path,
  branch,
  lock = FALSE,
  local = TRUE,
  repo = "."
) {
  repo <- git_open(repo)
  name <- as.character(name)
  path <- normalizePath(path.expand(path), mustWork = FALSE)
  stopifnot(length(path) == 1L)
  # Avoid footguns where worktree metadata can get created, but branch can't be
  # checked out into the worktree, resulting in broken worktree state requiring
  # manual touching of `.git/` files
  if (dir.exists(path)) {
    stop(sprintf("Path at '%s' must not exist.", path))
  }
  if (!dir.exists(dirname(path))) {
    stop(sprintf("Path at '%s' must exist.", dirname(path)))
  }
  branch <- as.character(branch)
  lock <- as.logical(lock)
  if (!is.null(local)) {
    local <- as.logical(local)
  }
  .Call(R_git_worktree_add, repo, name, path, branch, lock, local)
  invisible()
}

#' @export
#' @rdname git_worktree
git_worktree_remove <- function(name, repo = ".") {
  git_worktree_prune(
    name = name,
    repo = repo,
    prune_valid = TRUE,
    prune_locked = TRUE,
    prune_working_tree = TRUE
  )
}

#' @export
#' @rdname git_worktree
#' @param prune_valid Whether or not to forcibly prune a _valid_ worktree.
#' @param prune_locked Whether or not to forcibly prune a _locked_ worktree.
#' @param prune_working_tree Whether or not to also remove the folder that the
#'   worktree was using, i.e. the `path` supplied to `git_worktree_add()`.
#' @useDynLib gert R_git_worktree_prune
git_worktree_prune <- function(
  name,
  prune_valid = FALSE,
  prune_locked = FALSE,
  prune_working_tree = FALSE,
  repo = "."
) {
  repo <- git_open(repo)
  name <- as.character(name)
  prune_valid <- as.logical(prune_valid)
  prune_locked <- as.logical(prune_locked)
  prune_working_tree <- as.logical(prune_working_tree)
  .Call(
    R_git_worktree_prune,
    repo,
    name,
    prune_valid,
    prune_locked,
    prune_working_tree
  )
  invisible()
}

#' @export
#' @rdname git_worktree
#' @useDynLib gert R_git_worktree_is_prunable
git_worktree_is_prunable <- function(
  name,
  prune_valid = FALSE,
  prune_locked = FALSE,
  repo = "."
) {
  repo <- git_open(repo)
  name <- as.character(name)
  prune_valid <- as.logical(prune_valid)
  prune_locked <- as.logical(prune_locked)
  .Call(
    R_git_worktree_is_prunable,
    repo,
    name,
    prune_valid,
    prune_locked
  )
}

#' @export
#' @rdname git_worktree
#' @useDynLib gert R_git_worktree_lock
git_worktree_lock <- function(name, repo = ".") {
  repo <- git_open(repo)
  name <- as.character(name)
  .Call(R_git_worktree_lock, repo, name)
  invisible()
}

#' @export
#' @rdname git_worktree
#' @useDynLib gert R_git_worktree_unlock
git_worktree_unlock <- function(name, repo = ".") {
  repo <- git_open(repo)
  name <- as.character(name)
  .Call(R_git_worktree_unlock, repo, name)
  invisible()
}

#' @export
#' @rdname git_worktree
#' @useDynLib gert R_git_worktree_is_locked
git_worktree_is_locked <- function(name, repo = ".") {
  repo <- git_open(repo)
  name <- as.character(name)
  .Call(R_git_worktree_is_locked, repo, name)
}

#' @export
#' @rdname git_worktree
#' @useDynLib gert R_git_worktree_is_valid
git_worktree_is_valid <- function(name, repo = ".") {
  repo <- git_open(repo)
  name <- as.character(name)
  .Call(R_git_worktree_is_valid, repo, name)
}

#' Git Tag
#'
#' Create and list tags.
#'
#' @export
#' @rdname git_tag
#' @name git_tag
#' @family git
#' @inheritParams git_open
#' @param match pattern to filter tags (use `*` for wildcard)
#' @useDynLib gert R_git_tag_list
git_tag_list <- function(match = "*", repo = '.'){
  repo <- git_open(repo)
  match <- as.character(match)
  .Call(R_git_tag_list, repo, match)
}

#' @export
#' @rdname git_tag
#' @param name tag name
#' @param message tag message
#' @param ref target reference to tag
#' @useDynLib gert R_git_tag_create
git_tag_create <- function(name, message, ref = "HEAD", repo = '.'){
  repo <- git_open(repo)
  name <- as.character(name)
  message <- as.character(message)
  .Call(R_git_tag_create, repo, name, message, ref)
}

#' @export
#' @rdname git_tag
#' @useDynLib gert R_git_tag_delete
git_tag_delete <- function(name, repo = '.'){
  repo <- git_open(repo)
  name <- as.character(name)
  .Call(R_git_tag_delete, repo, name)
}

#' @export
#' @rdname git_tag
#' @param ... other arguments passed to [git_push]
git_tag_push <- function(name, ..., repo = '.'){
  ref <- paste0('refs/tags/', name)
  git_push(refspec = ref, ..., repo = repo)
}

#' Author Signature
#'
#' A signature contains the author and timestamp of a commit. Each commit
#' includes a signature of the author and committer (which can be identical).
#'
#' A signature string has format `"Real Name <email> timestamp tzoffset"`. The
#' `timestamp tzoffset` piece can be omitted in which case the current local
#' time is used. If not omitted, `timestamp` must contain the number
#' of seconds since the Unix epoch and `tzoffset` is the timezone offset in
#' `hhmm` format (note the lack of a colon separator)
#'
#' @export
#' @rdname git_signature
#' @name git_signature
#' @family git
#' @inheritParams git_open
#' @useDynLib gert R_git_signature_default
#' @examples # Your default user
#' try(git_signature_default())
#'
#' # Specify explicit name and email
#' git_signature("Some committer", "sarah@gmail.com")
#'
#' # Create signature for an hour ago
#' (sig <- git_signature("Han", "han@company.com", Sys.time() - 3600))
#'
#' # Parse a signature
#' git_signature_parse(sig)
#' git_signature_parse("Emma <emma@mu.edu>")
git_signature_default <- function(repo = '.'){
  repo <- git_open(repo)
  sig <- .Call(R_git_signature_default, repo)
  sprintf("%s <%s>", sig$name, sig$email)
}

#' @export
#' @rdname git_signature
#' @useDynLib gert R_git_signature_create
#' @param name Real name of the committer
#' @param email Email address of the committer
#' @param time timestamp of class POSIXt or NULL
git_signature <- function(name, email, time = NULL){
  assert_string(name)
  assert_string(email)
  if(length(time)){
    time <- as.POSIXct(time)
    tz <- format(time, "%z")
    minutes <- as.integer(substring(tz, 4))
    hours <- as.integer(substring(tz, 1, 3))
    offset <- hours * 60 + minutes
  }
  sig <- .Call(R_git_signature_create, name, email, time, offset)
  if(length(time)){
    sig_data_to_string(sig)
  } else {
    sig_data_to_string(list(name = sig$name, email = sig$email))
  }
}

#' @export
#' @rdname git_signature
#' @param sig string in proper `"First Last <your@email.com>"` format, see details.
#' @useDynLib gert R_git_signature_parse
git_signature_parse <- function(sig){
  assert_string(sig)
  .Call(R_git_signature_parse, sig)
}

offset_to_string <- function(offset){
  if(length(offset) && is.numeric(offset)){
    hours <- as.integer(offset %/% 60)
    mins <- as.integer(offset %% 60)
    sprintf('%+03d%02d', hours, mins)
  } else ""
}

sig_data_to_string <- function(x){
  sig <- sprintf("%s <%s>", x$name, x$email)
  if(length(x$time)){
    sig <- paste(sig, unclass(x$time), offset_to_string(x$offset))
  }
  structure(trimws(sig), class = "gert_signature")
}

#' @export
print.gert_signature <- function(x, ...){
  sig <- git_signature_parse(x)
  name <- sig$name
  email <- sig$email
  time <- format(sig$time, "%a %b %d %H:%M:%S %Y") #print as user local time
  offset <- offset_to_string(sig$offset)
  cat(sprintf('[git signature]\nAuthor: %s <%s>\nDate: %s %s\n', name, email, time, offset))
}

#' @importFrom openssl write_ssh write_pem write_pkcs1 write_openssh_pem
#' @importFrom credentials ssh_key_info git_credential_forget ssh_read_key
#' @importFrom askpass askpass
make_key_cb <- function(ssh_key = NULL, host = NULL, password = askpass) {
  function() {
    try({
      if (is.null(ssh_key)) {
        ssh_key <- try(ssh_key_info(host = host, auto_keygen = FALSE)$key)
        if (inherits(ssh_key, "try-error")) return(NULL)
      }
      key <- tryCatch(
        ssh_read_key(ssh_key, password = password),
        error = function(e) {
          stop(sprintf("Unable to load key: %s", ssh_key), call. = FALSE)
        }
      )
      # NB: need to add ID for buggy libssh2: https://github.com/libgit2/libgit2/issues/5162
      writeLines(
        paste(write_ssh(key$pubkey), "git@localhost"),
        tmp_pub <- tempfile()
      )

      # NB: pkcs1 is the only format that works on all libssh2 configurations
      tmp_key <- if (inherits(key, c("rsa", "dsa", "ecdsa"))) {
        write_pkcs1(key, tempfile())
      } else {
        write_openssh_pem(key, tempfile())
      }
      if (.Platform$OS.type == "unix") {
        Sys.chmod(tmp_pub, '0644')
        Sys.chmod(tmp_key, '0400')
      }
      c(tmp_pub, tmp_key, "")
    })
  }
}

#' @importFrom credentials git_credential_ask
make_cred_cb <- function(password = askpass, verbose = TRUE) {
  if (!is.character(password) && !is.function(password)) {
    stop("Password parameter must be string or callback function")
  }
  function(url, username, retries) {
    # Case of hardcoded (string) password
    if (is.character(password)) {
      if (!length(username) || is.na(username)) {
        stop(
          "To use a hardcoded password, include your username in the URL like 'https://jerry@github.com"
        )
      }
      return(c(username, password))
    }

    # Look for GITHUB_PAT variable
    if (retries < 2) {
      github_pat <- Sys.getenv('GITHUB_PAT')
      if (
        nchar(github_pat) > 0 && grepl('^https?://([^/]*@)?github.com', url)
      ) {
        return(c("git", github_pat))
      }
    }

    # Retrieve a password from the credential manager
    if (retries > 2) {
      try(git_credential_forget(url))
    }
    cred <- try(git_credential_ask(url, verbose = verbose), silent = !verbose)
    if (inherits(cred, 'git_credential')) {
      return(c(cred$username, cred$password))
    }

    # If that doesn't work try to manually prompt
    if (!length(username) || is.na(username)) {
      password(sprintf("Please enter username for %s", url))
    }
    pwd <- password(sprintf("Please enter a PAT or password for %s", url))
    as.character(c(username, pwd))
  }
}

remote_to_host <- function(repo, remote) {
  rms <- git_remote_list(repo = repo)
  url <- rms[rms$name == remote, ]$url
  if (length(url)) {
    url_to_host(url)
  }
}

url_to_host <- function(url) {
  #credentials:::parse_url(url, allow_ssh = TRUE)[['host']]
  parse_url <- utils::getFromNamespace('parse_url', 'credentials')
  parse_url(url, allow_ssh = TRUE)[['host']]
}

#' Create or discover a local Git repository
#'
#' * `git_init()` creates a new repository
#' * `git_find()` to discover an existing local repository.
#' * `git_info()` shows basic information about a repository, such as the SHA
#'   and branch of the current HEAD.
#'
#' @section Path:
#'
#' For `git_init()` the `path` parameter sets the directory of the git repository
#' to create. If this directory already exists, it must be empty. If it does
#' not exist, it is created, along with any intermediate directories that don't
#' yet exist.
#'
#' For `git_find()`, the `path` parameter specifies the directory at
#' which to start the search for a git repository. If it is not a git repository
#' itself, then its parent directory is consulted, then the parent's parent, and
#' so on.
#'
#' @export
#' @rdname git_repo
#' @name git_repo
#' @family git
#' @useDynLib gert R_git_repository_init
#' @inheritParams git_open
#' @param path the location of the git repository, see the "path" section.
#' @param bare if true, a Git repository without a working directory is created
#' @git repository
#' @return
#' * `git_find()` and `git_init()`: the path to the Git repository.
#' * `git_info()`: A list of information of the Git repository.
#' @examples
#' # directory does not yet exist
#' r <- tempfile(pattern = "gert")
#' git_init(r)
#' git_find(r)
#' git_info(r)
#'
#' # create a child directory, then a grandchild, then search
#' r_grandchild_dir <- file.path(r, "aaa", "bbb")
#' dir.create(r_grandchild_dir, recursive = TRUE)
#' git_find(r_grandchild_dir)
#'
#' # cleanup
#' unlink(r, recursive = TRUE)
#'
#' # directory exists but is empty
#' r <- tempfile(pattern = "gert")
#' dir.create(r)
#' git_init(r)
#' git_find(r)
#'
#' # cleanup
#' unlink(r, recursive = TRUE)
git_init <- function(path = '.', bare = FALSE) {
  path <- normalizePath(path.expand(path), mustWork = FALSE)
  repo <- .Call(R_git_repository_init, path, as.logical(bare))
  git_repo_path(repo)
}

#' @export
#' @rdname git_repo
#' @useDynLib gert R_git_repository_find
git_find <- function(path = '.') {
  path <- normalizePath(path.expand(path), mustWork = FALSE)
  out <- .Call(R_git_repository_find, path)
  dirname(out)
}

#' @export
#' @rdname git_repo
#' @useDynLib gert R_git_repository_info
#' @section Detached head:
#' If `git_info()$shorthand` is equal to `HEAD`,
#' it means the repository is in a [detached head state](https://jvns.ca/blog/2023/11/01/confusing-git-terminology/#detached-head-state).
git_info <- function(repo = '.') {
  repo <- git_open(repo)
  .Call(R_git_repository_info, repo)
}

#' Git Remotes
#'
#' List, add, configure, or remove remotes.
#'
#' @export
#' @rdname git_remote
#' @name git_remote
#' @family git
#' @inheritParams git_open
#' @param remote name of an existing remote. Default `NULL` means the remote
#' from the upstream of the current branch.
#' @param url server url (https or ssh)
#' @useDynLib gert R_git_remote_list
git_remote_list <- function(repo = '.'){
  repo <- git_open(repo)
  .Call(R_git_remote_list, repo)
}

#' @export
#' @rdname git_remote
#' @param name unique name for the new remote
#' @param refspec optional string with the remote fetch value
#' @useDynLib gert R_git_remote_add
git_remote_add <- function(url, name = "origin", refspec = NULL, repo = '.'){
  repo <- git_open(repo)
  name <- as.character(name)
  url <- as.character(url)
  refspec <- as.character(refspec)
  invisible(.Call(R_git_remote_add, repo, name, url, refspec))
}

#' @export
#' @rdname git_remote
#' @useDynLib gert R_git_remote_remove
git_remote_remove <- function(remote, repo = '.'){
  repo <- git_open(repo)
  remote <- as.character(remote)
  .Call(R_git_remote_remove, repo, remote)
  invisible()
}

#' @export
#' @rdname git_remote
#' @useDynLib gert R_git_remote_info
git_remote_info <- function(remote = NULL, repo = '.'){
  repo <- git_open(repo)
  remote <- as.character(remote)
  if(!length(remote))
    remote <- git_info(repo = repo)$remote
  .Call(R_git_remote_info, repo, remote)
}

#' @export
#' @rdname git_remote
#' @useDynLib gert R_git_remote_set_url
git_remote_set_url <- function(url, remote = NULL, repo = '.'){
  repo <- git_open(repo)
  name <- as.character(remote)
  if(!length(remote))
    remote <- git_info(repo = repo)$remote
  url <- as.character(url)
  .Call(R_git_remote_set_url, repo, remote, url)
  invisible()
}

#' @export
#' @rdname git_remote
#' @useDynLib gert R_git_remote_set_pushurl
git_remote_set_pushurl <- function(url, remote = NULL , repo = '.'){
  repo <- git_open(repo)
  remote <- as.character(remote)
  if(!length(remote))
    remote <- git_info(repo = repo)$remote
  url <- as.character(url)
  .Call(R_git_remote_set_pushurl, repo, remote, url)
  invisible()
}

#' @export
#' @rdname git_remote
#' @useDynLib gert R_git_remote_refspecs
git_remote_refspecs <- function(remote = NULL, repo = '.'){
  repo <- git_open(repo)
  remote <- as.character(remote)
  if(!length(remote))
    remote <- git_info(repo = repo)$remote
  .Call(R_git_remote_refspecs, repo, remote)
}

#' @useDynLib gert R_git_remote_add_fetch
git_remote_add_fetch <- function(refspec, remote = NULL, repo = '.'){
  repo <- git_open(repo)
  remote <- as.character(remote)
  if(!length(remote))
    remote <- git_info(repo = repo)$remote
  refspec <- as.character(refspec)
  .Call(R_git_remote_add_fetch, repo, remote, refspec)
}

test_that("cloning repositories works", {
  skip_if_offline('github.com')
  path <- file.path(tempdir(), 'gert')
  repo <- git_clone('https://github.com/r-lib/gert', path = path)
  expect_true(file.exists(file.path(path, 'DESCRIPTION')))
  info <- git_info(repo)
  default_head <- git_remote_ls('https://github.com/r-lib/gert', repo = repo)$symref[1]
  default_branch <- basename(default_head)
  expect_equal(info$head, default_head)
  expect_equal(info$shorthand, default_branch)
  repo2 <- git_open(path)
  info2 <- git_info(repo2)
  expect_equal(info, info2)
  expect_is(git_ls(repo), 'data.frame')
  expect_is(git_log(repo = repo), 'data.frame')
  heads <- git_remote_ls(repo = repo)
  expect_is(heads, 'data.frame')
  expect_equal(git_remote_info(repo = repo)$head, paste0("refs/remotes/origin/", default_branch))

  # Test remotes
  remotes <- git_remote_list(repo)
  expect_equal(remotes$name, "origin")
  expect_equal(remotes$url, "https://github.com/r-lib/gert")

  # Test archive
  expect_equal(git_archive_zip(repo = repo), 'gert.zip')
  expect_equal(zip::zip_list('gert.zip')$filename, git_ls(repo=repo)$path)
  unlink('gert.zip')
})

test_that("cloning repositories works", {
  skip_if_offline('github.com')
  path <- file.path(tempdir(), 'gert')
  repo <- git_clone('https://github.com/r-lib/gert', path = path)
  expect_true(file.exists(file.path(path, 'DESCRIPTION')))
  info <- git_info(repo)
  default_head <- git_remote_ls('https://github.com/r-lib/gert', repo = repo)$symref[1]
  default_branch <- basename(default_head)
  expect_equal(info$head, default_head)
  expect_equal(info$shorthand, default_branch)
  repo2 <- git_open(path)
  info2 <- git_info(repo2)
  expect_equal(info, info2)
  expect_is(git_ls(repo = repo), 'data.frame')
  expect_is(git_log(repo = repo), 'data.frame')
  heads <- git_remote_ls(repo = repo)
  expect_is(heads, 'data.frame')
  expect_equal(git_remote_info(repo = repo)$head, paste0("refs/remotes/origin/", default_branch))

  # Test remotes
  remotes <- git_remote_list(repo)
  expect_equal(remotes$name, "origin")
  expect_equal(remotes$url, "https://github.com/r-lib/gert")

  # Test archive
  expect_equal(git_archive_zip(repo = repo), 'gert.zip')
  expect_equal(zip::zip_list('gert.zip')$filename, git_ls(repo=repo)$path)
  unlink('gert.zip')
})

test_that("remotes from new repo",{
  skip_if_offline('github.com')
  repo <- git_init(tempfile("gert-tests-remote"))
  on.exit(unlink(repo, recursive = TRUE))
  expect_equal(nrow(git_remote_list(repo = repo)), 0)
  expect_error(git_remote_info(repo = repo))
  expect_error(git_remote_refspecs(repo = repo))
  expect_error(git_remote_set_url('https://github.com/foo/bar', repo = repo))
  expect_error(git_remote_set_pushurl('https://github.com/foo/bar', repo = repo))
  expect_equal(git_remote_add('https://github.com/jeroen/webp', name = 'jeroen', repo = repo), 'jeroen')
  git_fetch('jeroen', repo = repo)
  git_branch_create('master', 'jeroen/master', repo = repo)
  git_branch_create('testje', git_commit_id('jeroen/master', repo = repo), checkout = FALSE, repo = repo)
  git_remote_set_pushurl('https://github.com/foo/baz', repo = repo)
  info <- git_remote_info(repo = repo)
  expect_equal(info$name, 'jeroen')
  expect_equal(info$url, "https://github.com/jeroen/webp")
  expect_equal(info$push_url, "https://github.com/foo/baz")
  expect_equal(info$fetch, "+refs/heads/*:refs/remotes/jeroen/*")
  branches <- git_branch_list(repo = repo)
  expect_equal(branches$name, c("master", "testje", "jeroen/master"))
  expect_equal(branches$upstream, c("refs/remotes/jeroen/master", NA, NA))
})

test_that("remotes after clone", {
  skip_if_offline('github.com')
  repo <- file.path(tempdir(), 'gert')
  if(!file.exists(repo)) git_clone('https://github.com/r-lib/gert', path = repo)
  info <- git_remote_info(repo = repo)
  expect_equal(info$name, 'origin')
  expect_equal(info$url, "https://github.com/r-lib/gert")
  expect_equal(info$fetch, "+refs/heads/*:refs/remotes/origin/*")
  refspecs <- git_remote_refspecs(repo = repo)
  expect_equal(refspecs$direction, 'fetch')
  remotelist <- git_remote_list(repo = repo)
  expect_equal(remotelist$name, 'origin')
  expect_equal(remotelist$url, 'https://github.com/r-lib/gert')
  expect_error(git_remote_add('https://github.com/jeroen/gert', repo = repo))
  expect_equal(git_remote_add('https://github.com/jeroen/gert', name = 'myfork', repo = repo), 'myfork')
  remotelist <- git_remote_list(repo = repo)
  expect_equal(sort(remotelist$name), c("myfork", 'origin'))
  expect_equal(sort(remotelist$url), c("https://github.com/jeroen/gert", 'https://github.com/r-lib/gert'))
  expect_equal(git_remote_refspecs('myfork', repo = repo)$refspec, '+refs/heads/*:refs/remotes/myfork/*')
})

test_that("git_remote_set_pushurl with add = TRUE appends push URLs", {
  repo <- git_init(tempfile("gert-tests-pushurl"))
  on.exit(unlink(repo, recursive = TRUE))
  configure_local_user(repo)

  git_remote_add("https://example.com/fetch", name = "origin", repo = repo)

  git_remote_set_pushurl(
    "https://example.com/push1",
    remote = "origin",
    repo = repo
  )
  git_remote_set_pushurl(
    "https://example.com/push2",
    remote = "origin",
    add = TRUE,
    repo = repo
  )

  pushurls <- git_config_get("remote.origin.pushurl", repo = repo)
  expect_setequal(
    pushurls,
    c("https://example.com/push1", "https://example.com/push2")
  )
})

test_that("git_remote_set_pushurl without add replaces push URL", {
  repo <- git_init(tempfile("gert-tests-pushurl-replace"))
  on.exit(unlink(repo, recursive = TRUE))
  configure_local_user(repo)

  git_remote_add("https://example.com/fetch", name = "origin", repo = repo)

  git_remote_set_pushurl(
    "https://example.com/push1",
    remote = "origin",
    repo = repo
  )
  git_remote_set_pushurl(
    "https://example.com/push2",
    remote = "origin",
    repo = repo
  )

  pushurls <- git_config_get("remote.origin.pushurl", repo = repo)
  expect_equal(pushurls, "https://example.com/push2")
})

test_that("remotes from new repo", {
  skip_if_offline('github.com')
  repo <- git_init(tempfile("gert-tests-remote"))
  on.exit(unlink(repo, recursive = TRUE))
  expect_equal(nrow(git_remote_list(repo = repo)), 0)
  expect_error(git_remote_info(repo = repo))
  expect_error(git_remote_refspecs(repo = repo))
  expect_error(git_remote_set_url('https://github.com/foo/bar', repo = repo))
  expect_error(git_remote_info(repo = repo), "remote 'NA' does not exist")
  expect_error(git_remote_set_pushurl(
    'https://github.com/foo/bar',
    repo = repo
  ))
  expect_error(git_remote_info(repo = repo), "remote 'NA' does not exist")
  expect_equal(
    git_remote_add(
      'https://github.com/jeroen/webp',
      name = 'jeroen',
      repo = repo
    ),
    'jeroen'
  )
  # info should work even when no upstream:
  info <- git_remote_info(repo = repo)
  expect_equal(info$url, "https://github.com/jeroen/webp")
  git_fetch('jeroen', 'master', repo = repo)
  git_branch_create('master', 'jeroen/master', repo = repo)
  git_branch_create(
    'testje',
    git_commit_id('jeroen/master', repo = repo),
    checkout = FALSE,
    repo = repo
  )
  git_remote_set_pushurl('https://github.com/foo/baz', repo = repo)
  info <- git_remote_info(repo = repo)
  expect_equal(info$name, 'jeroen')
  expect_equal(info$url, "https://github.com/jeroen/webp")
  expect_equal(info$push_url, "https://github.com/foo/baz")
  expect_equal(info$fetch, "+refs/heads/*:refs/remotes/jeroen/*")
  branches <- git_branch_list(repo = repo)
  expect_equal(branches$name, c("master", "testje", "jeroen/master"))
  expect_equal(branches$upstream, c("refs/remotes/jeroen/master", NA, NA))
})

test_that("remotes after clone", {
  skip_if_offline('github.com')
  repo <- file.path(tempdir(), 'gert')
  if (!file.exists(repo)) {
    git_clone('https://github.com/r-lib/gert', path = repo)
  }
  info <- git_remote_info(repo = repo)
  expect_equal(info$name, 'origin')
  expect_equal(info$url, "https://github.com/r-lib/gert")
  expect_equal(info$fetch, "+refs/heads/*:refs/remotes/origin/*")
  refspecs <- git_remote_refspecs(repo = repo)
  expect_equal(refspecs$direction, 'fetch')
  remotelist <- git_remote_list(repo = repo)
  expect_equal(remotelist$name, 'origin')
  expect_equal(remotelist$url, 'https://github.com/r-lib/gert')
  expect_error(git_remote_add('https://github.com/jeroen/gert', repo = repo))
  expect_equal(
    git_remote_add(
      'https://github.com/jeroen/gert',
      name = 'myfork',
      repo = repo
    ),
    'myfork'
  )
  remotelist <- git_remote_list(repo = repo)
  expect_equal(sort(remotelist$name), c("myfork", 'origin'))
  expect_equal(
    sort(remotelist$url),
    c("https://github.com/jeroen/gert", 'https://github.com/r-lib/gert')
  )
  expect_equal(
    git_remote_refspecs('myfork', repo = repo)$refspec,
    '+refs/heads/*:refs/remotes/myfork/*'
  )
})

#' Get or set Git configuration
#'
#' @description
#' Get, set, or unset Git options, as `git config` does on the command line.
#' **Global** settings affect all of a user's Git operations
#' (`git config --global`), whereas **local** settings are scoped to a specific
#' repository (`git config --local`). When both exist, local options always win.
#'
#' ```{r echo = FALSE, results = "asis"}
#' dat <- data.frame(
#'   local = c("`git_config()`", "`git_config_get()`", "`git_config_local_get()`", "`git_config_set()`", "`git_config_unset()`"),
#'   global = c("`git_config_global()`", "`git_config_get()`", "`git_config_global_get()`", "`git_config_global_set()`", "`git_config_global_unset()`"),
#'   row.names = c("get all", "get one (local+global)", "get one (local or global only)", "set", "unset")
#' )
#' knitr::kable(dat, col.names = paste0("**", colnames(dat), "**"))
#' ```
#' @return
#' * `git_config()`: a `data.frame` of the Git options "in force" in the context
#'   of `repo`, one row per option. The `level` column reveals whether the
#'   option is determined from global or local config.
#' * `git_config_global()`: a `data.frame`, as for `git_config()`, except only
#'   for global Git options.
#' * `git_config_get()`: the value of the named option considering both local and
#'   global config (local wins), or `NULL` if unset.
#' * `git_config_local_get()`: as for `git_config_get()`, but restricted to
#'   local (repository-level) config only.
#' * `git_config_global_get()`: as for `git_config_get()`, but for global config only.
#' * `git_config_set()`, `git_config_global_set()`: The previous value(s) of
#'   `name` in local or global config, respectively. If this option was
#'   previously unset, returns `NULL`. Returns invisibly.
#' * `git_config_unset()`, `git_config_global_unset()`: The previous value(s) of
#'   `name` that were unset. Returns invisibly.
#'
#' @note All entries in the `name` column are automatically normalised to
#'   lowercase (see
#'   <https://libgit2.org/libgit2/#HEAD/type/git_config_entry> for details).
#'
#' @examples
#' # Set and inspect a local, custom Git option
#' r <- file.path(tempdir(), "gert-demo")
#' git_init(r)
#'
#' previous <- git_config_set("aaa.bbb", "ccc", repo = r)
#' previous
#' git_config_local_get("aaa.bbb", repo = r)
#'
#' previous <- git_config_set("aaa.bbb", NULL, repo = r)
#' previous
#' git_config_local_get("aaa.bbb", repo = r)
#'
#' # Get a single named option (returns NULL if unset)
#' git_config_get("aaa.bbb", repo = r)
#' git_config_set("aaa.bbb", "ccc", repo = r)
#' git_config_get("aaa.bbb", repo = r)
#'
#' unlink(r, recursive = TRUE)
#'
#' \dontrun{
#' # Set global Git options
#' git_config_global_set("user.name", "Your Name")
#' git_config_global_set("user.email", "your@email.com")
#' git_config_global()
#'
#' # Get a single global option (returns NULL if unset)
#' git_config_global_get("user.name")
#' git_config_global_get("gert.nonexistent")
#' }
#' @export
#' @family git
#' @inheritParams git_open
#' @useDynLib gert R_git_config_list
#' @git config
git_config <- function(repo = '.') {
  repo <- git_open(repo)
  .Call(R_git_config_list, repo)
}

#' @export
#' @rdname git_config
git_config_global <- function() {
  .Call(R_git_config_list, NULL)
}

#' @export
#' @rdname git_config
#' @param name Name of the option to get or set
git_config_get <- function(name, repo = '.') {
  cfg <- git_config(repo = repo)
  if (!name %in% cfg$name) {
    return(NULL)
  }
  cfg$value[cfg$name == name]
}

#' @export
#' @rdname git_config
git_config_local_get <- function(name, repo = '.') {
  cfg <- git_config(repo = repo)
  cfg <- cfg[cfg$level == "local", ]
  if (!name %in% cfg$name) {
    return(NULL)
  }
  cfg$value[cfg$name == name]
}

#' @export
#' @rdname git_config
git_config_global_get <- function(name) {
  cfg <- git_config_global()
  if (!name %in% cfg$name) {
    return(NULL)
  }
  cfg$value[cfg$name == name]
}

#' @export
#' @rdname git_config
#' @useDynLib gert R_git_config_set
#' @param value Value to set. Must be a string, logical, number or `NULL` (to
#'   unset).
#' @param add if `TRUE`, append a new entry for `name` instead of replacing
#'   existing one(s). Equivalent to `git config --add`. Only supported for
#'   string values.
git_config_set <- function(name, value, add = FALSE, repo = '.') {
  if (!is.logical(add) || length(add) != 1) {
    stop("Argument add must be a logical of length 1.", call. = FALSE)
  }
  repo <- git_open(repo)
  name <- as.character(name)
  out <- git_config_local_get(name, repo = repo)
  .Call(R_git_config_set, repo, name, value, add)
  if (length(out) > 0) {
    invisible(out)
  } else {
    invisible(NULL)
  }
}

#' @export
#' @rdname git_config
git_config_global_set <- function(name, value, add = FALSE) {
  out <- git_config_global_get(name)
  .Call(R_git_config_set, NULL, name, value, add)
  if (length(out) > 0) {
    invisible(out)
  } else {
    invisible(NULL)
  }
}

#' @export
#' @rdname git_config
#' @useDynLib gert R_git_config_unset
#' @param pattern optional regular expression, for matching values to unset in
#' case of multiple values
git_config_unset <- function(name, pattern = NULL, repo = '.'){
  .Call(R_git_config_unset, git_open(repo), name, pattern)
}

#' @export
#' @rdname git_config
git_config_global_unset <- function(name, pattern = NULL){
  .Call(R_git_config_unset, NULL, name, pattern)
}

#' Show libgit2 version and capabilities
#'
#' `libgit2_config()` reveals which version of libgit2 gert is using and which
#' features are supported, such whether you are able to use ssh remotes.
#'
#' @export
#' @useDynLib gert R_libgit2_config
#' @examples
#' libgit2_config()
libgit2_config <- function() {
  res <- .Call(R_libgit2_config)
  res$version <- as.numeric_version(res$version)
  res
}

# helpers used in tests
configure_local_user <- function(repo = ".") {
  git_config_set('user.name', "Jerry Johnson", repo = repo)
  git_config_set('user.email', "jerry@gmail.com", repo = repo)
}

local_author <- function(repo = ".") {
  user <- git_signature_parse(git_signature_default(repo))
  sprintf("%s <%s>", user$name, user$email)
}

# helpers used in pkgdown setup
configure_global_user <- function() {
  git_config_global_set('user.name', "Jerry Johnson")
  git_config_global_set('user.email', "jerry@gmail.com")
}

global_user_is_configured <- function() {
  user_name_exists <- !is.null(git_config_global_get("user.name"))
  user_email_exists <- !is.null(git_config_global_get("user.email"))
  user_name_exists && user_email_exists
}

# helpers used in examples

#' Test if a Git user is configured
#'
#' This function exists mostly to guard examples that rely on having a user
#' configured, in order to make commits. `user_is_configured()` makes no
#' distinction between local or global user config.
#'
#' @param repo An optional `repo`, in the sense of [git_open()].
#'
#' @return `TRUE` if `user.name` and `user.email` are set locally or globally,
#'   `FALSE` otherwise.
#'
#' @export
#' @examplesIf interactive()
#' user_is_configured()
user_is_configured <- function(repo = ".") {
  user_name_exists <- !is.null(tryCatch(
    git_config_get("user.name", repo = repo),
    error = function(e) git_config_global_get("user.name")
  ))
  user_email_exists <- !is.null(tryCatch(
    git_config_get("user.email", repo = repo),
    error = function(e) git_config_global_get("user.email")
  ))
  user_name_exists && user_email_exists
}

test_that("git_config_get returns value or NULL", {
  repo <- git_init(tempfile("gert-tests-config"))
  on.exit(unlink(repo, recursive = TRUE))

  expect_null(git_config_get("aaa.bbb", repo = repo))
  git_config_set("aaa.bbb", "ccc", repo = repo)
  expect_equal(git_config_get("aaa.bbb", repo = repo), "ccc")
})

test_that("git_config_global_get returns value or NULL", {
  expect_null(git_config_global_get("gert.nonexistent.option.xyzzy"))
  # Note: avoid setting/unsetting real global config in tests
})

test_that("git_config_local_get returns local value or NULL", {
  repo <- git_init(tempfile("gert-tests-config"))
  on.exit(unlink(repo, recursive = TRUE))

  expect_null(git_config_local_get("aaa.bbb", repo = repo))
  git_config_set("aaa.bbb", "ccc", repo = repo)
  expect_equal(git_config_local_get("aaa.bbb", repo = repo), "ccc")
  # global-only option should not be visible at local level
  expect_null(git_config_local_get(
    "gert.nonexistent.option.xyzzy",
    repo = repo
  ))
})

test_that("local, custom config roundtrip", {
  repo <- git_init(tempfile("gert-tests-config"))
  on.exit(unlink(repo, recursive = TRUE))

  orig <- git_config_set("aaa.bbb", "ccc", repo = repo)
  expect_null(orig)
  expect_equal(git_config_get("aaa.bbb", repo = repo), "ccc")

  orig <- git_config_set("aaa.bbb", NULL, repo = repo)
  expect_equal(orig, "ccc")
  expect_null(git_config_get("aaa.bbb", repo = repo))
})

test_that("git_config_unset() works for multivar options", {
  repo <- git_init(tempfile("gert-tests-config"))
  on.exit(unlink(repo, recursive = TRUE))

  git_config_set("aaa.bbb", "ccc", add = TRUE, repo = repo)
  git_config_set("aaa.bbb", "ccccc", add = TRUE, repo = repo)
  git_config_set("aaa.bbb", "ddd", add = TRUE, repo = repo)

  git_config_unset("aaa.bbb", "ccc", repo = repo)
  expect_equal(git_config_get("aaa.bbb", repo = repo), "ddd")

  git_config_unset("aaa.bbb", repo = repo)
  expect_null(git_config_get("aaa.bbb", repo = repo))
})

#' Submodules
#'
#' Interact with submodules in the repository.
#'
#' @export
#' @rdname git_submodule
#' @inheritParams git_open
#' @useDynLib gert R_git_submodule_list
git_submodule_list <- function(repo = '.'){
  repo <- git_open(repo)
  .Call(R_git_submodule_list, repo)
}

#' @export
#' @rdname git_submodule
#' @useDynLib gert R_git_submodule_info
git_submodule_info <- function(submodule, repo = '.'){
  repo <- git_open(repo)
  submodule <- as.character(submodule)
  .Call(R_git_submodule_info, repo, submodule)
}

#' @export
#' @rdname git_submodule
#' @useDynLib gert R_git_submodule_init
#' @param submodule name of the submodule
#' @param overwrite overwrite existing entries
git_submodule_init <- function(submodule, overwrite = FALSE, repo = '.'){
  repo <- git_open(repo)
  submodule <- as.character(submodule)
  overwrite <- as.logical(overwrite)
  .Call(R_git_submodule_init, repo, submodule, overwrite)
}

#' @export
#' @rdname git_submodule
#' @useDynLib gert R_git_submodule_set_to
#' @param ref branch or tag to point the submodule at. If checkout = FALSE, you
#' can pass a commit hash before downloading the submodule.
#' @param checkout actually switch the contents of the directory to this commit
git_submodule_set_to <- function(submodule, ref, checkout = TRUE, repo = '.'){
  repo <- git_open(repo)
  submodule <- as.character(submodule)
  info <- git_submodule_info(submodule, repo = repo)
  if(isTRUE(checkout)){
    git_reset_hard(ref = ref, repo = I(info$path))
    ref <- git_info(repo = I(info$path))$commit
  } else if(!is_full_hash(ref)) {
    ref <- git_commit_info(ref, repo = I(info$path))$id
  }
  if(!is_full_hash(ref))
    stop("When checkout = FALSE, parameter ref must be a full hash")
 .Call(R_git_submodule_set_to, repo, submodule, ref)
}

#' @export
#' @rdname git_submodule
#' @param url full git url of the submodule
#' @param path relative of the submodule
#' @param ref a branch or tag or hash with
#' @param ... extra arguments for [git_fetch] for authentication things
git_submodule_add <- function(url, path = basename(url), ref = 'HEAD', ..., repo = '.'){
  if(!is_a_hash(ref)){
    upstream_refs <- git_remote_ls(url, ..., repo = repo)
    ref_match <- sub("refs/(heads|tags)/","", upstream_refs$ref) == ref
    if(!any(ref_match)){
      stop(sprintf("Upstream repo %s does not have a branch or tag named '%s'",
                   basename(url), ref))
    }
    ref <- upstream_refs$oid[ref_match]
  }
  submodule <- git_submodule_setup(url = url, path = path, repo = repo)
  git_fetch('origin', ..., repo = submodule)
  git_reset_hard(ref, repo = submodule)
  git_submodule_save(path, repo = repo)
  git_submodule_info(path, repo = repo)
}

#' @export
#' @rdname git_submodule
git_submodule_fetch <- function(submodule, ..., repo = '.'){
  sm <- git_submodule_info(submodule = submodule, repo = repo)
  subrepo = I(sm$path)
  tryCatch({
    git_fetch('origin', ..., repo = subrepo)
  }, GIT_ENOTFOUND = function(e){
    inform("Initial clone for submodule '%s'", submodule)
    git_clone(sm$url, ..., path = subrepo)
  })
  if(length(sm$branch) && !is.na(sm$branch)){
    git_reset_hard(sm$branch, repo = subrepo)
  } else {
    remote_head <- git_remote_info('origin', repo = subrepo)$head
    if(!length(remote_head)){
      git_remote_ls('origin', ..., repo = subrepo)
      remote_head <- git_remote_info('origin', repo = subrepo)$head
    }
    git_reset_hard(remote_head, repo = I(sm$path))
  }
  git_commit_id(repo = subrepo)
}

#' @useDynLib gert R_git_submodule_setup
git_submodule_setup <- function(url, path, repo){
  repo <- git_open(repo)
  path <- as.character(path)
  url <- as.character(url)
  .Call(R_git_submodule_setup, repo, url, path)
}

#' @useDynLib gert R_git_submodule_save
git_submodule_save <- function(submodule, repo){
  repo <- git_open(repo)
  submodule <- as.character(submodule)
  .Call(R_git_submodule_save, repo, submodule)
}

# I find this confusing, also doesn't support auth.
# Better use git_submodule_fetch()
#' @useDynLib gert R_git_submodule_update
git_submodule_update <- function(submodule, init = TRUE, repo = '.'){
  repo <- git_open(repo)
  submodule <- as.character(submodule)
  init <- as.logical(init)
  .Call(R_git_submodule_update, repo, submodule, init)
}

is_a_hash <- function(x){
  grepl('^[a-f0-9]{7,}$', tolower(x))
}

is_full_hash <- function(x){
  grepl('^[a-f0-9]{40}$', tolower(x))
}

#' Restore working tree files
#'
#' Restores specified paths in the working tree from a given ref, equivalent
#' to `git restore --source=<ref> <path>` (or the older
#' `git checkout <ref> -- <path>`). The ref must be reachable from the
#' current HEAD. By default restores from HEAD, discarding any local
#' modifications.
#'
#' @export
#' @name git_restore
#' @rdname git_restore
#' @family git
#' @inheritParams git_open
#' @param path character vector with file paths to restore, relative to the
#'   repository root. Use `"."` to restore all tracked files.
#' @param ref revision string with a branch/tag/commit to restore from.
#'   Defaults to `"HEAD"`.
#' @return Invisibly, the [git_status()] after restoring.
#' @examplesIf interactive()
#' repo <- file.path(tempdir(), "myrepo")
#' git_init(repo)
#'
#' # Set a user if no default
#' if (!user_is_configured()) {
#'   git_config_set("user.name", "Jerry")
#'   git_config_set("user.email", "jerry@gmail.com")
#' }
#'
#' writeLines("hello", file.path(repo, "hello.txt"))
#' git_add("hello.txt", repo = repo)
#' git_commit("First commit", repo = repo)
#'
#' # Modify the file, then restore it from HEAD
#' writeLines("oops", file.path(repo, "hello.txt"))
#' git_restore("hello.txt", repo = repo)
#' readLines(file.path(repo, "hello.txt"))  # "hello"
#'
#' unlink(repo, recursive = TRUE)
#' @useDynLib gert R_git_restore
git_restore <- function(path, ref = "HEAD", repo = ".") {
  repo <- git_open(repo)
  path <- check_path_tracked(as.character(path), repo)
  ref <- check_ref_in_history(ref, repo)
  .Call(R_git_restore, repo, path, ref)
  invisible(git_status(repo = repo))
}

check_path_tracked <- function(path, repo) {
  if (identical(path, ".")) {
    return(character(0))
  }
  tracked <- git_ls(repo = repo)$path
  untracked <- setdiff(path, tracked)
  if (length(untracked) > 0) {
    stop("Path(s) not tracked by git: ", toString(untracked))
  }
  path
}

#' Git Branch
#'
#' Create, list, and checkout branches.
#'
#' @export
#' @rdname git_branch
#' @name git_branch
#' @family git
#' @inheritParams git_open
#' @useDynLib gert R_git_branch_current
git_branch <- function(repo = '.'){
  repo <- git_open(repo)
  .Call(R_git_branch_current, repo)
}

#' @export
#' @rdname git_branch
#' @useDynLib gert R_git_branch_list
git_branch_list <- function(local = NULL, repo = '.'){
  repo <- git_open(repo)
  local <- as.logical(local)
  .Call(R_git_branch_list, repo, local)
}

#' @export
#' @rdname git_branch
#' @param branch name of branch to check out
#' @param force ignore conflicts and overwrite modified files
#' @param orphan if branch does not exist, checkout unborn branch
#' @useDynLib gert R_git_checkout_branch R_git_checkout_unborn
git_branch_checkout <- function(branch, force = FALSE, orphan = FALSE, repo = '.'){
  repo <- git_open(repo)
  branch <- as.character(branch)
  force <- as.logical(force)
  if(!git_branch_exists(branch, repo = repo)){
    if(isTRUE(orphan)){
      ref <- paste0('refs/heads/', branch)
      .Call(R_git_checkout_unborn, repo, ref)
      return(ref)
    }
    all_branches <- subset(git_branch_list(repo = repo), local == FALSE)$name
    candidate <- sub("^[^/]+/", "", all_branches) == branch
    if(sum(candidate) > 1){
      stop(sprintf("Local branch '%s' does not exist and multiple remote candidates found.", branch))
    } else if(sum(candidate) == 0){
      stop(sprintf("No local or remote branch '%s' found.", branch))
    } else {
      remote_branch <- unname(all_branches[candidate])
      inform("Creating local branch %s from %s", branch, remote_branch)
      git_branch_create(branch, remote_branch, checkout = FALSE, repo = repo)
    }
  }
  .Call(R_git_checkout_branch, repo, branch, force)
}

#' @export
#' @rdname git_branch
#' @useDynLib gert R_git_create_branch
#' @param ref string with a branch/tag/commit
#' @param checkout move HEAD to the newly created branch
git_branch_create <- function(branch, ref = "HEAD", checkout = TRUE, repo = '.'){
  repo <- git_open(repo)
  branch <- as.character(branch)
  ref <- as.character(ref)
  checkout <- as.logical(checkout)
  invisible(.Call(R_git_create_branch, repo, branch, ref, checkout))
}

#' @export
#' @rdname git_branch
#' @useDynLib gert R_git_delete_branch
git_branch_delete <- function(branch, repo = '.'){
  repo <- git_open(repo)
  branch <- as.character(branch)
  .Call(R_git_delete_branch, repo, branch)
  git_repo_path(repo)
}

#' @export
#' @rdname git_branch
#' @useDynLib gert R_git_branch_move
#' @param new_branch target name of the branch once the move is performed; this name is validated for consistency.
git_branch_move <- function(branch, new_branch, force = FALSE, repo = '.'){
  repo <- git_open(repo)
  branch <- as.character(branch)
  new_branch <- as.character(new_branch)
  .Call(R_git_branch_move, repo, branch, new_branch, as.logical(force))
  git_repo_path(repo)
}

#' @export
#' @rdname git_branch
git_branch_fast_forward <- function(ref, repo = '.'){
  analysis <- git_merge_analysis(ref = ref, repo = repo)
  if(analysis != "fastforward")
    stop("Branch cannot be fast-forwarded. Use git_merge() instead")
  git_branch_set_target(ref = ref, repo = repo)
}

#' @export
#' @rdname git_branch
#' @param upstream remote branch from [git_branch_list], for example `"origin/master"`
#' @useDynLib gert R_git_branch_set_upstream
git_branch_set_upstream <- function(upstream, branch = git_branch(repo), repo = '.'){
  repo <- git_open(repo)
  stopifnot(is.character(upstream))
  if(!git_branch_exists(upstream, local = FALSE, repo = repo))
    stop(sprintf("No remote branch found: %s, maybe fetch first?", upstream))
  .Call(R_git_branch_set_upstream, repo, upstream, branch)
  git_repo_path(repo)
}

#' @export
#' @rdname git_branch
#' @param local set TRUE to only check for local branches, FALSE to check for remote
#' branches. Use NULL to return all branches.
#' @useDynLib gert R_git_branch_exists
git_branch_exists <- function(branch, local = TRUE, repo = '.'){
  repo <- git_open(repo)
  branch <- as.character(branch)
  local <- as.logical(local)
  .Call(R_git_branch_exists, repo, branch, local)
}

#' @useDynLib gert R_git_branch_set_target
git_branch_set_target <- function(ref, branch, repo){
  repo <- git_open(repo)
  ref <- as.character(ref)
  .Call(R_git_branch_set_target, repo, ref)
  git_repo_path(repo)
}

test_that("rebasing things", {
  skip_if_offline('github.com')
  repo <- file.path(tempdir(), 'gert')
  if (!file.exists(repo)) {
    git_clone('https://github.com/r-lib/gert', path = repo)
  }
  git_branch_create('backup', checkout = FALSE, repo = repo)

  # Original log
  orig <- git_log(max = 10, repo = repo)

  # Drop some commits, and fast-forward them back
  git_reset_hard('HEAD~5', repo = repo)
  expect_equal(
    git_log(max = 5, repo = repo)$commit,
    utils::tail(orig$commit, 5)
  )
  expect_equal(
    git_ahead_behind(repo = repo),
    list(
      ahead = 0,
      behind = 5,
      local = git_commit_id(repo = repo),
      upstream = git_commit_id('origin/HEAD', repo = repo)
    )
  )
  git_pull(repo = repo)
  expect_equal(orig, git_log(max = 10, repo = repo))

  # Same with rebase
  git_reset_hard('HEAD~5', repo = repo)
  expect_equal(
    git_log(max = 5, repo = repo)$commit,
    utils::tail(orig$commit, 5)
  )
  expect_equal(
    git_ahead_behind(repo = repo),
    list(
      ahead = 0,
      behind = 5,
      local = git_commit_id(repo = repo),
      upstream = git_commit_id('origin/HEAD', repo = repo)
    )
  )
  git_pull(rebase = TRUE, repo = repo)
  expect_equal(orig, git_log(max = 10, repo = repo))

  # Now rebase a commit
  git_reset_hard('HEAD~5', repo = repo)
  expect_equal(
    git_log(max = 5, repo = repo)$commit,
    utils::tail(orig$commit, 5)
  )
  writeLines("some random change", file.path(repo, 'randomfile.txt'))
  git_add(".", repo = repo)
  commit_id <- git_commit(
    "Added a local change",
    repo = repo,
    author = "Jerry <test@jerry.nl>"
  )
  commit_msg <- git_commit_info(commit_id, repo = repo)$message

  # We are 1 ahead and 5 behind
  expect_equal(
    git_ahead_behind(repo = repo),
    list(
      ahead = 1,
      behind = 5,
      local = git_commit_id(repo = repo),
      upstream = git_commit_id('origin/HEAD', repo = repo)
    )
  )

  # Confirm that we cannot fast forward
  upstream <- git_info(repo = repo)$upstream
  expect_error(git_branch_fast_forward(upstream, repo = repo))

  # Pull with rebase
  rebase_info <- git_rebase_list(repo = repo)
  expect_equal(rebase_info$commit, commit_id)
  git_pull(rebase = TRUE, repo = repo)
  newlog <- git_log(max = 11, repo = repo)
  expect_equal(c(commit_msg, orig$message), newlog$message)
  expect_equal(orig$commit, newlog$commit[-1])
  expect_equal(
    git_ahead_behind(repo = repo),
    list(
      ahead = 1,
      behind = 0,
      local = git_commit_id(repo = repo),
      upstream = git_commit_id('origin/HEAD', repo = repo)
    )
  )

  # Check that patch matches
  expect_equal(
    git_diff(commit_id, repo = repo),
    git_diff(newlog$commit[1], repo = repo)
  )

  # Merge changes into another branch
  main <- git_info(repo = repo)$shorthand
  git_branch_checkout('backup', repo = repo)
  git_merge(main, repo = repo)
  expect_equal(newlog, git_log(max = 11, repo = repo))

  # Cleanup from master
  git_branch_checkout(main, repo = repo)
  git_reset_hard("HEAD^", repo = repo)
  git_branch_delete('backup', repo = repo)
})

test_that("cherry-picking things", {
  repo <- tempfile()
  gert::git_init(path = repo)
  writeLines("hello", file.path(repo, 'hello.txt'))
  git_add('hello.txt', repo = repo)
  first_commit <- git_commit(
    "First commit",
    author = "jeroen <jeroen@blabla.nl>",
    repo = repo
  )

  # Create a feature branch with a new commit
  mainbranch <- gert::git_branch(repo = repo)
  git_branch_create('feature', repo = repo)
  write.csv(iris, file.path(repo, 'iris.csv'))
  git_add('iris.csv', repo = repo)
  commit <- git_commit(
    "Added iris.csv file",
    author = "maelle <maelle@salmon.fish>",
    repo = repo
  )

  # Cherry pick the commit onto main
  git_branch_checkout(mainbranch, repo = repo)
  expect_equal(git_log(repo = repo)$commit, first_commit)
  short_commit <- substr(commit, 1, 7)
  expect_equal(git_cherry_pick(short_commit, repo = repo), commit)
  expect_length(git_log(repo = repo)$commit, 2)
})

test_that("git_branch_create force applies to checkout", {
  repo <- git_init(tempfile("gert-tests-branch-force"))
  on.exit(unlink(repo, recursive = TRUE))
  configure_local_user(repo)

  # First commit: hello.txt = "v1"
  writeLines("v1", file.path(repo, "hello.txt"))
  git_add("hello.txt", repo = repo)
  first <- git_commit("First commit", repo = repo)

  # Second commit: hello.txt = "v2"
  writeLines("v2", file.path(repo, "hello.txt"))
  git_add("hello.txt", repo = repo)
  git_commit("Second commit", repo = repo)

  # Unstaged local change: hello.txt = "v3"
  writeLines("v3", file.path(repo, "hello.txt"))

  # force = FALSE: checkout to first commit conflicts with local change
  expect_error(
    git_branch_create("oldbranch", ref = first, checkout = TRUE, force = FALSE, repo = repo)
  )

  # force = TRUE: checkout succeeds, overwriting local change
  git_branch_create("oldbranch2", ref = first, checkout = TRUE, force = TRUE, repo = repo)
  expect_equal(git_branch(repo = repo), "oldbranch2")
  expect_equal(readLines(file.path(repo, "hello.txt")), "v1")
})

#' Push and pull
#'
#' Functions to connect with a git server (remote) to fetch or push changes.
#' The 'credentials' package is used to handle authentication, the
#' [credentials vignette](https://docs.ropensci.org/credentials/articles/intro.html)
#' explains the various authentication methods for SSH and HTTPS remotes.
#'
#' Use [git_fetch()] and [git_push()] to sync a local branch with a remote
#' branch. Here [git_pull()] is a wrapper for [git_fetch()] which then tries to
#' [fast-forward][git_branch_fast_forward()] the local branch after fetching.
#'
#' @export
#' @family git
#' @name git_fetch
#' @rdname git_fetch
#' @inheritParams git_open
#' @useDynLib gert R_git_remote_fetch
#' @param remote Optional. Name of a remote listed in [git_remote_list()]. If
#'   unspecified and the current branch is already tracking branch a remote
#'   branch, that remote is honored. Otherwise, defaults to `origin`.
#' @param refspec string with mapping between remote and local refs. Default
#' uses the default refspec from the remote, which usually fetches all branches.
#' @param mirror use the `--mirror` flag
#' @param bare use the `--bare` flag
#' @param force use the `--force` flag
#' @param prune delete tracking branches that no longer exist on the remote, or
#' are not in the refspec (such as pull requests).
git_fetch <- function(remote = NULL, refspec = NULL, password = askpass, ssh_key = NULL,
                      prune = FALSE, verbose = interactive(), repo = '.'){
  repo <- git_open(repo)
  info <- git_info(repo)
  if(!length(remote))
    remote <- info$remote
  remote <- as.character(remote)
  if(!length(remote) || is.na(remote)){
    if(is.na(match("origin", git_remote_list(repo = repo)$name))){
      stop("No remote is set for this branch")
    } else {
      inform("No remote set for this branch, using default remote 'origin'")
      remote <- "origin"
    }
  }
  refspec <- as.character(refspec)
  prune <- as.logical(prune)
  verbose <- as.logical(verbose)
  host <- remote_to_host(repo, remote)
  key_cb <- make_key_cb(ssh_key, host = host, password = password)
  cred_cb <- make_cred_cb(password = password, verbose = verbose)
  .Call(R_git_remote_fetch, repo, remote, refspec, key_cb, cred_cb, prune, verbose)
  git_repo_path(repo)
}

#' @export
#' @rdname git_fetch
#' @useDynLib gert R_git_remote_ls
git_remote_ls <- function(remote = NULL, password = askpass, ssh_key = NULL,
                                   verbose = interactive(), repo = '.'){
  repo <- git_open(repo)
  info <- git_info(repo)
  if(!length(remote))
    remote <- info$remote
  remote <- as.character(remote)
  if(!length(remote) || is.na(remote)){
    if(is.na(match("origin", git_remote_list(repo = repo)$name))){
      stop("No remote is set for this branch")
    } else {
      inform("No remote set for this branch, using default remote 'origin'")
      remote <- "origin"
    }
  }
  verbose <- as.logical(verbose)
  host <- remote_to_host(repo, remote)
  key_cb <- make_key_cb(ssh_key, host = host, password = password)
  cred_cb <- make_cred_cb(password = password, verbose = verbose)
  .Call(R_git_remote_ls, repo, remote, key_cb, cred_cb, verbose)
}

#' @export
#' @rdname git_fetch
#' @param set_upstream change the branch default upstream to `remote`.
#' If `NULL`, this will set the branch upstream only if the push was
#' successful and if the branch does not have an upstream set yet.
#' @useDynLib gert R_git_remote_push
git_push <- function(remote = NULL, refspec = NULL, set_upstream = NULL,
                     password = askpass, ssh_key = NULL, mirror = FALSE,
                     force = FALSE, verbose = interactive(), repo = '.'){
  repo <- git_open(repo)
  info <- git_info(repo)
  verbose <- as.logical(verbose)

  if(!length(remote))
    remote <- info$remote

  remote <- as.character(remote)

  if(!length(remote) || is.na(remote)){
    if(is.na(match("origin", git_remote_list(repo = repo)$name))){
      stop("No remote is set for this branch")
    } else {
      if (verbose) {
        inform("No remote set for this branch, using default remote 'origin'")
      }
      remote <- "origin"
    }
  }

  if(isTRUE(mirror)) {
    refs <- info$reflist
    # Ignore github's special refs
    refs <- refs[!grepl("^refs/pull", refs)]
    refspec <- paste0(refs, ":", refs)
  }
  if(!length(refspec))
    refspec <- info$head
  refspec <- as.character(refspec)
  if(isTRUE(force))
    refspec <- sub("^\\+?","+", refspec)

  host <- remote_to_host(repo, remote)
  key_cb <- make_key_cb(ssh_key, host = host, password = password)
  cred_cb <- make_cred_cb(password = password, verbose = verbose)

  .Call(R_git_remote_push, repo, remote, refspec, key_cb, cred_cb, verbose)
  if(is.null(set_upstream))
    set_upstream <- isTRUE(is.na(info$upstream)) && !isTRUE(info$bare)

  if(isTRUE(set_upstream)){
    git_branch_set_upstream(paste0(remote, "/", info$shorthand), repo = repo)
  }
  git_repo_path(repo)
}

#' @export
#' @rdname git_fetch
#' @useDynLib gert R_git_repository_clone
#' @param url remote url. Typically starts with `https://github.com/` for public
#' repositories, and `https://yourname@github.com/` or `git@github.com/` for
#' private repos. You will be prompted for a password or pat when needed.
#' @param path Directory of the Git repository to create.
#' @param ssh_key path or object containing your ssh private key. By default we
#' look for keys in `ssh-agent` and [credentials::ssh_key_info].
#' @param branch name of branch to check out locally
#' @param password a string or a callback function to get passwords for authentication
#' or password protected ssh keys. Defaults to [askpass][askpass::askpass] which
#' checks `getOption('askpass')`.
#' @param verbose display some progress info while downloading
#' @examples {# Clone a small repository
#' git_dir <- file.path(tempdir(), 'antiword')
#' git_clone('https://github.com/ropensci/antiword', git_dir)
#'
#' # Change into the repo directory
#' olddir <- getwd()
#' setwd(git_dir)
#'
#' # Show some stuff
#' git_log()
#' git_branch_list()
#' git_remote_list()
#'
#' # Add a file
#' write.csv(iris, 'iris.csv')
#' git_add('iris.csv')
#'
#' # Commit the change
#' jerry <- git_signature("Jerry", "jerry@hotmail.com")
#' git_commit('added the iris file', author = jerry)
#'
#' # Now in the log:
#' git_log()
#'
#' # Cleanup
#' setwd(olddir)
#' unlink(git_dir, recursive = TRUE)
#' }
git_clone <- function(url, path = NULL, branch = NULL, password = askpass, ssh_key = NULL,
                      bare = FALSE, mirror = FALSE, verbose = interactive()){
  stopifnot(is.character(url))
  if(!length(path))
    path <- file.path(getwd(), basename(url))
  stopifnot(is.character(path))
  stopifnot(is.null(branch) || is.character(branch))
  verbose <- as.logical(verbose)
  path <- normalizePath(path.expand(path), mustWork = FALSE)
  host <- url_to_host(url)
  key_cb <- make_key_cb(ssh_key, host = host, password = password)
  cred_cb <- make_cred_cb(password = password, verbose = verbose)
  repo <- .Call(R_git_repository_clone, url, path, branch, key_cb, cred_cb, bare, mirror, verbose)
  git_repo_path(repo)
}

#' @export
#' @rdname git_fetch
#' @param rebase if TRUE we try to rebase instead of merge local changes. This
#' is not possible in case of conflicts (you will get an error).
#' @param ... arguments passed to [git_fetch]
git_pull <- function(remote = NULL, rebase = FALSE, ..., repo = '.'){
  repo <- git_open(repo)
  info <- git_info(repo)
  branch <- info$shorthand
  if (branch == "HEAD")
    stop("Repository is currently in a detached head state")

  upstream <- if(length(remote) && nchar(remote)){
    paste0(remote, '/', branch)
  } else {
    info$upstream
  }

  if(!length(upstream) || is.na(upstream) || !nchar(upstream))
    stop("No upstream configured for current branch, please specify a remote")

  if(grepl(".*/pr/\\d+$", upstream)){
    pr <- utils::tail(strsplit(upstream, '/pr/', fixed = TRUE)[[1]], 1)
    try(git_fetch_pull_requests(pr = pr, remote = remote, repo = repo))
  }
  if(git_branch_exists(upstream, local = TRUE, repo = repo)){
    inform("Local upstream, skipping fetch")
  } else {
    git_fetch(remote, ..., repo = repo)
    if(!git_branch_exists(upstream, local = FALSE, repo = repo))
      stop("Failed to fetch upstream branch: ", upstream)
  }
  if(isTRUE(rebase)){
    rebase_df <- git_rebase_list(upstream = upstream, repo = repo)
    if(any(rebase_df$conflicts))
      stop("Found conflicts, rebase not possible. Retry with rebase = FALSE")
    git_rebase_commit(upstream = upstream, repo = repo)
  } else {
    git_merge(upstream, repo = repo)
  }
  git_repo_path(repo)
}

# Even for public repos, Github only allows keys that it knows.
test_that("public ssh remotes with random key", {
  skip_if_offline('github.com')
  remote <- 'git@github.com:jeroen/webp.git'
  target <- file.path(tempdir(), basename(remote))
  repo <- git_clone(remote, path = target, ssh_key = 'key.pem', password = 'testingjerry')
  expect_true(file.exists(file.path(target, 'DESCRIPTION')))
})

isOldWindows <- Sys.info()[["sysname"]] == "Windows" && grepl('Windows Server 2008', osVersion)

# Even for public repos, Github only allows keys that it knows.
test_that("private ssh remotes with key", {
  skip_if_offline('github.com')
  remote <- 'git@github.com:ropensci/testprivate.git'
  target <- file.path(tempdir(), basename(remote))

  # Also test password as a callback function
  repo <- git_clone(remote, path = target, ssh_key = 'key.pem', password = function(...){ 'testingjerry'})
  expect_true(file.exists(file.path(target, 'hello')))

  # Test errors
  expect_error(git_clone(remote, path = tempfile(), ssh_key = 'doesnotexist'), 'load key', class = 'GIT_EAUTH')
  expect_error(git_clone(remote, path = tempfile(), ssh_key = 'pat.bin'), 'load key', class = 'GIT_EAUTH')

  # Test ls-remote auth
  git_remote_ls(repo = target, ssh_key = 'key.pem', password = function(...){ 'testingjerry'})
})

# Access token for dummy account with minimal rights
test_that("HTTP user/pass auth", {
  skip_if_offline('github.com')
  # Disable user PAT
  Sys.unsetenv("GITHUB_PAT")

  # Test with password
  enc <- readBin('pat.bin', raw(), 1e3)
  dec <- openssl::rsa_decrypt(enc, 'key.pem', password = 'testingjerry')
  target2 <- file.path(tempdir(), 'testprivate2')
  repo <- git_clone('https://testingjerry@github.com/ropensci/testprivate',
                    path = target2, password = rawToChar(dec))
  expect_true(file.exists(file.path(target2, 'hello')))

  # Test with password in URL
  target3 <- file.path(tempdir(), 'testprivate3')
  repo <- git_clone(sprintf('https://testingjerry:%s@github.com/ropensci/testprivate',
                            rawToChar(dec)), path = target3)
  expect_true(file.exists(file.path(target3, 'hello')))


  # This seems to fail on WinBuilder due to some syscall
  if(!isTRUE(isOldWindows)){
    # Test that repo is private
    expect_error(git_clone('https://github.com/ropensci/testprivate',
                           password = "bla", path = tempfile()), 'Authentication', class = 'GIT_EAUTH')
  }

  # Test with PAT
  Sys.setenv(GITHUB_PAT = rawToChar(dec))
  on.exit(Sys.unsetenv("GITHUB_PAT"))
  target4 <- file.path(tempdir(), 'testprivate4')
  repo <- git_clone('https://github.com/ropensci/testprivate', path = target4)
  expect_true(file.exists(file.path(target4, 'hello')))
  heads <- git_remote_ls(repo = repo)
  expect_is(heads, 'data.frame')
  expect_equal(git_remote_info(repo = repo)$head, "refs/remotes/origin/master")

  # Try with user in URL
  target5 <- file.path(tempdir(), 'testprivate5')
  repo <- git_clone('https://nobody@github.com/ropensci/testprivate', path = target5)
  expect_true(file.exists(file.path(target5, 'hello')))
  heads <- git_remote_ls(repo = repo)
  expect_is(heads, 'data.frame')
  expect_equal(git_remote_info(repo = repo)$head, "refs/remotes/origin/master")
})

# Even for public repos, Github only allows keys that it knows.
test_that("public ssh remotes with random key", {
  skip_if_offline('github.com')  #Also skips on_cran
  remote <- 'git@github.com:jeroen/webp.git'
  target <- file.path(tempdir(), basename(remote))
  repo <- git_clone(remote, path = target, ssh_key = 'ecdsa.key', password = 'testingjerry')
  expect_true(file.exists(file.path(target, 'DESCRIPTION')))
})

isOldWindows <- Sys.info()[["sysname"]] == "Windows" && grepl('Windows Server 2008', osVersion)

# Even for public repos, Github only allows keys that it knows.
test_that("private ssh remotes with key", {
  skip_if_offline('github.com') #Also skips on_cran
  remote <- 'git@github.com:jeroenooms/testprivate.git'

  # Test errors
  expect_error(git_clone(remote, path = tempfile(), ssh_key = 'doesnotexist'), 'load key', class = 'GIT_EAUTH')
  expect_error(git_clone(remote, path = tempfile(), ssh_key = 'pat.bin'), 'load key', class = 'GIT_EAUTH')

  # Also test password as a callback function
  for(keyfile in c("ecdsa.key", "rsa3072.key", "ed25519.key")){
    target <- tempfile()
    repo <- git_clone(remote, path = target, ssh_key = keyfile, password = function(...){ 'testingjerry'})
    expect_true(file.exists(file.path(target, 'hello')))

    # Test ls-remote auth
    git_remote_ls(repo = target, ssh_key = keyfile, password = function(...){ 'testingjerry'})
  }
})

# Access token for dummy account with minimal rights
test_that("HTTP user/pass auth", {
  skip_if_offline('github.com')
  # Disable user PAT
  Sys.unsetenv("GITHUB_PAT")

  # Test with password
  enc <- readBin('pat.bin', raw(), 1e3)
  dec <- openssl::rsa_decrypt(enc, 'key.pem', password = 'testingjerry')
  target2 <- file.path(tempdir(), 'testprivate2')
  repo <- git_clone('https://testingjerry@github.com/jeroenooms/testprivate',
                    path = target2, password = rawToChar(dec))
  expect_true(file.exists(file.path(target2, 'hello')))

  # Test with password in URL
  target3 <- file.path(tempdir(), 'testprivate3')
  repo <- git_clone(sprintf('https://testingjerry:%s@github.com/jeroenooms/testprivate',
                            rawToChar(dec)), path = target3)
  expect_true(file.exists(file.path(target3, 'hello')))


  # This seems to fail on WinBuilder due to some syscall
  if(!isTRUE(isOldWindows)){
    # Test that repo is private
    expect_error(git_clone('https://github.com/jeroenooms/testprivate',
                           password = "bla", path = tempfile()), 'Authentication', class = 'GIT_EAUTH')
  }

  # Test with PAT
  Sys.setenv(GITHUB_PAT = rawToChar(dec))
  on.exit(Sys.unsetenv("GITHUB_PAT"))
  target4 <- file.path(tempdir(), 'testprivate4')
  repo <- git_clone('https://github.com/jeroenooms/testprivate', path = target4)
  expect_true(file.exists(file.path(target4, 'hello')))
  heads <- git_remote_ls(repo = repo)
  expect_is(heads, 'data.frame')
  expect_equal(git_remote_info(repo = repo)$head, "refs/remotes/origin/master")

  # Try with user in URL
  target5 <- file.path(tempdir(), 'testprivate5')
  repo <- git_clone('https://nobody@github.com/jeroenooms/testprivate', path = target5)
  expect_true(file.exists(file.path(target5, 'hello')))
  heads <- git_remote_ls(repo = repo)
  expect_is(heads, 'data.frame')
  expect_equal(git_remote_info(repo = repo)$head, "refs/remotes/origin/master")
})

# Even for public repos, Github only allows keys that it knows.
test_that("public ssh remotes with random key", {
  skip_if_offline('github.com') #Also skips on_cran
  remote <- 'git@github.com:jeroen/webp.git'
  target <- file.path(tempdir(), basename(remote))
  repo <- git_clone(
    remote,
    path = target,
    ssh_key = 'ecdsa.key',
    password = 'testingjerry'
  )
  expect_true(file.exists(file.path(target, 'DESCRIPTION')))
})

# Even for public repos, Github only allows keys that it knows.
test_that("private ssh remotes with key", {
  skip_if_offline('github.com') #Also skips on_cran
  remote <- 'git@github.com:jeroenooms/testprivate.git'

  # Test errors
  expect_error(
    git_clone(remote, path = tempfile(), ssh_key = 'doesnotexist'),
    'load key',
    class = 'GIT_EAUTH'
  )
  expect_error(
    git_clone(remote, path = tempfile(), ssh_key = 'pat.bin'),
    'load key',
    class = 'GIT_EAUTH'
  )

  # Also test password as a callback function
  for (keyfile in c("ecdsa.key", "rsa3072.key", "ed25519.key")) {
    target <- tempfile()
    repo <- git_clone(
      remote,
      path = target,
      ssh_key = keyfile,
      password = function(...) {
        'testingjerry'
      }
    )
    expect_true(file.exists(file.path(target, 'hello')))

    # Test ls-remote auth
    git_remote_ls(repo = target, ssh_key = keyfile, password = function(...) {
      'testingjerry'
    })
  }
})

# Access token for dummy account with minimal rights
test_that("HTTP user/pass auth", {
  skip_if_offline('github.com')
  # Disable user PAT
  Sys.unsetenv("GITHUB_PAT")

  # Test with password
  enc <- readBin('pat.bin', raw(), 1e3)
  dec <- openssl::rsa_decrypt(enc, 'key.pem', password = 'testingjerry')
  target2 <- file.path(tempdir(), 'testprivate2')
  repo <- git_clone(
    'https://testingjerry@github.com/jeroenooms/testprivate',
    path = target2,
    password = rawToChar(dec)
  )
  expect_true(file.exists(file.path(target2, 'hello')))

  # Test with password in URL
  target3 <- file.path(tempdir(), 'testprivate3')
  repo <- git_clone(
    sprintf(
      'https://testingjerry:%s@github.com/jeroenooms/testprivate',
      rawToChar(dec)
    ),
    path = target3
  )
  expect_true(file.exists(file.path(target3, 'hello')))

  # Test that repo is private
  expect_error(
    git_clone(
      'https://github.com/jeroenooms/testprivate',
      password = "bla",
      path = tempfile()
    ),
    'Authentication',
    class = 'GIT_EAUTH'
  )

  # Test with PAT
  Sys.setenv(GITHUB_PAT = rawToChar(dec))
  on.exit(Sys.unsetenv("GITHUB_PAT"))
  target4 <- file.path(tempdir(), 'testprivate4')
  repo <- git_clone('https://github.com/jeroenooms/testprivate', path = target4)
  expect_true(file.exists(file.path(target4, 'hello')))
  heads <- git_remote_ls(repo = repo)
  expect_is(heads, 'data.frame')
  expect_equal(git_remote_info(repo = repo)$head, "refs/remotes/origin/master")

  # Try with user in URL
  target5 <- file.path(tempdir(), 'testprivate5')
  repo <- git_clone(
    'https://nobody@github.com/jeroenooms/testprivate',
    path = target5
  )
  expect_true(file.exists(file.path(target5, 'hello')))
  heads <- git_remote_ls(repo = repo)
  expect_is(heads, 'data.frame')
  expect_equal(git_remote_info(repo = repo)$head, "refs/remotes/origin/master")
})

#' Git Diff
#'
#' @description
#' View changes in a commit or in the current working directory.
#'
#' * `git_diff()` returns a data frame with information about a commit patch.
#' * `git_diff_patch()` is shortcode for `git_diff()$patch`.
#'
#' @export
#' @inheritParams git_open
#' @family git
#' @param ref a reference such as `"HEAD"`, or a commit id, or `NULL`
#' to the diff the working directory against the repository index.
#' @returns
#' * `git_diff()` returns a data frame.
#' * `git_diff_patch()` returns a character vector.
#' @useDynLib gert R_git_diff_list
#' @git diff
git_diff <- function(ref = NULL, repo = '.') {
  repo <- git_open(repo)
  ref <- as.character(ref)
  .Call(R_git_diff_list, repo, ref)
}

#' @export
#' @rdname git_diff
git_diff_patch <- function(ref = NULL, repo = '.') {
  git_diff(ref = ref, repo = repo)$patch
}

if(!file.exists('clone.o') && !file.exists("../.deps/libgit2/include/git2.h")){
  unlink("../.deps", recursive = TRUE)
  url <- if(grepl("aarch", R.version$platform)){
    "https://github.com/r-windows/bundles/releases/download/libgit2-1.9.1/libgit2-1.9.1-clang-aarch64.tar.xz"
  } else if(grepl("clang", Sys.getenv('R_COMPILED_BY'))){
    "https://github.com/r-windows/bundles/releases/download/libgit2-1.9.1/libgit2-1.9.1-clang-x86_64.tar.xz"
  } else if(getRversion() >= "4.2") {
    "https://github.com/r-windows/bundles/releases/download/libgit2-1.9.1/libgit2-1.9.1-ucrt-x86_64.tar.xz"
  } else {
    "https://github.com/rwinlib/libgit2/archive/v1.7.1.tar.gz"
  }
  download.file(url, basename(url), quiet = TRUE)
  dir.create("../.deps", showWarnings = FALSE)
  untar(basename(url), exdir = "../.deps", tar = 'internal')
  unlink(basename(url))
  setwd("../.deps")
  file.rename(list.files(), 'libgit2')
}

#' Git Archive
#'
#' Exports the files in your repository to a zip file that
#' is returned by the function.
#'
#' @export
#' @rdname git_archive
#' @name git_archive
#' @family git
#' @inheritParams git_open
#' @param file name of the output zip file. Default is returned
#' by the function
#' @return path to the zip file that was created
git_archive_zip <- function(file = NULL, repo = "."){
  repo <- git_open(repo = repo)
  tmp <- tempfile(fileext = '.zip')
  git_archive_internal(tmp, repo = repo)
  if(!length(file)){
    file <- paste0(basename(git_info(repo)$path), ".zip")
  }
  file.copy(tmp, file, overwrite = TRUE)
  return(file)
}

git_archive_internal <- function(outfile, repo){
  tryCatch({
    git_stash_save(repo = repo)
    on.exit(git_stash_pop(repo = repo))
  }, GIT_ENOTFOUND = function(e){})
  files <- git_ls(repo = repo)$path
  wd <- getwd()
  on.exit(setwd(wd), add = TRUE)
  setwd(git_info(repo = repo)$path)
  zip::zip(outfile, files = files, recurse = FALSE)
}

#' Stashing changes
#'
#' Temporary stash away changed from the working directory.
#'
#' @export
#' @rdname git_stash
#' @name git_stash
#' @family git
#' @inheritParams git_open
#' @useDynLib gert R_git_stash_save
#' @param message optional message to store the stash
#' @param keep_index changes already added to the index are left intact in
#' the working directory
#' @param include_untracked untracked files are also stashed and then
#' cleaned up from the working directory
#' @param include_ignored ignored files are also stashed and then cleaned
#' up from the working directory
git_stash_save <- function(message = "", keep_index = FALSE, include_untracked = FALSE,
                           include_ignored = FALSE, repo = "."){
  repo <- git_open(repo)
  keep_index <- as.logical(keep_index)
  include_untracked <- as.logical(include_untracked)
  include_ignored <- as.logical(include_ignored)
  .Call(R_git_stash_save, repo, message, keep_index, include_untracked, include_ignored)
}

#' @export
#' @rdname git_stash
#' @useDynLib gert R_git_stash_pop
#' @param index The position within the stash list. 0 points to the
#' most recent stashed state.
git_stash_pop <- function(index = 0, repo = "."){
  repo <- git_open(repo)
  .Call(R_git_stash_pop, repo, index)
}

#' @export
#' @rdname git_stash
#' @useDynLib gert R_git_stash_drop
git_stash_drop <- function(index = 0, repo = "."){
  repo <- git_open(repo)
  .Call(R_git_stash_drop, repo, index)
}

#' @export
#' @rdname git_stash
#' @useDynLib gert R_git_stash_list
git_stash_list <- function(repo = "."){
  repo <- git_open(repo)
  .Call(R_git_stash_list, repo)
}

#' Stage and commit changes
#'
#' @description
#' To commit changes, start by *staging* the files to be included in the commit
#' using `git_add()` or `git_rm()`. Use `git_status()` to see an overview of
#' staged and unstaged changes, and finally `git_commit()` creates a new commit
#' with currently staged files.
#'
#' `git_commit_all()` is a convenience function that automatically stages and
#' commits all modified files. Note that `git_commit_all()` does **not** add
#' new, untracked files to the repository. You need to make an explicit call to
#' `git_add()` to start tracking new files.
#'
#' `git_log()` shows the most recent commits and `git_ls()` lists all the files
#' that are being tracked in the repository. `git_stat_files()`
#'
#' @export
#' @rdname git_commit
#' @name git_commit
#' @family git
#' @inheritParams git_open
#' @param message a commit message
#' @param author A [git_signature] value, default is [git_signature_default()].
#' @param committer A [git_signature] value, default is same as `author`
#' @return
#' * `git_status()`, `git_ls()`: A data frame with one row per file
#' * `git_log()`: A data frame with one row per commit
#' * `git_commit()`, `git_commit_all()`: A SHA
#' @useDynLib gert R_git_commit_create
#' @examples
#' oldwd <- getwd()
#' repo <- file.path(tempdir(), "myrepo")
#' git_init(repo)
#' setwd(repo)
#'
#' # Set a user if no default
#' if(!user_is_configured()){
#'   git_config_set("user.name", "Jerry")
#'   git_config_set("user.email", "jerry@gmail.com")
#' }
#'
#' writeLines(letters[1:6], "alphabet.txt")
#' git_status()
#'
#' git_add("alphabet.txt")
#' git_status()
#'
#' git_commit("Start alphabet file")
#' git_status()
#'
#' git_ls()
#'
#' git_log()
#'
#' cat(letters[7:9], file = "alphabet.txt", sep = "\n", append = TRUE)
#' git_status()
#'
#' git_commit_all("Add more letters")
#'
#' # cleanup
#' setwd(oldwd)
#' unlink(repo, recursive = TRUE)
git_commit <- function(message, author = NULL, committer = NULL, repo = '.'){
  repo <- git_open(repo)
  if(!length(author)) {
    author <- git_signature_default(repo = repo)
  } else {
    git_signature_parse(author) #validate
  }
  if(!length(committer)){
    committer <- author
  } else {
    git_signature_parse(committer)  #validate
  }
  stopifnot(is.character(message), length(message) == 1)
  status <- git_status(repo = repo)
  if(!any(status$staged))
    stop("No staged files to commit. Run git_add() to select files.")
  merge_parents <- git_merge_parent_heads(repo = repo)
  .Call(R_git_commit_create, repo, message, author, committer, merge_parents)
}

#' @export
#' @rdname git_commit
git_commit_all <- function(message, author = NULL, committer = NULL, repo = '.'){
  repo <- git_open(repo)
  unstaged <- git_status(staged = FALSE, repo = repo)

  changes <- unstaged$file[unstaged$status %in% c("modified", "renamed", "typechange")]
  if(length(changes))
    git_add(changes, repo = repo)

  deleted <- unstaged$file[unstaged$status == "deleted"]
  if(length(deleted))
    git_rm(deleted, repo = repo)

  git_commit(message = message, author = author, committer = committer, repo = repo)
}

#' @export
#' @rdname git_commit
#' @useDynLib gert R_git_commit_info
git_commit_info <- function(ref = "HEAD", repo = '.'){
  repo <- git_open(repo)
  .Call(R_git_commit_info, repo, ref)
}

#' @export
#' @rdname git_commit
#' @useDynLib gert R_git_commit_id
git_commit_id <- function(ref = "HEAD", repo = '.'){
  repo <- git_open(repo)
  .Call(R_git_commit_id, repo, ref)
}

#' @export
#' @rdname git_commit
#' @param ancestor a reference to a potential ancestor commit
#' @useDynLib gert R_git_commit_descendant
git_commit_descendant_of <- function(ancestor, ref = 'HEAD', repo = '.'){
  repo <- git_open(repo)
  .Call(R_git_commit_descendant, repo, ref, ancestor)
}

#' @export
#' @rdname git_commit
#' @param files vector of paths relative to the git root directory.
#' Use `"."` to stage all changed files.
#' @param force add files even if in gitignore
#' @useDynLib gert R_git_repository_add
git_add <- function(files, force = FALSE, repo = '.'){
  repo <- git_open(repo)
  info <- git_info(repo)
  normalizePath(file.path(info$path, files), mustWork = FALSE)
  force <- as.logical(force)
  .Call(R_git_repository_add, repo, files, force)
  git_status(repo = repo)
}

#' @export
#' @rdname git_commit
#' @useDynLib gert R_git_repository_rm
git_rm <- function(files, repo = '.'){
  repo <- git_open(repo)
  info <- git_info(repo)
  normalizePath(file.path(info$path, files), mustWork = FALSE)
  .Call(R_git_repository_rm, repo, files)
  git_status(repo = repo)
}

#' @export
#' @rdname git_commit
#' @useDynLib gert R_git_status_list
#' @param staged return only staged (TRUE) or unstaged files (FALSE).
#' Use `NULL` or `NA` to show both (default).
git_status <- function(staged = NULL, repo = '.'){
  repo <- git_open(repo)
  staged <- as.logical(staged)
  df <- .Call(R_git_status_list, repo, staged)
  df[order(df$file), ,drop = FALSE]
}

#' @export
#' @rdname git_commit
#' @useDynLib gert R_git_conflict_list
git_conflicts <- function(repo = '.'){
  repo <- git_open(repo)
  .Call(R_git_conflict_list, repo)
}

#' @export
#' @rdname git_commit
#' @useDynLib gert R_git_repository_ls
git_ls <- function(repo = '.'){
  repo <- git_open(repo)
  .Call(R_git_repository_ls, repo)
}

#' @export
#' @rdname git_commit
#' @useDynLib gert R_git_commit_log
#' @param ref revision string with a branch/tag/commit value
#' @param max lookup at most latest n parent commits
git_log <- function(ref = "HEAD", max = 100, repo = "."){
  repo <- git_open(repo)
  ref <- as.character(ref)
  max <- as.integer(max)
  .Call(R_git_commit_log, repo, ref, max)
}

#' @export
#' @rdname git_commit
#' @useDynLib gert R_git_stat_files
git_stat_files <- function(files, ref = "HEAD", repo = '.'){
  repo <- git_open(repo)
  files <- as.character(files)
  .Call(R_git_stat_files, repo, files, ref)
}

assert_string <- function(x){
  if(!is.character(x) || !length(x))
    stop("Argument must be a string of length 1")
}

test_that("rebasing things", {
  skip_if_offline('github.com')
  repo <- file.path(tempdir(), 'gert')
  if(!file.exists(repo)) git_clone('https://github.com/r-lib/gert', path = repo)
  git_branch_create('backup', checkout = FALSE, repo = repo)

  # Original log
  orig <- git_log(max = 10, repo = repo)

  # Drop some commits, and fast-forward them back
  git_reset_hard('HEAD~5', repo = repo)
  expect_equal(git_log(max = 5, repo = repo)$commit, utils::tail(orig$commit, 5))
  expect_equal(git_ahead_behind(repo = repo), list(ahead = 0, behind = 5,
    local = git_commit_id(repo = repo), upstream = git_commit_id('origin/HEAD', repo = repo)))
  git_pull(repo = repo)
  expect_equal(orig, git_log(max = 10, repo = repo))

  # Same with rebase
  git_reset_hard('HEAD~5', repo = repo)
  expect_equal(git_log(max = 5, repo = repo)$commit, utils::tail(orig$commit, 5))
  expect_equal(git_ahead_behind(repo = repo), list(ahead = 0, behind = 5,
    local = git_commit_id(repo = repo), upstream = git_commit_id('origin/HEAD', repo = repo)))
  git_pull(rebase = TRUE, repo = repo)
  expect_equal(orig, git_log(max = 10, repo = repo))

  # Now rebase a commit
  git_reset_hard('HEAD~5', repo = repo)
  expect_equal(git_log(max = 5, repo = repo)$commit, utils::tail(orig$commit, 5))
  writeLines("some random change", file.path(repo, 'randomfile.txt'))
  git_add(".", repo = repo)
  commit_id <- git_commit("Added a local change", repo = repo, author = "Jerry <test@jerry.nl>")
  commit_msg <- git_commit_info(commit_id, repo = repo)$message

  # We are 1 ahead and 5 behind
  expect_equal(git_ahead_behind(repo = repo), list(ahead = 1, behind = 5,
    local = git_commit_id(repo = repo), upstream = git_commit_id('origin/HEAD', repo = repo)))

  # Confirm that we cannot fast forward
  upstream <- git_info(repo = repo)$upstream
  expect_error(git_branch_fast_forward(upstream, repo = repo))

  # Pull with rebase
  rebase_info <- git_rebase_list(repo = repo)
  expect_equal(rebase_info$commit, commit_id)
  git_pull(rebase = TRUE, repo = repo)
  newlog <- git_log(max = 11, repo = repo)
  expect_equal(c(commit_msg, orig$message), newlog$message)
  expect_equal(orig$commit, newlog$commit[-1])
  expect_equal(git_ahead_behind(repo = repo), list(ahead = 1, behind = 0,
    local = git_commit_id(repo = repo), upstream = git_commit_id('origin/HEAD', repo = repo)))

  # Check that patch matches
  expect_equal(git_diff(commit_id, repo = repo), git_diff(newlog$commit[1], repo = repo))

  # Merge changes into another branch
  main <- git_info(repo = repo)$shorthand
  git_branch_checkout('backup', repo = repo)
  git_merge(main, repo = repo)
  expect_equal(newlog, git_log(max = 11, repo = repo))

  # Cleanup from master
  git_branch_checkout(main, repo = repo)
  git_reset_hard("HEAD^", repo = repo)
  git_branch_delete('backup', repo = repo)
})

test_that("creating signatures", {
  now <- Sys.time()
  name <- 'Testing Jerry'
  email <- 'test@jerry.com'
  author <- sprintf("%s <%s>", name, email)

  sig <- git_signature(name, email)
  sigdata <- git_signature_parse(sig)
  expect_equal(sigdata$name, name)
  expect_equal(sigdata$email, email)
  expect_lt(difftime(sigdata$time, now, 'secs'), 1)

  yesterday <- now - 24*60*60
  sig <- git_signature(name, email, time = yesterday)
  sigdata <- git_signature_parse(sig)
  expect_equal(sigdata$name, name)
  expect_lt(difftime(sigdata$time, yesterday, 'secs'), 1)
})

test_that("adding and removing files", {
  repo <- git_init(tempfile("gert-tests-commit"))
  on.exit(unlink(repo, recursive = TRUE))

  expect_equal(nrow(git_ls(repo)), 0)
  write.csv(iris, file.path(repo, 'iris.csv'))
  write.csv(cars, file.path(repo, 'cars.csv'))
  git_add(c('cars.csv', 'iris.csv'), repo = repo)
  expect_equal(nrow(git_ls(repo)), 2)
  expect_equal(git_ls(repo)$path, c('cars.csv', 'iris.csv'))
  git_rm(c('cars.csv', 'iris.csv'), repo = repo)
  expect_equal(nrow(git_ls(repo)), 0)
  remotes <- git_remote_list(repo)
  expect_equal(nrow(remotes), 0)
})

test_that("creating a commit", {
  repo <- git_init(tempfile("gert-tests-commit"))
  on.exit(unlink(repo, recursive = TRUE))
  configure_local_user(repo)

  expect_equal(nrow(git_ls(repo)), 0)
  dir.create(file.path(repo, 'src'))
  write.csv(cars, file.path(repo, 'src', 'cars.csv'))
  git_add('src/cars.csv', repo = repo)
  git_commit("Added cars.csv file", repo = repo)
  stats <- git_stat_files('src/cars.csv', repo = repo)
  expect_equal(stats$file, 'src/cars.csv')
  expect_equal(stats$commits, 1)
  expect_equal(stats$created, stats$modified)

  # Another commit before that
  write.csv(iris, file.path(repo, 'iris.csv'))
  git_add("iris.csv", repo = repo)
  timestamp <- round(Sys.time() - 48*60*60)
  sig2 <- git_signature('nobody', 'nobody@gmail.com', time = timestamp)
  git_commit("Added iris.csv also", author = sig2, repo = repo)

  # Inspect the log file
  log <- git_log(repo = repo)
  expect_equal(log$author[2], local_author(repo))
  expect_equal(log$time[1], timestamp)
})

test_that("creating a commit in another directory without author works", {
  path <- tempfile("gert-tests-commit")
  on.exit(unlink(path, recursive = TRUE))

  dir.create(path)
  repo <- git_init(path)
  configure_local_user(repo)
  writeLines("content", file.path(path, "file"))
  git_add("file", repo = path)
  git_commit("Added file", repo = repo)

  log <- git_log(repo = repo)
  expect_equal(log$author, local_author(repo))
})

test_that("status reports a conflicted file", {
  repo <- git_init(tempfile("gert-test-conflicts"))
  on.exit(unlink(repo, recursive = TRUE))
  configure_local_user(repo)

  foo_path <- file.path(repo, "foo.txt")
  bar_path <- file.path(repo, "bar.txt")

  writeLines("cranky-crab-legs", foo_path)
  git_add("foo.txt", repo = repo)
  base <- git_commit("Add a file", repo = repo)

  git_branch_create("my-branch", repo = repo)
  writeLines("cranky-CRAB-LEGS", foo_path)
  git_add("foo.txt", repo = repo)
  git_commit("Uppercase last 2 words", repo = repo)

  git_branch_checkout("master", repo = repo)
  expect_equal(git_merge_analysis("my-branch", repo = repo), "fastforward")

  writeLines("CRANKY-CRAB-legs", foo_path)
  git_add("foo.txt", repo = repo)
  git_commit("Uppercase first 2 words", repo = repo)

  writeLines(c("CRANKY-CRAB-legs", "One more line"), foo_path)
  git_add("foo.txt", repo = repo)
  git_commit("Add another commit", repo = repo)

  writeLines("Nothing special", bar_path)
  git_add("bar.txt", repo = repo)
  git_commit("Non-conflicting commit", repo = repo)

  expect_equal(base, git_merge_find_base("my-branch", repo = repo))
  expect_equal(git_merge_analysis(base, repo = repo), "up_to_date")
  expect_equal(git_merge_analysis('my-branch', repo = repo), "normal")

  # What if we would rebase
  rebase_info <- git_rebase_list("my-branch", repo = repo)
  expect_equal(rebase_info$type, rep("pick", 3))
  expect_equal(rebase_info$commit, rev(head(git_log(repo = repo), -1)$commit))
  expect_equal(rebase_info$conflicts, c(T,T,F))
  expect_error(git_rebase_commit("my-branch", repo = repo), class = "GIT_EMERGECONFLICT")
  expect_error(git_branch_fast_forward("my-branch", repo = repo))

  # Merge returns FALSE due to conflicts
  git_merge("my-branch", repo = repo)
  status <- git_status(repo = repo)
  expect_equal(status$file, "foo.txt")
  expect_equal(status$status, "conflicted")
  expect_equal(git_conflicts(repo = repo)$our, "foo.txt")
  expect_false(status$staged)

  # Abort and cleanup
  git_merge_abort(repo = repo)
  expect_equal(nrow(git_status(repo = repo)), 0)
  expect_equal(nrow(git_conflicts(repo = repo)), 0)

  # Merge again :)
  git_merge("my-branch", repo = repo)
  status <- git_status(repo = repo)
  expect_equal(status$file, "foo.txt")
  expect_equal(status$status, "conflicted")
  expect_equal(git_conflicts(repo = repo)$our, "foo.txt")
  expect_false(status$staged)

  # Resolve the conflict
  conflicts <- git_conflicts(repo = repo)
  expect_equal(conflicts$our, status$file)
  writeLines("Conflict is resolved", foo_path)
  git_add('foo.txt', repo = repo)
  git_commit("Resolve merge conflict", repo = repo)
  expect_length(git_conflicts(repo = repo)$our, 0)
  expect_length(git_status(repo = repo)$file, 0)
})

test_that("creating signatures", {
  now <- Sys.time()
  name <- 'Testing Jerry'
  email <- 'test@jerry.com'
  author <- sprintf("%s <%s>", name, email)

  sig <- git_signature(name, email)
  sigdata <- git_signature_parse(sig)
  expect_equal(sigdata$name, name)
  expect_equal(sigdata$email, email)
  expect_lt(difftime(sigdata$time, now, 'secs'), 1)

  yesterday <- now - 24*60*60
  sig <- git_signature(name, email, time = yesterday)
  sigdata <- git_signature_parse(sig)
  expect_equal(sigdata$name, name)
  expect_lt(difftime(sigdata$time, yesterday, 'secs'), 1)
})

test_that("adding and removing files", {
  repo <- git_init(tempfile("gert-tests-commit"))
  on.exit(unlink(repo, recursive = TRUE))

  expect_equal(nrow(git_ls(repo = repo)), 0)
  write.csv(iris, file.path(repo, 'iris.csv'))
  write.csv(cars, file.path(repo, 'cars.csv'))
  git_add(c('cars.csv', 'iris.csv'), repo = repo)
  expect_equal(nrow(git_ls(repo = repo)), 2)
  expect_equal(git_ls(repo = repo)$path, c('cars.csv', 'iris.csv'))
  git_rm(c('cars.csv', 'iris.csv'), repo = repo)
  expect_equal(nrow(git_ls(repo = repo)), 0)
  remotes <- git_remote_list(repo)
  expect_equal(nrow(remotes), 0)
})

test_that("creating a commit", {
  repo <- git_init(tempfile("gert-tests-commit"))
  on.exit(unlink(repo, recursive = TRUE))
  configure_local_user(repo)

  expect_equal(nrow(git_ls(repo = repo)), 0)
  dir.create(file.path(repo, 'src'))
  write.csv(cars, file.path(repo, 'src', 'cars.csv'))
  git_add('src/cars.csv', repo = repo)
  git_commit("Added cars.csv file", repo = repo)
  stats <- git_stat_files('src/cars.csv', repo = repo)
  expect_equal(stats$file, 'src/cars.csv')
  expect_equal(stats$commits, 1)
  expect_equal(stats$created, stats$modified)

  # Another commit before that
  write.csv(iris, file.path(repo, 'iris.csv'))
  git_add("iris.csv", repo = repo)
  timestamp <- round(Sys.time() - 48*60*60)
  sig2 <- git_signature('nobody', 'nobody@gmail.com', time = timestamp)
  git_commit("Added iris.csv also", author = sig2, repo = repo)

  # Inspect the log file
  log <- git_log(repo = repo)
  expect_equal(log$author[2], local_author(repo))
  expect_equal(log$time[1], timestamp)
})

test_that("creating a commit in another directory without author works", {
  path <- tempfile("gert-tests-commit")
  on.exit(unlink(path, recursive = TRUE))

  dir.create(path)
  repo <- git_init(path)
  configure_local_user(repo)
  writeLines("content", file.path(path, "file"))
  git_add("file", repo = path)
  git_commit("Added file", repo = repo)

  log <- git_log(repo = repo)
  expect_equal(log$author, local_author(repo))
})

test_that("status reports a conflicted file", {
  repo <- git_init(tempfile("gert-test-conflicts"))
  on.exit(unlink(repo, recursive = TRUE))
  configure_local_user(repo)

  foo_path <- file.path(repo, "foo.txt")
  bar_path <- file.path(repo, "bar.txt")

  writeLines("cranky-crab-legs", foo_path)
  git_add("foo.txt", repo = repo)
  base <- git_commit("Add a file", repo = repo)

  git_branch_create("my-branch", repo = repo)
  writeLines("cranky-CRAB-LEGS", foo_path)
  git_add("foo.txt", repo = repo)
  git_commit("Uppercase last 2 words", repo = repo)

  git_branch_checkout("master", repo = repo)
  expect_equal(git_merge_analysis("my-branch", repo = repo), "fastforward")

  writeLines("CRANKY-CRAB-legs", foo_path)
  git_add("foo.txt", repo = repo)
  git_commit("Uppercase first 2 words", repo = repo)

  writeLines(c("CRANKY-CRAB-legs", "One more line"), foo_path)
  git_add("foo.txt", repo = repo)
  git_commit("Add another commit", repo = repo)

  writeLines("Nothing special", bar_path)
  git_add("bar.txt", repo = repo)
  git_commit("Non-conflicting commit", repo = repo)

  expect_equal(base, git_merge_find_base("my-branch", repo = repo))
  expect_equal(git_merge_analysis(base, repo = repo), "up_to_date")
  expect_equal(git_merge_analysis('my-branch', repo = repo), "normal")

  # What if we would rebase
  rebase_info <- git_rebase_list("my-branch", repo = repo)
  expect_equal(rebase_info$type, rep("pick", 3))
  expect_equal(rebase_info$commit, rev(head(git_log(repo = repo), -1)$commit))
  expect_equal(rebase_info$conflicts, c(T,T,F))
  expect_error(git_rebase_commit("my-branch", repo = repo), class = "GIT_EMERGECONFLICT")
  expect_error(git_branch_fast_forward("my-branch", repo = repo))

  # Merge returns FALSE due to conflicts
  git_merge("my-branch", repo = repo)
  status <- git_status(repo = repo)
  expect_equal(status$file, "foo.txt")
  expect_equal(status$status, "conflicted")
  expect_equal(git_conflicts(repo = repo)$our, "foo.txt")
  expect_false(status$staged)

  # Abort and cleanup
  git_merge_abort(repo = repo)
  expect_equal(nrow(git_status(repo = repo)), 0)
  expect_equal(nrow(git_conflicts(repo = repo)), 0)

  # Merge again :)
  git_merge("my-branch", repo = repo)
  status <- git_status(repo = repo)
  expect_equal(status$file, "foo.txt")
  expect_equal(status$status, "conflicted")
  expect_equal(git_conflicts(repo = repo)$our, "foo.txt")
  expect_false(status$staged)

  # Resolve the conflict
  conflicts <- git_conflicts(repo = repo)
  expect_equal(conflicts$our, status$file)
  writeLines("Conflict is resolved", foo_path)
  git_add('foo.txt', repo = repo)
  git_commit("Resolve merge conflict", repo = repo)
  expect_length(git_conflicts(repo = repo)$our, 0)
  expect_length(git_status(repo = repo)$file, 0)
})

test_that("git_restore restores a modified file from HEAD", {
  repo <- git_init(tempfile("gert-tests-restore"))
  on.exit(unlink(repo, recursive = TRUE))
  configure_local_user(repo)

  writeLines("original", file.path(repo, "hello.txt"))
  git_add("hello.txt", repo = repo)
  git_commit("First commit", repo = repo)

  writeLines("modified", file.path(repo, "hello.txt"))
  expect_equal(readLines(file.path(repo, "hello.txt")), "modified")

  git_restore("hello.txt", repo = repo)
  expect_equal(readLines(file.path(repo, "hello.txt")), "original")
})

test_that("git_restore restores from a specific ref", {
  repo <- git_init(tempfile("gert-tests-restore-ref"))
  on.exit(unlink(repo, recursive = TRUE))
  configure_local_user(repo)

  writeLines("v1", file.path(repo, "hello.txt"))
  writeLines("v1", file.path(repo, "keep.txt"))
  writeLines("v1", file.path(repo, "wip.txt"))
  writeLines("v1", file.path(repo, "untracked.txt"))

  git_add(c("hello.txt", "keep.txt", "wip.txt"), repo = repo)
  first <- git_commit("First commit", repo = repo)

  writeLines("v2", file.path(repo, "hello.txt"))
  writeLines("v2", file.path(repo, "keep.txt"))
  writeLines("v2", file.path(repo, "wip.txt"))
  git_add(c("hello.txt", "keep.txt"), repo = repo)
  git_commit("Second commit", repo = repo)

  git_restore("hello.txt", ref = first, repo = repo)
  expect_equal(readLines(file.path(repo, "hello.txt")), "v1")
  expect_equal(readLines(file.path(repo, "keep.txt")), "v2")
  expect_equal(readLines(file.path(repo, "wip.txt")), "v2")
  expect_equal(readLines(file.path(repo, "untracked.txt")), "v1")

})

test_that("git_restore restores all files with path = '.'", {
  repo <- git_init(tempfile("gert-tests-restore-dot"))
  on.exit(unlink(repo, recursive = TRUE))
  configure_local_user(repo)

  writeLines("original a", file.path(repo, "a.txt"))
  writeLines("original b", file.path(repo, "b.txt"))
  git_add(c("a.txt", "b.txt"), repo = repo)
  git_commit("First commit", repo = repo)

  writeLines("modified a", file.path(repo, "a.txt"))
  writeLines("modified b", file.path(repo, "b.txt"))

  git_restore(".", repo = repo)
  expect_equal(readLines(file.path(repo, "a.txt")), "original a")
  expect_equal(readLines(file.path(repo, "b.txt")), "original b")
})

test_that("git_restore raises an error for an untracked path", {
  repo <- git_init(tempfile("gert-tests-restore-untracked"))
  on.exit(unlink(repo, recursive = TRUE))
  configure_local_user(repo)

  writeLines("hello", file.path(repo, "hello.txt"))
  git_add("hello.txt", repo = repo)
  git_commit("First commit", repo = repo)

  expect_error(git_restore("nottracked.txt", repo = repo), "not tracked by git")
})

test_that("git_restore raises an error for an invalid ref", {
  repo <- git_init(tempfile("gert-tests-restore-invalid"))
  on.exit(unlink(repo, recursive = TRUE))
  configure_local_user(repo)

  writeLines("hello", file.path(repo, "hello.txt"))
  git_add("hello.txt", repo = repo)
  git_commit("First commit", repo = repo)

  expect_error(git_restore("hello.txt", ref = "notaref", repo = repo), "notaref")
})

#' Open local repository
#'
#' Returns a pointer to a libgit2 repository object.This function is mainly
#' for internal use; users should simply reference a repository in gert by
#' by the path to the directory.
#'
#' @export
#' @param repo The path to the git repository. If the directory is not a
#' repository, parent directories are considered (see [git_find]). To disable
#' this search, provide the filepath protected with [I()]. When using this
#' parameter, always explicitly call by name (i.e. `repo = `) because future
#' versions of gert may have additional parameters.
#' @return an pointer to the libgit2 repository
#' @useDynLib gert R_git_repository_open
#' @examples
#' r <- tempfile(pattern = "gert")
#' git_init(r)
#' r_ptr <- git_open(r)
#' r_ptr
#' git_open(r_ptr)
#' git_info(r)
#'
#' # cleanup
#' unlink(r, recursive = TRUE)
git_open <- function(repo = '.') {
  if (inherits(repo, 'git_repo_ptr')) {
    return(repo)
  } else if (!is.character(repo)) {
    stop("repo argument must be a filepath or an existing repository object")
  }
  search <- !inherits(repo, 'AsIs')
  path <- normalizePath(path.expand(repo), mustWork = FALSE)
  out <- git_repository_open(path, search)
  if (is_rstudio_ide()) {
    cl <- substitute(rstudioapi::executeCommand("vcsRefresh"))
    do.call(on.exit, list(cl, add = TRUE), envir = parent.frame())
  }
  return(out)
}


git_repository_open <- local({
  cache <- new.env(parent = emptyenv())
  function(path, search) {
    key <- digest(list(path, search))
    if (
      !exists(key, envir = cache) || !isTRUE(getOption('gert.use.repo.cache'))
    ) {
      val <- .Call(R_git_repository_open, path, search)
      assign(key, val, envir = cache)
    }
    get0(key, cache)
  }
})

is_rstudio_ide <- function() {
  interactive() &&
    identical(Sys.getenv('RSTUDIO'), '1') &&
    !nchar(Sys.getenv('Disable_vcsRefresh')) &&
    rstudioapi::isAvailable()
}

#' @export
print.git_repo_ptr <- function(x, ...) {
  info <- git_info(x)

  type = "git repository"
  if (info$bare) {
    type = paste(type, "(bare)")
  }

  cat(sprintf(
    "<%s>: %s[@%s]\n",
    type,
    normalizePath(info$path),
    info$shorthand
  ))
}

#' @useDynLib gert R_git_repository_path
git_repo_path <- function(repo) {
  invisible(.Call(R_git_repository_path, repo))
}

digest <- function(x) {
  as.character(openssl::sha1(serialize(x, NULL)))
}

#' Author Signature
#'
#' A signature contains the author and timestamp of a commit. Each commit
#' includes a signature of the author and committer (which can be identical).
#'
#' A signature string has format `"Real Name <email> timestamp tzoffset"`. The
#' `timestamp tzoffset` piece can be omitted in which case the current local
#' time is used. If not omitted, `timestamp` must contain the number
#' of seconds since the Unix epoch and `tzoffset` is the timezone offset in
#' `hhmm` format (note the lack of a colon separator)
#'
#' @export
#' @rdname git_signature
#' @name git_signature
#' @family git
#' @inheritParams git_open
#' @useDynLib gert R_git_signature_default
#' @git signature
#' @examples # Your default user
#' try(git_signature_default())
#'
#' # Specify explicit name and email
#' git_signature("Some committer", "sarah@gmail.com")
#'
#' # Create signature for an hour ago
#' (sig <- git_signature("Han", "han@company.com", Sys.time() - 3600))
#'
#' # Parse a signature
#' git_signature_parse(sig)
#' git_signature_parse("Emma <emma@mu.edu>")
git_signature_default <- function(repo = '.') {
  repo <- git_open(repo)
  sig <- .Call(R_git_signature_default, repo)
  sprintf("%s <%s>", sig$name, sig$email)
}

#' @export
#' @rdname git_signature
#' @useDynLib gert R_git_signature_create
#' @param name Real name of the committer
#' @param email Email address of the committer
#' @param time timestamp of class POSIXt or NULL
git_signature <- function(name, email, time = NULL) {
  assert_string(name)
  assert_string(email)
  if (length(time)) {
    time <- as.POSIXct(time)
    tz <- format(time, "%z")
    minutes <- as.integer(substring(tz, 4))
    hours <- as.integer(substring(tz, 1, 3))
    offset <- hours * 60 + minutes
  }
  sig <- .Call(R_git_signature_create, name, email, time, offset)
  if (length(time)) {
    sig_data_to_string(sig)
  } else {
    sig_data_to_string(list(name = sig$name, email = sig$email))
  }
}

#' @export
#' @rdname git_signature
#' @param sig string in proper `"First Last <your@email.com>"` format, see details.
#' @useDynLib gert R_git_signature_parse
git_signature_parse <- function(sig) {
  assert_string(sig)
  .Call(R_git_signature_parse, sig)
}

offset_to_string <- function(offset) {
  if (length(offset) && is.numeric(offset)) {
    hours <- as.integer(offset %/% 60)
    mins <- as.integer(offset %% 60)
    sprintf('%+03d%02d', hours, mins)
  } else {
    ""
  }
}

sig_data_to_string <- function(x) {
  sig <- sprintf("%s <%s>", x$name, x$email)
  if (length(x$time)) {
    sig <- paste(sig, unclass(x$time), offset_to_string(x$offset))
  }
  structure(trimws(sig), class = "gert_signature")
}

#' @export
print.gert_signature <- function(x, ...) {
  sig <- git_signature_parse(x)
  name <- sig$name
  email <- sig$email
  time <- format(sig$time, "%a %b %d %H:%M:%S %Y") #print as user local time
  offset <- offset_to_string(sig$offset)
  cat(sprintf(
    '[git signature]\nAuthor: %s <%s>\nDate: %s %s\n',
    name,
    email,
    time,
    offset
  ))
}

.onAttach <- function(libname, pkgname){
  config <- libgit2_config()
  ssh <- ifelse(config$ssh, "YES", "NO")
  https <- ifelse(config$https, "YES", "NO")
  packageStartupInform(
    "Linking to libgit2 v%s, ssh support: %s",
    as.character(config$version), ssh)
  user_config <- config$config.global
  if(length(user_config) && nchar(user_config)){
    if(is_windows()){
      user_config <- normalizePath(user_config, mustWork = FALSE)
    }
    packageStartupInform("Global config: %s", user_config)
  } else {
    packageStartupInform("No global .gitconfig found in: %s", config$config.home)
  }
  if(length(config$config.system) && nchar(config$config.system))
    packageStartupInform(paste0("System config: ", config$config.system))
  try({
    settings <- git_config_global()
    name <- subset(settings, name == 'user.name')$value
    email <- subset(settings, name == 'user.email')$value
    if(length(name) || length(email)){
      packageStartupInform("Default user: %s <%s>", as_string(name), as_string(email))
    } else {
      packageStartupInform("No default user configured")
    }
  })
}

.onLoad <- function(libname, pkgname) {
  # This should usually correspond to 'openssl version -d'
  if(isTRUE(have_static_libgit2())){
    certpath <- find_cert_dir()
    if(length(certpath)){
      cafile <- file.path(dirname(certpath), 'cert.pem')
      if(!file.exists(cafile))
        cafile <- NULL
      set_cert_locations(cafile, certpath)
    } else {
      warning("Unable to find directory with certificates", immediate. = TRUE)
    }
  }

  if (identical(Sys.getenv("IN_PKGDOWN"), "true") &&
      !global_user_is_configured()) {
    configure_global_user()
  }

  # Load tibble (if available) for pretty printing
  if(interactive() && is.null(.getNamespace('tibble'))){
    tryCatch({
      getNamespace('tibble')
    }, error = function(e){})
  }

  invisible()
}

as_string <- function(x){
  ifelse(length(x) > 0, x[1], NA_character_)
}

is_windows <- function(){
  identical(.Platform$OS.type, "windows")
}

is_solaris <- function(){
  isTRUE(grepl('solaris', R.version$platform, ignore.case = TRUE))
}

inform_impl <- function(message, subclass = NULL) {
  cnd <- structure(
    list(message = paste0(message, "\n")),
    class = c(subclass, "simpleMessage", "message", "condition")
  )
  withRestarts(
    muffleMessage = function() NULL,
    {
      signalCondition(cnd)
      cat(
        conditionMessage(cnd),
        file = if (interactive()) stdout() else stderr()
      )
    }
  )
}


inform <- function(...) inform_impl(sprintf(...), subclass = "gertMessage")

packageStartupInform <- function(...) {
  inform_impl(sprintf(...), subclass = c("gertMessage", "packageStartupMessage"))
}

message <- function(...) {
  stop("Internal error: use inform() instead of message()")
}

packageStartupMessage <- function(...) {
  stop("Internal error: use packageStartupInform() instead of packageStartupMessage()")
}

# libgit2 wants all paths utf-8
normalizePath <- function(...){
  enc2utf8(base::normalizePath(...))
}

#' Stage and commit changes
#'
#' @description
#' To commit changes, start by *staging* the files to be included in the commit
#' using `git_add()` or `git_rm()`. Use `git_status()` to see an overview of
#' staged and unstaged changes, and finally `git_commit()` creates a new commit
#' with currently staged files.
#'
#' `git_commit_all()` is a convenience function that automatically stages and
#' commits all modified files. Note that `git_commit_all()` does **not** add
#' new, untracked files to the repository. You need to make an explicit call to
#' `git_add()` to start tracking new files.
#'
#' `git_log()` shows the most recent commits and `git_ls()` lists all the files
#' that are being tracked in the repository. `git_stat_files()`
#'
#' @export
#' @rdname git_commit
#' @name git_commit
#' @family git
#' @inheritParams git_open
#' @param message a commit message
#' @param author A [git_signature] value, default is [git_signature_default()].
#' @param committer A [git_signature] value, default is same as `author`
#' @return
#' * `git_status()`, `git_ls()`: A data frame with one row per file
#' * `git_log()`: A data frame with one row per commit
#' * `git_commit()`, `git_commit_all()`: A SHA
#' @useDynLib gert R_git_commit_create
#' @examples
#' oldwd <- getwd()
#' repo <- file.path(tempdir(), "myrepo")
#' git_init(repo)
#' setwd(repo)
#'
#' # Set a user if no default
#' if(!user_is_configured()){
#'   git_config_set("user.name", "Jerry")
#'   git_config_set("user.email", "jerry@gmail.com")
#' }
#'
#' writeLines(letters[1:6], "alphabet.txt")
#' git_status()
#'
#' git_add("alphabet.txt")
#' git_status()
#'
#' git_commit("Start alphabet file")
#' git_status()
#'
#' git_ls()
#'
#' git_log()
#'
#' cat(letters[7:9], file = "alphabet.txt", sep = "\n", append = TRUE)
#' git_status()
#'
#' git_commit_all("Add more letters")
#'
#' # cleanup
#' setwd(oldwd)
#' unlink(repo, recursive = TRUE)
git_commit <- function(message, author = NULL, committer = NULL, repo = '.'){
  repo <- git_open(repo)
  if(!length(author)) {
    author <- git_signature_default(repo = repo)
  } else {
    git_signature_parse(author) #validate
  }
  if(!length(committer)){
    committer <- author
  } else {
    git_signature_parse(committer)  #validate
  }
  stopifnot(is.character(message), length(message) == 1)
  status <- git_status(repo = repo)
  if(!any(status$staged))
    stop("No staged files to commit. Run git_add() to select files.")
  merge_parents <- git_merge_parent_heads(repo = repo)
  .Call(R_git_commit_create, repo, message, author, committer, merge_parents)
}

#' @export
#' @rdname git_commit
git_commit_all <- function(message, author = NULL, committer = NULL, repo = '.'){
  repo <- git_open(repo)
  unstaged <- git_status(staged = FALSE, repo = repo)

  changes <- unstaged$file[unstaged$status %in% c("modified", "renamed", "typechange")]
  if(length(changes))
    git_add(changes, repo = repo)

  deleted <- unstaged$file[unstaged$status == "deleted"]
  if(length(deleted))
    git_rm(deleted, repo = repo)

  git_commit(message = message, author = author, committer = committer, repo = repo)
}

#' @export
#' @rdname git_commit
#' @useDynLib gert R_git_commit_info
git_commit_info <- function(ref = "HEAD", repo = '.'){
  repo <- git_open(repo)
  .Call(R_git_commit_info, repo, ref)
}

#' @export
#' @rdname git_commit
#' @useDynLib gert R_git_commit_id
git_commit_id <- function(ref = "HEAD", repo = '.'){
  repo <- git_open(repo)
  .Call(R_git_commit_id, repo, ref)
}

#' @export
#' @rdname git_commit
#' @useDynLib gert R_git_commit_stats
git_commit_stats <- function(ref = "HEAD", repo = '.'){
  repo <- git_open(repo)
  .Call(R_git_commit_stats, repo, ref)
}

#' @export
#' @rdname git_commit
#' @param ancestor a reference to a potential ancestor commit
#' @useDynLib gert R_git_commit_descendant
git_commit_descendant_of <- function(ancestor, ref = 'HEAD', repo = '.'){
  repo <- git_open(repo)
  .Call(R_git_commit_descendant, repo, ref, ancestor)
}

#' @export
#' @rdname git_commit
#' @param files vector of paths relative to the git root directory.
#' Use `"."` to stage all changed files.
#' @param force add files even if in gitignore
#' @useDynLib gert R_git_repository_add
git_add <- function(files, force = FALSE, repo = '.'){
  repo <- git_open(repo)
  info <- git_info(repo)
  normalizePath(file.path(info$path, files), mustWork = FALSE)
  force <- as.logical(force)
  .Call(R_git_repository_add, repo, files, force)
  git_status(repo = repo)
}

#' @export
#' @rdname git_commit
#' @useDynLib gert R_git_repository_rm
git_rm <- function(files, repo = '.'){
  repo <- git_open(repo)
  info <- git_info(repo)
  normalizePath(file.path(info$path, files), mustWork = FALSE)
  .Call(R_git_repository_rm, repo, files)
  git_status(repo = repo)
}

#' @export
#' @rdname git_commit
#' @useDynLib gert R_git_status_list
#' @param staged return only staged (TRUE) or unstaged files (FALSE).
#' Use `NULL` or `NA` to show both (default).
#' @param pathspec character vector with paths to match
git_status <- function(staged = NULL, pathspec = NULL, repo = '.'){
  repo <- git_open(repo)
  staged <- as.logical(staged)
  pathspec <- as.character(pathspec)
  df <- .Call(R_git_status_list, repo, staged, pathspec)
  df[order(df$file), ,drop = FALSE]
}

#' @export
#' @rdname git_commit
#' @useDynLib gert R_git_conflict_list
git_conflicts <- function(repo = '.'){
  repo <- git_open(repo)
  .Call(R_git_conflict_list, repo)
}

#' @export
#' @rdname git_commit
#' @useDynLib gert R_git_repository_ls
git_ls <- function(repo = '.', ref = NULL){
  repo <- git_open(repo)
  if(!length(ref) && isTRUE(git_info(repo = repo)$bare)){
    ref <- 'HEAD'
  }
  .Call(R_git_repository_ls, repo, ref = ref)
}

#' @export
#' @rdname git_commit
#' @useDynLib gert R_git_commit_log
#' @param ref revision string with a branch/tag/commit value
#' @param max lookup at most latest n parent commits
#' @param after date or timestamp: only include commits starting this date
git_log <- function(ref = "HEAD", max = 100, after = NULL, repo = "."){
  repo <- git_open(repo)
  ref <- as.character(ref)
  max <- as.integer(max)
  if(length(after))
    after <- as.POSIXct(after)
  .Call(R_git_commit_log, repo, ref, max, after)
}

#' @export
#' @rdname git_commit
#' @useDynLib gert R_git_stat_files
git_stat_files <- function(files, ref = "HEAD", repo = '.'){
  repo <- git_open(repo)
  files <- as.character(files)
  .Call(R_git_stat_files, repo, files, ref)
}

assert_string <- function(x){
  if(!is.character(x) || !length(x))
    stop("Argument must be a string of length 1")
}

#' GitHub Wrappers
#'
#' Fetch and checkout pull requests.
#'
#' By default `git_fetch_pull_requests` will download all PR branches. To
#' remove these again simply use `git_fetch(prune = TRUE)`.
#'
#' @export
#' @rdname github
#' @inheritParams git_fetch
#' @param pr number with PR to fetch or check out. Use `"*"` to fetch all
#' pull requests.
git_checkout_pull_request <- function(pr = 1, remote = NULL, repo = '.'){
  pr <- as.character(pr)
  if(!length(remote))
    remote <- git_info(repo)$remote
  local_branch <- sprintf("pr/%s", pr)
  remote_branch <- sprintf("%s/pr/%s", remote, pr)
  git_fetch_pull_requests(pr = pr, remote = remote, repo = repo)
  if(git_branch_exists(local_branch)){
    inform("Continuing on existing pr branch: %s", local_branch)
    git_branch_checkout(local_branch, repo = repo)
    git_pull(repo = repo)
  } else {
    inform("Creating new branch: %s", local_branch)
    git_branch_create(local_branch, remote_branch, checkout = TRUE, repo = repo)
  }
}

#' @export
#' @rdname github
git_fetch_pull_requests <- function(pr = '*', remote = NULL, repo = '.'){
  pr <- as.character(pr)
  if(!length(remote))
    remote <- git_info(repo)$remote
  refspec <- sprintf('+refs/pull/%s/head:refs/remotes/%s/pr/%s', pr, remote, pr)
  git_fetch(remote = remote, refspec = refspec, repo = repo)
  invisible(refspec)
}

#' Git Branch
#'
#' Create, list, and checkout branches.
#'
#' @export
#' @rdname git_branch
#' @name git_branch
#' @family git
#' @inheritParams git_open
#' @useDynLib gert R_git_branch_current
#' @git branch
git_branch <- function(repo = '.') {
  repo <- git_open(repo)
  .Call(R_git_branch_current, repo)
}

#' @export
#' @rdname git_branch
#' @useDynLib gert R_git_branch_list
git_branch_list <- function(local = NULL, repo = '.') {
  repo <- git_open(repo)
  local <- as.logical(local)
  .Call(R_git_branch_list, repo, local)
}

#' @export
#' @rdname git_branch
#' @param branch name of branch to check out
#' @param force ignore conflicts and overwrite modified files
#' @param orphan if branch does not exist, checkout unborn branch
#' @useDynLib gert R_git_checkout_branch R_git_checkout_unborn
#' @git checkout
#' @param force overwrite existing branch
#' @examplesIf interactive()
#' # Creating a branch
#' repo <- file.path(tempdir(), "myrepo")
#' git_init(repo)
#' writeLines("hello", file.path(repo, "hello.txt"))
#' git_add("hello.txt", repo = repo)
#' first_commit <- git_commit("First commit", repo = repo)
#'
#' git_branch_create("new-feat", repo = repo)
#' writeLines("world", file.path(repo, "hello.txt"))
#' git_add("hello.txt", repo = repo)
#' second_commit <- git_commit("Second commit", repo = repo)
#'
#' # Changing branches
#' git_branch(repo = repo)
#' git_branch_switch("main", repo = repo)
#' git_branch(repo = repo)
#'
#' git_branch_exists("new-feat", repo = repo)
#'
#' # Listing branches
#' git_branch_list(repo = repo)
#'
#' # Renaming a branch
#' git_branch_move("new-feat", "better-name", repo = repo)
#' git_branch_exists("new-feat", repo = repo)
#' git_branch_list(repo = repo)
#'
#' # Deleting a branch
#' git_branch_delete("better-name", repo = repo)
#' git_branch_exists("better-name", repo = repo)
#'
#' # clean up
#' unlink(repo, recursive = TRUE)
#' # ------------------------
#' # Creating an orphan branch
#' repo <- file.path(tempdir(), "myrepo")
#' git_init(repo)
#'
#' # Set a user if no default
#' if (!user_is_configured()) {
#'   git_config_set("user.name", "Jerry")
#'   git_config_set("user.email", "jerry@gmail.com")
#' }
#' writeLines("hello", file.path(repo, "hello.txt"))
#' git_add("hello.txt", repo = repo)
#' first_commit <- git_commit("First commit", repo = repo)
#'
#' writeLines("world", file.path(repo, "hello.txt"))
#' git_add("hello.txt", repo = repo)
#' second_commit <- git_commit("Second commit", repo = repo)
#'
#' # Check out first commit
#' git_branch_checkout(first_commit, orphan = TRUE, repo = repo)
#' git_status(repo)
#' # clean up
#' unlink(repo, recursive = TRUE)
#'
git_branch_checkout <- function(
  branch,
  force = FALSE,
  orphan = FALSE,
  repo = '.'
) {
  repo <- git_open(repo)
  branch <- as.character(branch)
  force <- as.logical(force)
  if (!git_branch_exists(branch, repo = repo)) {
    if (isTRUE(orphan)) {
      ref <- paste0('refs/heads/', branch)
      .Call(R_git_checkout_unborn, repo, ref)
      return(ref)
    }
    all_branches <- subset(git_branch_list(repo = repo), local == FALSE)$name
    candidate <- sub("^[^/]+/", "", all_branches) == branch
    if (sum(candidate) > 1) {
      stop(sprintf(
        "Local branch '%s' does not exist and multiple remote candidates found.",
        branch
      ))
    } else if (sum(candidate) == 0) {
      stop(sprintf("No local or remote branch '%s' found.", branch))
    } else {
      remote_branch <- unname(all_branches[candidate])
      inform("Creating local branch %s from %s", branch, remote_branch)
      git_branch_create(branch, remote_branch, checkout = FALSE, repo = repo)
    }
  }
  .Call(R_git_checkout_branch, repo, branch, force)
}

#' @rdname git_branch
#' @export
git_branch_switch <- git_branch_checkout

#' @export
#' @rdname git_branch
#' @useDynLib gert R_git_create_branch
#' @param ref string with a branch/tag/commit
#' @param checkout move HEAD to the newly created branch
#' @param force overwrite existing branch
git_branch_create <- function(
  branch,
  ref = "HEAD",
  checkout = TRUE,
  force = FALSE,
  repo = '.'
) {
  repo <- git_open(repo)
  branch <- as.character(branch)
  ref <- as.character(ref)
  checkout <- as.logical(checkout)
  force <- as.logical(force)
  invisible(.Call(R_git_create_branch, repo, branch, ref, checkout, force))
}

#' @export
#' @rdname git_branch
#' @useDynLib gert R_git_delete_branch
git_branch_delete <- function(branch, repo = '.') {
  repo <- git_open(repo)
  branch <- as.character(branch)
  .Call(R_git_delete_branch, repo, branch)
  git_repo_path(repo)
}

#' @export
#' @rdname git_branch
#' @useDynLib gert R_git_branch_move
#' @param new_branch target name of the branch once the move is performed; this name is validated for consistency.
git_branch_move <- function(branch, new_branch, force = FALSE, repo = '.') {
  repo <- git_open(repo)
  branch <- as.character(branch)
  new_branch <- as.character(new_branch)
  .Call(R_git_branch_move, repo, branch, new_branch, as.logical(force))
  git_repo_path(repo)
}

#' @export
#' @rdname git_branch
git_branch_fast_forward <- function(ref, repo = '.') {
  analysis <- git_merge_analysis(ref = ref, repo = repo)
  if (analysis != "fastforward") {
    stop("Branch cannot be fast-forwarded. Use git_merge() instead")
  }
  git_branch_set_target(ref = ref, repo = repo)
}

#' @export
#' @rdname git_branch
#' @param upstream remote branch from [git_branch_list], for example `"origin/master"`
#' @useDynLib gert R_git_branch_set_upstream
git_branch_set_upstream <- function(
  upstream,
  branch = git_branch(repo),
  repo = '.'
) {
  repo <- git_open(repo)
  stopifnot(is.character(upstream))
  if (!git_branch_exists(upstream, local = FALSE, repo = repo)) {
    stop(sprintf("No remote branch found: %s, maybe fetch first?", upstream))
  }
  .Call(R_git_branch_set_upstream, repo, upstream, branch)
  git_repo_path(repo)
}

#' @export
#' @rdname git_branch
#' @param local set TRUE to only check for local branches, FALSE to check for remote
#' branches. Use NULL to return all branches.
#' @useDynLib gert R_git_branch_exists
git_branch_exists <- function(branch, local = TRUE, repo = '.') {
  repo <- git_open(repo)
  branch <- as.character(branch)
  local <- as.logical(local)
  .Call(R_git_branch_exists, repo, branch, local)
}

#' @useDynLib gert R_git_branch_set_target
git_branch_set_target <- function(ref, branch, repo) {
  repo <- git_open(repo)
  ref <- as.character(ref)
  .Call(R_git_branch_set_target, repo, ref)
  git_repo_path(repo)
}

#' Git Ignore
#'
#' Test if files would be ignored by `.gitignore` rules
#'
#' @export
#' @rdname git_ignore
#' @name git_ignore
#' @family git
#' @inheritParams git_open
#' @param path A character vector of paths to test within the repo
#' @return A logical vector the same length as `path`, indicating if the
#' paths would be ignored.
#' @useDynLib gert R_git_ignore_path_is_ignored
git_ignore_path_is_ignored <- function(path, repo = '.') {
  repo <- git_open(repo)
  path <- as.character(path)
  .Call(R_git_ignore_path_is_ignored, repo, path)
}

#' Cherry-Pick and Rebase
#'
#' A cherry-pick applies the changes from a given commit (from another branch)
#' onto the current branch. A rebase resets the branch to the state of another
#' branch (upstream) and then re-applies your local changes by cherry-picking
#' each of your local commits onto the upstream commit history.
#'
#' `git_rebase_list` shows your local commits that are missing from the `upstream`
#' history, and if they conflict with upstream changes. It does so by performing
#' a rebase dry-run, without committing anything. If there are no conflicts, you
#' can use `git_rebase_commit` to rewind and rebase your branch onto `upstream`.
#' Gert only support a clean rebase; it never leaves the repository in unfinished
#' "rebasing" state. If conflicts arise, `git_rebase_commit` will raise an error
#' without making changes.
#'
#' @export
#' @rdname git_rebase
#' @name git_rebase
#' @family git
#' @param upstream branch to which you want to rewind and re-apply your
#' local commits. The default uses the remote upstream branch with the
#' current state on the git server, simulating [git_pull].
#' @inheritParams git_open
#' @inheritParams git_branch
git_rebase_list <- function(upstream = NULL, repo = '.'){
  git_rebase(upstream = upstream, commit_changes = FALSE, repo = repo)
}

#' @export
#' @rdname git_rebase
git_rebase_commit <- function(upstream = NULL, repo = '.'){
  git_rebase(upstream = upstream, commit_changes = TRUE, repo = repo)
}

#' @useDynLib gert R_git_rebase
git_rebase <- function(upstream, commit_changes, repo){
  repo <- git_open(repo)
  info <- git_info(repo = repo)
  if(!length(upstream)){
    if(!length(info$upstream) || is.na(info$upstream) || !nchar(info$upstream))
      stop("No upstream configured for current HEAD")
    git_fetch(info$remote, repo = repo)
    upstream <- info$upstream
  }
  df <- .Call(R_git_rebase, repo, upstream, commit_changes)
  if(commit_changes){
    new_head <- ifelse(nrow(df) > 0, utils::tail(df$commit, 1), upstream[1])
    git_branch_set_target(ref = new_head, repo = repo)
    inform("Resetting %s to %s", info$shorthand, new_head)
  }
  return(df)
}

#' @export
#' @rdname git_rebase
git_reset_hard <- function(ref = "HEAD", repo = "."){
  git_reset("hard", ref = ref, repo = repo)
}

#' @export
#' @rdname git_rebase
git_reset_soft <- function(ref = "HEAD", repo = "."){
  git_reset("soft", ref = ref, repo = repo)
}

#' @export
#' @rdname git_rebase
git_reset_mixed <- function(ref = "HEAD", repo = "."){
  git_reset("mixed", ref = ref, repo = repo)
}

#' @useDynLib gert R_git_reset
git_reset <- function(type = c("soft", "hard", "mixed"), ref = "HEAD", repo = "."){
  typenum <- switch(match.arg(type), soft = 1L, mixed = 2L, hard = 3L)
  repo <- git_open(repo)
  ref <- as.character(ref)
  .Call(R_git_reset, repo, ref, typenum)
  git_status(repo = repo)
}

#' @export
#' @rdname git_rebase
#' @useDynLib gert R_git_cherry_pick
#' @param commit id of the commit to cherry pick
git_cherry_pick <- function(commit, repo = '.'){
  repo <- git_open(repo)
  assert_string(commit)
  .Call(R_git_cherry_pick, repo, commit)
}

#' @export
#' @rdname git_rebase
#' @useDynLib gert R_git_ahead_behind
git_ahead_behind <- function(upstream = NULL, ref = 'HEAD', repo = '.'){
  repo <- git_open(repo)
  if(!length(upstream))
    upstream <- git_info(repo = repo)$upstream
  if(!length(upstream))
    stop("No upstream set or specified")
  .Call(R_git_ahead_behind, repo, ref, upstream)
}

# This is only used for static builds and on Solaris.
# It is not needed if you're using a system libgit2.
#' @useDynLib gert R_static_libgit2
have_static_libgit2 <- function(){
  .Call(R_static_libgit2)
}

#' @useDynLib gert R_set_cert_locations
set_cert_locations <- function(file, path){
  file <- as.character(file)
  path <- as.character(path)
  .Call(R_set_cert_locations, file, path)
}

file_exists <- function(x){
  length(x) && file.exists(x)
}

find_cert_dir <- function(){
  if(nchar(Sys.which('openssl')) > 0){
    out <- sys::exec_internal('openssl', c('version', '-d'), error = FALSE)
    if(out$status == 0){
      txt <- sys::as_text(out$stdout)
      path <- utils::tail(strsplit(txt, ' ', fixed = TRUE)[[1]], 1)
      path <- gsub('"', "", path, fixed = TRUE)
      path <- file.path(path, 'certs')
      if(file_exists(path)){
        return(path)
      }
    }
  }
  default_paths <- c('/etc/pki/tls/certs', '/etc/ssl/certs/', '/etc/opt/csw/ssl/certs')
  for(x in default_paths){
    if(file_exists(x)){
      return(x)
    }
  }
  return(NULL)
}

if (
  !file.exists('clone.o') && !file.exists("../.deps/libgit2/include/git2.h")
) {
  unlink("../.deps", recursive = TRUE)
  url <- if (grepl("aarch", R.version$platform)) {
    "https://github.com/r-windows/bundles/releases/download/libgit2-1.9.1/libgit2-1.9.1-clang-aarch64.tar.xz"
  } else if (grepl("clang", Sys.getenv('R_COMPILED_BY'))) {
    "https://github.com/r-windows/bundles/releases/download/libgit2-1.9.1/libgit2-1.9.1-clang-x86_64.tar.xz"
  } else if (getRversion() >= "4.2") {
    "https://github.com/r-windows/bundles/releases/download/libgit2-1.9.1/libgit2-1.9.1-ucrt-x86_64.tar.xz"
  } else {
    "https://github.com/rwinlib/libgit2/archive/v1.7.1.tar.gz"
  }
  download.file(url, basename(url), quiet = TRUE)
  dir.create("../.deps", showWarnings = FALSE)
  untar(basename(url), exdir = "../.deps", tar = 'internal')
  unlink(basename(url))
  setwd("../.deps")
  file.rename(list.files(), 'libgit2')
}

.onAttach <- function(libname, pkgname){
  config <- libgit2_config()
  ssh <- ifelse(config$ssh, "YES", "NO")
  https <- ifelse(config$https, "YES", "NO")
  packageStartupInform(
    "Linking to libgit2 v%s, ssh support: %s",
    as.character(config$version), ssh)
  user_config <- config$config.global
  if(length(user_config) && nchar(user_config)){
    if(is_windows()){
      user_config <- normalizePath(user_config, mustWork = FALSE)
    }
    packageStartupInform("Global config: %s", user_config)
  } else {
    packageStartupInform("No global .gitconfig found in: %s", config$config.home)
  }
  if(length(config$config.system) && nchar(config$config.system))
    packageStartupInform(paste0("System config: ", config$config.system))
  try({
    settings <- git_config_global()
    name <- subset(settings, name == 'user.name')$value
    email <- subset(settings, name == 'user.email')$value
    if(length(name) || length(email)){
      packageStartupInform("Default user: %s <%s>", as_string(name), as_string(email))
    } else {
      packageStartupInform("No default user configured")
    }
  })
}

.onLoad <- function(libname, pkgname) {
  # This should usually correspond to 'openssl version -d'
  if(isTRUE(have_static_libgit2())){
    certpath <- find_cert_dir()
    if(length(certpath)){
      cafile <- file.path(dirname(certpath), 'cert.pem')
      if(!file.exists(cafile))
        cafile <- NULL
      set_cert_locations(cafile, certpath)
    } else {
      warning("Unable to find directory with certificates", immediate. = TRUE)
    }
  }

  # Load tibble (if available) for pretty printing
  if(interactive() && is.null(.getNamespace('tibble'))){
    tryCatch({
      getNamespace('tibble')
    }, error = function(e){})
  }

  invisible()
}

as_string <- function(x){
  ifelse(length(x) > 0, x[1], NA_character_)
}

is_windows <- function(){
  identical(.Platform$OS.type, "windows")
}

is_solaris <- function(){
  isTRUE(grepl('solaris', R.version$platform, ignore.case = TRUE))
}

inform_impl <- function(message, subclass = NULL) {
  cnd <- structure(
    list(message = paste0(message, "\n")),
    class = c(subclass, "simpleMessage", "message", "condition")
  )
  withRestarts(
    muffleMessage = function() NULL,
    {
      signalCondition(cnd)
      cat(
        conditionMessage(cnd),
        file = if (interactive()) stdout() else stderr()
      )
    }
  )
}


inform <- function(...) inform_impl(sprintf(...), subclass = "gertMessage")

packageStartupInform <- function(...) {
  inform_impl(sprintf(...), subclass = c("gertMessage", "packageStartupMessage"))
}

message <- function(...) {
  stop("Internal error: use inform() instead of message()")
}

packageStartupMessage <- function(...) {
  stop("Internal error: use packageStartupInform() instead of packageStartupMessage()")
}

# libgit2 wants all paths utf-8
normalizePath <- function(...){
  enc2utf8(base::normalizePath(...))
}

#' Open local repository
#'
#' Returns a pointer to a libgit2 repository object. This function is mainly
#' for internal use; users should simply reference a repository in gert by
#' by the path to the directory.
#'
#' @export
#' @param repo The path to the git repository. If the directory is not a
#' repository, parent directories are considered (see [git_find()]). To disable
#' this search, provide the filepath protected with [I()]. When using this
#' parameter, always explicitly call by name (i.e. `repo = `) because future
#' versions of gert may have additional parameters.
#' @return an pointer to the libgit2 repository
#' @useDynLib gert R_git_repository_open
#' @examples
#' r <- tempfile(pattern = "gert")
#' git_init(r)
#' r_ptr <- git_open(r)
#' r_ptr
#' git_open(r_ptr)
#' git_info(r)
#'
#' # cleanup
#' unlink(r, recursive = TRUE)
git_open <- function(repo = '.') {
  if (inherits(repo, 'git_repo_ptr')) {
    return(repo)
  } else if (!is.character(repo)) {
    stop("repo argument must be a filepath or an existing repository object")
  }
  search <- !inherits(repo, 'AsIs')
  path <- normalizePath(path.expand(repo), mustWork = FALSE)
  out <- git_repository_open(path, search)
  if (is_rstudio_ide()) {
    cl <- substitute(rstudioapi::executeCommand("vcsRefresh"))
    do.call(on.exit, list(cl, add = TRUE), envir = parent.frame())
  }
  return(out)
}


git_repository_open <- local({
  cache <- new.env(parent = emptyenv())
  function(path, search) {
    key <- digest(list(path, search))
    if (
      !exists(key, envir = cache) || !isTRUE(getOption('gert.use.repo.cache'))
    ) {
      val <- .Call(R_git_repository_open, path, search)
      assign(key, val, envir = cache)
    }
    get0(key, cache)
  }
})

is_rstudio_ide <- function() {
  interactive() &&
    identical(Sys.getenv('RSTUDIO'), '1') &&
    !nchar(Sys.getenv('Disable_vcsRefresh')) &&
    rstudioapi::isAvailable()
}

#' @export
print.git_repo_ptr <- function(x, ...) {
  info <- git_info(x)

  type = "git repository"
  if (info$bare) {
    type = paste(type, "(bare)")
  }

  cat(sprintf(
    "<%s>: %s[@%s]\n",
    type,
    normalizePath(info$path),
    info$shorthand
  ))
}

#' @useDynLib gert R_git_repository_path
git_repo_path <- function(repo) {
  invisible(.Call(R_git_repository_path, repo))
}

digest <- function(x) {
  as.character(openssl::sha1(serialize(x, NULL)))
}

# nocov start

git_links <- function() {
  base_url <- "https://libgit2.org/docs/reference/main/%s/index.html"
  dat <- as.data.frame(
    rbind(
      c("branch", sprintf(base_url, "branch")),
      c("cherrypick", sprintf(base_url, "cherrypick")),
      c("checkout", sprintf(base_url, "checkout")),
      c("clone", sprintf(base_url, "clone")),
      c("commit", sprintf(base_url, "commit")),
      c("config", sprintf(base_url, "config")),
      c("diff", sprintf(base_url, "diff")),
      c("graph", sprintf(base_url, "graph")),
      c("ignore", sprintf(base_url, "ignore")),
      c("index", sprintf(base_url, "index")),
      c("merge", sprintf(base_url, "merge")),
      c("rebase", sprintf(base_url, "rebase")),
      c("remote", sprintf(base_url, "remote")),
      c("repository", sprintf(base_url, "repository")),
      c("reset", sprintf(base_url, "reset")),
      c("revert", sprintf(base_url, "revert")),
      c("signature", sprintf(base_url, "signature")),
      c("stash", sprintf(base_url, "stash")),
      c("status", sprintf(base_url, "status")),
      c("submodule", sprintf(base_url, "submodule")),
      c("tag", sprintf(base_url, "tag")),
      c("worktree", sprintf(base_url, "worktree"))
    ),
    stringsAsFactors = FALSE
  )
  names(dat) <- c("command", "url")
  dat
}

#' @exportS3Method roxygen2::roxy_tag_parse
roxy_tag_parse.roxy_tag_git <- function(x) {
  roxygen2::tag_words(x)
}

#' @exportS3Method roxygen2::roxy_tag_rd
roxy_tag_rd.roxy_tag_git <- function(x, base_path, env) {
  roxygen2::rd_section("gitcommands", x$val)
}

#' @export
format.rd_section_gitcommands <- function(x, ...) {
  paste0(
    "\\section{Related libgit2 documentation}{",
    present_git_link(x[["value"]]),
    ".}\n"
  )
}

present_git_link <- function(value) {
  links <- git_links()
  format_git_single_link <- function(x) {
    df <- links[links$command == x, ]
    if (nrow(df) == 0) {
      warning(sprintf("Can't find libgit2 entry for %s!", x))
    }
    sprintf("\\href{%s}{\\code{%s}}", df$url, df$command)
  }

  strings <- vapply(unique(value), format_git_single_link, character(1))

  paste(strings, collapse = ", ")
}
# nocov end

# Classed error handling
# Todo: kclass is ignored right now, would this be useful to report?
raise_libgit2_error <- function(code, message, where = "", kclass = 0){
  cl <- call("::", substitute(libgit2), as.name(where))
  e <- structure(
    class = c(libgit2_error_name(code), "libgit2_error", "error", "condition"),
    list(message = message, call = cl) #call must be an R expression
  )
  stop(e)
}

libgit2_error_name <- function(x){
  out <- which(libgit2_error_codes == x)
  if(length(out))
    return(names(out))
  return("UNKNOWN_ERROR_CODE")
}

# Eror codes copied from https://github.com/libgit2/libgit2/blame/master/include/git2/errors.h
# We don't do this in C because the list of error codes changes from version to version
libgit2_error_codes <- c(
  GIT_OK         =  0,
  GIT_ERROR      = -1,
  GIT_ENOTFOUND  = -3,
  GIT_EEXISTS    = -4,
  GIT_EAMBIGUOUS = -5,
  GIT_EBUFS      = -6,
  GIT_EUSER      = -7,
  GIT_EBAREREPO       =  -8,
  GIT_EUNBORNBRANCH   =  -9,
  GIT_EUNMERGED       = -10,
  GIT_ENONFASTFORWARD = -11,
  GIT_EINVALIDSPEC    = -12,
  GIT_ECONFLICT       = -13,
  GIT_ELOCKED         = -14,
  GIT_EMODIFIED       = -15,
  GIT_EAUTH           = -16,
  GIT_ECERTIFICATE    = -17,
  GIT_EAPPLIED        = -18,
  GIT_EPEEL           = -19,
  GIT_EEOF            = -20,
  GIT_EINVALID        = -21,
  GIT_EUNCOMMITTED    = -22,
  GIT_EDIRECTORY      = -23,
  GIT_EMERGECONFLICT  = -24,
  GIT_PASSTHROUGH     = -30,
  GIT_ITEROVER        = -31,
  GIT_RETRY           = -32,
  GIT_EMISMATCH       = -33,
  GIT_EINDEXDIRTY     = -34,
  GIT_EAPPLYFAIL      = -35
)

#' Git Archive
#'
#' Exports the files in your repository to a zip file that
#' is returned by the function.
#'
#' @export
#' @rdname git_archive
#' @name git_archive
#' @family git
#' @inheritParams git_open
#' @param file name of the output zip file. Default is returned
#' by the function
#' @return path to the zip file that was created
git_archive_zip <- function(file = NULL, repo = ".") {
  repo <- git_open(repo = repo)
  tmp <- tempfile(fileext = '.zip')
  git_archive_internal(tmp, repo = repo)
  if (!length(file)) {
    file <- paste0(basename(git_info(repo)$path), ".zip")
  }
  file.copy(tmp, file, overwrite = TRUE)
  return(file)
}

git_archive_internal <- function(outfile, repo) {
  tryCatch(
    {
      git_stash_save(repo = repo)
      on.exit(git_stash_pop(repo = repo))
    },
    GIT_ENOTFOUND = function(e) {}
  )
  files <- git_ls(repo = repo)$path
  wd <- getwd()
  on.exit(setwd(wd), add = TRUE)
  setwd(git_info(repo = repo)$path)
  zip::zip(outfile, files = files, recurse = FALSE)
}

#' Stashing changes
#'
#' Temporary stash away changed from the working directory.
#'
#' @export
#' @rdname git_stash
#' @name git_stash
#' @family git
#' @inheritParams git_open
#' @useDynLib gert R_git_stash_save
#' @git stash
#' @param message optional message to store the stash
#' @param keep_index changes already added to the index are left intact in
#' the working directory
#' @param include_untracked untracked files are also stashed and then
#' cleaned up from the working directory
#' @param include_ignored ignored files are also stashed and then cleaned
#' up from the working directory
git_stash_save <- function(
  message = "",
  keep_index = FALSE,
  include_untracked = FALSE,
  include_ignored = FALSE,
  repo = "."
) {
  repo <- git_open(repo)
  keep_index <- as.logical(keep_index)
  include_untracked <- as.logical(include_untracked)
  include_ignored <- as.logical(include_ignored)
  .Call(
    R_git_stash_save,
    repo,
    message,
    keep_index,
    include_untracked,
    include_ignored
  )
}

#' @export
#' @rdname git_stash
#' @useDynLib gert R_git_stash_pop
#' @param index The position within the stash list. 0 points to the
#' most recent stashed state.
git_stash_pop <- function(index = 0, repo = ".") {
  repo <- git_open(repo)
  .Call(R_git_stash_pop, repo, index)
}

#' @export
#' @rdname git_stash
#' @useDynLib gert R_git_stash_drop
git_stash_drop <- function(index = 0, repo = ".") {
  repo <- git_open(repo)
  .Call(R_git_stash_drop, repo, index)
}

#' @export
#' @rdname git_stash
#' @useDynLib gert R_git_stash_list
git_stash_list <- function(repo = ".") {
  repo <- git_open(repo)
  .Call(R_git_stash_list, repo)
}

# Classed error handling
# Todo: kclass is ignored right now, would this be useful to report?
raise_libgit2_error <- function(code, message, where = "", kclass = 0) {
  cl <- call("::", substitute(libgit2), as.name(where))
  e <- structure(
    class = c(libgit2_error_name(code), "libgit2_error", "error", "condition"),
    list(message = message, call = cl) #call must be an R expression
  )
  stop(e)
}

libgit2_error_name <- function(x) {
  out <- which(libgit2_error_codes == x)
  if (length(out)) return(names(out))
  return("UNKNOWN_ERROR_CODE")
}

# Eror codes copied from https://github.com/libgit2/libgit2/blame/master/include/git2/errors.h
# We don't do this in C because the list of error codes changes from version to version
libgit2_error_codes <- c(
  GIT_OK = 0,
  GIT_ERROR = -1,
  GIT_ENOTFOUND = -3,
  GIT_EEXISTS = -4,
  GIT_EAMBIGUOUS = -5,
  GIT_EBUFS = -6,
  GIT_EUSER = -7,
  GIT_EBAREREPO = -8,
  GIT_EUNBORNBRANCH = -9,
  GIT_EUNMERGED = -10,
  GIT_ENONFASTFORWARD = -11,
  GIT_EINVALIDSPEC = -12,
  GIT_ECONFLICT = -13,
  GIT_ELOCKED = -14,
  GIT_EMODIFIED = -15,
  GIT_EAUTH = -16,
  GIT_ECERTIFICATE = -17,
  GIT_EAPPLIED = -18,
  GIT_EPEEL = -19,
  GIT_EEOF = -20,
  GIT_EINVALID = -21,
  GIT_EUNCOMMITTED = -22,
  GIT_EDIRECTORY = -23,
  GIT_EMERGECONFLICT = -24,
  GIT_PASSTHROUGH = -30,
  GIT_ITEROVER = -31,
  GIT_RETRY = -32,
  GIT_EMISMATCH = -33,
  GIT_EINDEXDIRTY = -34,
  GIT_EAPPLYFAIL = -35
)

#' Stage and commit changes
#'
#' @description
#' To commit changes, start by *staging* the files to be included in the commit
#' using `git_add()` or `git_rm()`. Use `git_status()` to see an overview of
#' staged and unstaged changes, and finally `git_commit()` creates a new commit
#' with currently staged files.
#'
#' `git_commit_all()` is a convenience function that automatically stages and
#' commits all modified files. Note that `git_commit_all()` does **not** add
#' new, untracked files to the repository. You need to make an explicit call to
#' `git_add()` to start tracking new files.
#'
#' @export
#' @family git
#' @inheritParams git_open
#' @param message a commit message
#' @param ref revision string with a branch/tag/commit value
#' @param author A [git_signature] value, default is [git_signature_default()].
#' @param committer A [git_signature] value, default is same as `author`
#' @return
#' * `git_status()`, `git_ls()`: A data frame with one row per file
#' * `git_commit()`, `git_commit_all()`: A SHA
#' @useDynLib gert R_git_commit_create
#' @git commit index status
#' @examples
#' oldwd <- getwd()
#' repo <- file.path(tempdir(), "myrepo")
#' git_init(repo)
#' setwd(repo)
#'
#' # Set a user if no default
#' if(!user_is_configured()){
#'   git_config_set("user.name", "Jerry")
#'   git_config_set("user.email", "jerry@gmail.com")
#' }
#'
#' writeLines(letters[1:6], "alphabet.txt")
#' git_status()
#'
#' git_add("alphabet.txt")
#' git_status()
#'
#' git_commit("Start alphabet file")
#' git_status()
#'
#' git_ls()
#'
#' git_log()
#'
#' cat(letters[7:9], file = "alphabet.txt", sep = "\n", append = TRUE)
#' git_status()
#'
#' git_commit_all("Add more letters")
#'
#' # cleanup
#' setwd(oldwd)
#' unlink(repo, recursive = TRUE)
git_commit <- function(message, author = NULL, committer = NULL, repo = '.') {
  repo <- git_open(repo)
  if (!length(author)) {
    author <- git_signature_default(repo = repo)
  } else {
    git_signature_parse(author) #validate
  }
  if (!length(committer)) {
    committer <- author
  } else {
    git_signature_parse(committer) #validate
  }
  stopifnot(is.character(message), length(message) == 1)
  status <- git_status(repo = repo)
  if (!any(status$staged)) {
    stop("No staged files to commit. Run git_add() to select files.")
  }
  merge_parents <- git_merge_parent_heads(repo = repo)
  .Call(R_git_commit_create, repo, message, author, committer, merge_parents)
}

#' @export
#' @rdname git_commit
git_commit_all <- function(
  message,
  author = NULL,
  committer = NULL,
  repo = '.'
) {
  repo <- git_open(repo)
  unstaged <- git_status(staged = FALSE, repo = repo)

  changes <- unstaged$file[
    unstaged$status %in% c("modified", "renamed", "typechange")
  ]
  if (length(changes)) {
    git_add(changes, repo = repo)
  }

  deleted <- unstaged$file[unstaged$status == "deleted"]
  if (length(deleted)) {
    git_rm(deleted, repo = repo)
  }

  git_commit(
    message = message,
    author = author,
    committer = committer,
    repo = repo
  )
}

#' View commit history
#'
#' @description
#'
#' * `git_commit_stats()` returns information about commit insertion and deletion
#' * `git_commit_info()` a list of commit info
#' * `git_commit_id()` is a shortcut for `git_commit_info()$id`
#' * `git_log()` shows the most recent commits
#' * `git_ls()` lists all the files that are being tracked in the repository.
#' * `git_stat_files()` shows information of when `files` was last modified.
#'
#' @export
#' @inheritParams git_commit
#' @family git
#' @name git_history
#' @returns
#' * `git_commit_info()` and `git_commit_stats()` return a list.
#' @useDynLib gert R_git_commit_info
#' @git commit
git_commit_info <- function(ref = "HEAD", repo = '.') {
  repo <- git_open(repo)
  .Call(R_git_commit_info, repo, ref)
}

#' @export
#' @rdname git_history
#' @useDynLib gert R_git_commit_id
git_commit_id <- function(ref = "HEAD", repo = '.') {
  repo <- git_open(repo)
  .Call(R_git_commit_id, repo, ref)
}

#' @export
#' @rdname git_history
#' @useDynLib gert R_git_commit_stats
git_commit_stats <- function(ref = "HEAD", repo = '.') {
  repo <- git_open(repo)
  .Call(R_git_commit_stats, repo, ref)
}

#' @export
#' @rdname git_merge
#' @param ancestor a reference to a potential ancestor commit
#' @useDynLib gert R_git_commit_descendant
git_commit_descendant_of <- function(ancestor, ref = 'HEAD', repo = '.') {
  repo <- git_open(repo)
  .Call(R_git_commit_descendant, repo, ref, ancestor)
}

#' @export
#' @order 1
#' @rdname git_commit
#' @name git_commit
#' @param files vector of paths relative to the git root directory.
#' Use `"."` to stage all changed files.
#' @param force add files even if in gitignore
#' @useDynLib gert R_git_repository_add
git_add <- function(files, force = FALSE, repo = '.') {
  repo <- git_open(repo)
  info <- git_info(repo)
  normalizePath(file.path(info$path, files), mustWork = FALSE)
  force <- as.logical(force)
  .Call(R_git_repository_add, repo, files, force)
  git_status(repo = repo)
}

#' @export
#' @order 2
#' @rdname git_commit
#' @useDynLib gert R_git_repository_rm
git_rm <- function(files, repo = '.') {
  repo <- git_open(repo)
  info <- git_info(repo)
  normalizePath(file.path(info$path, files), mustWork = FALSE)
  .Call(R_git_repository_rm, repo, files)
  git_status(repo = repo)
}

#' @export
#' @rdname git_commit
#' @useDynLib gert R_git_status_list
#' @param staged return only staged (TRUE) or unstaged files (FALSE).
#' Use `NULL` or `NA` to show both (default).
#' @param pathspec character vector with paths to match
git_status <- function(staged = NULL, pathspec = NULL, repo = '.') {
  repo <- git_open(repo)
  staged <- as.logical(staged)
  pathspec <- as.character(pathspec)
  df <- .Call(R_git_status_list, repo, staged, pathspec)
  df[order(df$file), , drop = FALSE]
}

#' @export
#' @rdname git_merge
#' @useDynLib gert R_git_conflict_list
git_conflicts <- function(repo = '.') {
  repo <- git_open(repo)
  .Call(R_git_conflict_list, repo)
}

#' @export
#' @rdname git_commit
#' @useDynLib gert R_git_repository_ls
git_ls <- function(repo = '.', ref = NULL) {
  repo <- git_open(repo)
  if (!length(ref) && isTRUE(git_info(repo = repo)$bare)) {
    ref <- 'HEAD'
  }
  .Call(R_git_repository_ls, repo, ref = ref)
}

#' @export
#' @rdname git_history
#' @useDynLib gert R_git_commit_log
#' @param max lookup at most latest n parent commits
#' @param after date or timestamp: only include commits starting this date
#' @param path character vector with paths to filter on; only commits that
#' touch these paths are included
git_log <- function(
  ref = "HEAD",
  max = 100,
  after = NULL,
  path = NULL,
  repo = "."
) {
  repo <- git_open(repo)
  ref <- as.character(ref)
  max <- as.integer(max)
  if (length(after)) {
    after <- as.POSIXct(after)
  }
  path <- as.character(path)
  .Call(R_git_commit_log, repo, ref, max, after, path)
}

#' @export
#' @rdname git_history
#' @useDynLib gert R_git_stat_files
git_stat_files <- function(files, ref = "HEAD", max = NULL, repo = '.') {
  repo <- git_open(repo)
  files <- as.character(files)
  max <- as.integer(max)
  .Call(R_git_stat_files, repo, files, ref, max)
}

#' Revert a commit
#'
#' Applies the inverse of the changes introduced by a given commit, equivalent
#' to `git revert <commit>`. The commit must be reachable from the current HEAD.
#'
#' By default, a new revert commit is created immediately. Set `commit = FALSE`
#' to only stage the reverted changes without committing, leaving you free to
#' amend or combine them before calling [git_commit()].
#'
#' @export
#' @name git_revert
#' @rdname git_revert
#' @family git
#' @inheritParams git_open
#' @inheritParams git_history
#' @param commit if `FALSE`, stage the reverted changes without creating a
#'   commit. Default is `TRUE`, that is to say, by default a commit is made.
#' @param ... parameters passed to `git_commit` such as `message` or `author`
#' @return The SHA of the new revert commit (invisibly), or `NULL` when
#'   `commit = FALSE`.
#' @examplesIf interactive()
#' repo <- file.path(tempdir(), "myrepo")
#' git_init(repo)
#'
#' # Set a user if no default
#' if (!user_is_configured()) {
#'   git_config_set("user.name", "Jerry")
#'   git_config_set("user.email", "jerry@gmail.com")
#' }
#'
#' writeLines("hello", file.path(repo, "hello.txt"))
#' git_add("hello.txt", repo = repo)
#' git_commit("First commit", repo = repo)
#'
#' writeLines("world", file.path(repo, "hello.txt"))
#' git_add("hello.txt", repo = repo)
#' bad_commit <- git_commit("Second commit", repo = repo)
#'
#' # Default: revert and commit with an auto-generated message
#' git_revert(bad_commit, repo = repo)
#' git_log(repo = repo)
#'
#' # Revert with a custom message
#' writeLines("oops", file.path(repo, "hello.txt"))
#' git_add("hello.txt", repo = repo)
#' bad_commit2 <- git_commit("Third commit", repo = repo)
#' git_revert(bad_commit2, message = "Undo third commit\n", repo = repo)
#' git_log(repo = repo)
#'
#' # Stage the revert without committing
#' writeLines("again", file.path(repo, "hello.txt"))
#' git_add("hello.txt", repo = repo)
#' bad_commit3 <- git_commit("Fourth commit", repo = repo)
#' git_revert(bad_commit3, commit = FALSE, repo = repo)
#' git_status(repo = repo)
#'
#' unlink(repo, recursive = TRUE)
#' @useDynLib gert R_git_revert
#' @git revert
git_revert <- function(
  ref,
  commit = TRUE,
  ...,
  repo = '.'
) {
  repo <- git_open(repo)
  assert_string(ref)
  stopifnot(is.logical(commit), length(commit) == 1)

  sha <- check_ref_in_history(ref, repo)

  .Call(R_git_revert, repo, sha)

  if (isTRUE(commit)) {
    git_revert_commit(repo = repo, sha = sha, ...)
  }
}

git_revert_commit <- function(
  sha,
  message = revert_message(sha, repo),
  ...,
  repo
) {
  git_commit(message, ..., repo = repo)
}

assert_string <- function(x) {
  if (!is.character(x) || !length(x)) {
    stop("Argument must be a string of length 1")
  }
}

revert_message <- function(sha, repo) {
  info <- git_commit_info(sha, repo = repo)
  subject <- strsplit(info$message, "\n")[[1]][1]
  subject <- sub("\\s+$", "", subject)

  sprintf(
    'Revert "%s"\n\nThis reverts commit %s.\n',
    subject,
    info$id
  )
}

#' Git Ignore
#'
#' Test if files would be ignored by `.gitignore` rules
#'
#' @export
#' @rdname git_ignore
#' @name git_ignore
#' @family git
#' @inheritParams git_open
#' @param path A character vector of paths to test within the repo
#' @return A logical vector the same length as `path`, indicating if the
#' paths would be ignored.
#' @useDynLib gert R_git_ignore_path_is_ignored
#' @git ignore
#' @seealso `usethis::use_git_ignore()` to add a path to `.gitignore`.
git_ignore_path_is_ignored <- function(path, repo = '.') {
  repo <- git_open(repo)
  path <- as.character(path)
  .Call(R_git_ignore_path_is_ignored, repo, path)
}

#' Cherry-Pick and Rebase
#'
#' @description
#' * `git_cherry_pick()` applies the changes from a given commit (from another branch)
#' onto the current branch.
#' *`git_rebase_commit()` resets the branch to the state of another branch (upstream)
#' and then re-applies your local changes by cherry-picking each of your local
#' commits onto the upstream commit history.
#' *`git_rebase_list()` shows your local commits that are missing from the `upstream`
#' history, and if they conflict with upstream changes.
#'
#'
#' @details
#' To find if your local commits are missing from `upstream`,
#' `git_rebase_list()` first performs a rebase dry-run, without committing
#' anything. If there are no conflicts, you can use `git_rebase_commit()`
#' to rewind and rebase your branch onto `upstream`.
#'
#' Gert only support a clean rebase; it never leaves the repository in unfinished
#' "rebasing" state. If conflicts arise, `git_rebase_commit()` will raise an error
#' without making changes.
#'
#' @export
#' @rdname git_rebase
#' @name git_rebase
#' @family git
#' @param upstream branch to which you want to rewind and re-apply your
#' local commits. The default uses the remote upstream branch with the
#' current state on the git server, simulating [git_pull()].
#' @inheritParams git_open
#' @inheritParams git_branch
#' @git rebase
git_rebase_list <- function(upstream = NULL, repo = '.') {
  git_rebase(upstream = upstream, commit_changes = FALSE, repo = repo)
}

#' @export
#' @rdname git_rebase
git_rebase_commit <- function(upstream = NULL, repo = '.') {
  git_rebase(upstream = upstream, commit_changes = TRUE, repo = repo)
}

#' @useDynLib gert R_git_rebase
git_rebase <- function(upstream, commit_changes, repo) {
  repo <- git_open(repo)
  info <- git_info(repo = repo)
  if (!length(upstream)) {
    if (
      !length(info$upstream) || is.na(info$upstream) || !nchar(info$upstream)
    ) {
      stop("No upstream configured for current HEAD")
    }
    git_fetch(info$remote, repo = repo)
    upstream <- info$upstream
  }
  df <- .Call(R_git_rebase, repo, upstream, commit_changes)
  if (commit_changes) {
    new_head <- ifelse(nrow(df) > 0, utils::tail(df$commit, 1), upstream[1])
    git_branch_set_target(ref = new_head, repo = repo)
    inform("Resetting %s to %s", info$shorthand, new_head)
  }
  return(df)
}

#' Reset your repo to a previous state
#'
#' * `git_reset_hard()` resets the index and working tree
#' * `git_reset_soft()` does not touch the index file or the working tree
#' * `git_reset_mixed()` resets the index but not the working tree.
#'
#' @family git
#' @inheritParams git_rebase
#'
#' @export
#' @name git_reset
#' @rdname git_reset
#' @git reset
git_reset_hard <- function(ref = "HEAD", repo = ".") {
  git_reset("hard", ref = ref, repo = repo)
}

#' @export
#' @rdname git_reset
git_reset_soft <- function(ref = "HEAD", repo = ".") {
  git_reset("soft", ref = ref, repo = repo)
}

#' @export
#' @rdname git_reset
git_reset_mixed <- function(ref = "HEAD", repo = ".") {
  git_reset("mixed", ref = ref, repo = repo)
}

#' @useDynLib gert R_git_reset
git_reset <- function(
  type = c("soft", "hard", "mixed"),
  ref = "HEAD",
  repo = "."
) {
  typenum <- switch(match.arg(type), soft = 1L, mixed = 2L, hard = 3L)
  repo <- git_open(repo)
  ref <- as.character(ref)
  .Call(R_git_reset, repo, ref, typenum)
  git_status(repo = repo)
}

#' @export
#' @rdname git_rebase
#' @useDynLib gert R_git_cherry_pick
#' @param commit id of the commit to cherry pick
#' @git cherrypick
git_cherry_pick <- function(commit, repo = '.') {
  repo <- git_open(repo)
  assert_string(commit)
  .Call(R_git_cherry_pick, repo, commit)
}

#' @export
#' @rdname git_rebase
#' @useDynLib gert R_git_ahead_behind
#' @git graph
git_ahead_behind <- function(upstream = NULL, ref = 'HEAD', repo = '.') {
  repo <- git_open(repo)
  if (!length(upstream)) {
    upstream <- git_info(repo = repo)$upstream
  }
  if (!length(upstream)) {
    stop("No upstream set or specified")
  }
  .Call(R_git_ahead_behind, repo, ref, upstream)
}

.onAttach <- function(libname, pkgname) {
  config <- libgit2_config()
  ssh <- ifelse(config$ssh, "YES", "NO")
  https <- ifelse(config$https, "YES", "NO")
  packageStartupInform(
    "Linking to libgit2 v%s, ssh support: %s",
    as.character(config$version),
    ssh
  )
  user_config <- config$config.global
  if (length(user_config) && nchar(user_config)) {
    if (is_windows()) {
      user_config <- normalizePath(user_config, mustWork = FALSE)
    }
    packageStartupInform("Global config: %s", user_config)
  } else {
    packageStartupInform(
      "No global .gitconfig found in: %s",
      config$config.home
    )
  }
  if (length(config$config.system) && nchar(config$config.system)) {
    packageStartupInform(paste0("System config: ", config$config.system))
  }
  try({
    settings <- git_config_global()
    name <- subset(settings, name == 'user.name')$value
    email <- subset(settings, name == 'user.email')$value
    if (length(name) || length(email)) {
      packageStartupInform(
        "Default user: %s <%s>",
        as_string(name),
        as_string(email)
      )
    } else {
      packageStartupInform("No default user configured")
    }
  })
}

.onLoad <- function(libname, pkgname) {
  # This should usually correspond to 'openssl version -d'
  if (isTRUE(have_static_libgit2())) {
    certpath <- find_cert_dir()
    if (length(certpath)) {
      cafile <- file.path(dirname(certpath), 'cert.pem')
      if (!file.exists(cafile)) {
        cafile <- NULL
      }
      set_cert_locations(cafile, certpath)
    } else {
      warning("Unable to find directory with certificates", immediate. = TRUE)
    }
  }

  # Load tibble (if available) for pretty printing
  if (interactive() && is.null(.getNamespace('tibble'))) {
    tryCatch(
      {
        getNamespace('tibble')
      },
      error = function(e) {}
    )
  }

  invisible()
}

as_string <- function(x) {
  ifelse(length(x) > 0, x[1], NA_character_)
}

is_windows <- function() {
  identical(.Platform$OS.type, "windows")
}

is_solaris <- function() {
  isTRUE(grepl('solaris', R.version$platform, ignore.case = TRUE))
}

inform_impl <- function(message, subclass = NULL) {
  cnd <- structure(
    list(message = paste0(message, "\n")),
    class = c(subclass, "simpleMessage", "message", "condition")
  )
  withRestarts(
    muffleMessage = function() NULL,
    {
      signalCondition(cnd)
      cat(
        conditionMessage(cnd),
        file = if (interactive()) stdout() else stderr()
      )
    }
  )
}


inform <- function(...) inform_impl(sprintf(...), subclass = "gertMessage")

packageStartupInform <- function(...) {
  inform_impl(
    sprintf(...),
    subclass = c("gertMessage", "packageStartupMessage")
  )
}

message <- function(...) {
  stop("Internal error: use inform() instead of message()")
}

packageStartupMessage <- function(...) {
  stop(
    "Internal error: use packageStartupInform() instead of packageStartupMessage()"
  )
}

# libgit2 wants all paths utf-8
normalizePath <- function(...) {
  enc2utf8(base::normalizePath(...))
}

#' GitHub Wrappers
#'
#' Fetch and checkout pull requests.
#'
#' By default `git_fetch_pull_requests()` will download all PR branches. To
#' remove these again simply use [git_fetch(prune = TRUE)][git_fetch()].
#'
#' @export
#' @rdname github
#' @inheritParams git_fetch
#' @param pr number with PR to fetch or check out. Use `"*"` to fetch all
#' pull requests.
git_checkout_pull_request <- function(pr = 1, remote = NULL, repo = '.') {
  pr <- as.character(pr)
  if (!length(remote)) {
    remote <- git_info(repo)$remote
  }
  local_branch <- sprintf("pr/%s", pr)
  remote_branch <- sprintf("%s/pr/%s", remote, pr)
  git_fetch_pull_requests(pr = pr, remote = remote, repo = repo)
  if (git_branch_exists(local_branch, repo = repo)) {
    inform("Continuing on existing pr branch: %s", local_branch)
    git_branch_checkout(local_branch, repo = repo)
    git_pull(repo = repo)
  } else {
    inform("Creating new branch: %s", local_branch)
    git_branch_create(local_branch, remote_branch, checkout = TRUE, repo = repo)
  }
}

#' @export
#' @rdname github
git_fetch_pull_requests <- function(pr = '*', remote = NULL, repo = '.') {
  pr <- as.character(pr)
  if (!length(remote)) {
    remote <- git_info(repo)$remote
  }
  refspec <- sprintf('+refs/pull/%s/head:refs/remotes/%s/pr/%s', pr, remote, pr)
  git_fetch(remote = remote, refspec = refspec, repo = repo)
  invisible(refspec)
}

# Classed error handling
# Todo: kclass is ignored right now, would this be useful to report?
raise_libgit2_error <- function(code, message, where = "", kclass = 0) {
  cl <- call("::", substitute(libgit2), as.name(where))
  e <- structure(
    class = c(libgit2_error_name(code), "libgit2_error", "error", "condition"),
    list(message = message, call = cl) #call must be an R expression
  )
  stop(e)
}

libgit2_error_name <- function(x) {
  out <- which(libgit2_error_codes == x)
  if (length(out)) {
    return(names(out))
  }
  return("UNKNOWN_ERROR_CODE")
}

# Eror codes copied from https://github.com/libgit2/libgit2/blame/master/include/git2/errors.h
# We don't do this in C because the list of error codes changes from version to version
libgit2_error_codes <- c(
  GIT_OK = 0,
  GIT_ERROR = -1,
  GIT_ENOTFOUND = -3,
  GIT_EEXISTS = -4,
  GIT_EAMBIGUOUS = -5,
  GIT_EBUFS = -6,
  GIT_EUSER = -7,
  GIT_EBAREREPO = -8,
  GIT_EUNBORNBRANCH = -9,
  GIT_EUNMERGED = -10,
  GIT_ENONFASTFORWARD = -11,
  GIT_EINVALIDSPEC = -12,
  GIT_ECONFLICT = -13,
  GIT_ELOCKED = -14,
  GIT_EMODIFIED = -15,
  GIT_EAUTH = -16,
  GIT_ECERTIFICATE = -17,
  GIT_EAPPLIED = -18,
  GIT_EPEEL = -19,
  GIT_EEOF = -20,
  GIT_EINVALID = -21,
  GIT_EUNCOMMITTED = -22,
  GIT_EDIRECTORY = -23,
  GIT_EMERGECONFLICT = -24,
  GIT_PASSTHROUGH = -30,
  GIT_ITEROVER = -31,
  GIT_RETRY = -32,
  GIT_EMISMATCH = -33,
  GIT_EINDEXDIRTY = -34,
  GIT_EAPPLYFAIL = -35
)

---
title: "gert"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{gert}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  error = TRUE
)

if (identical(Sys.getenv("IN_PKGDOWN"), "true") &&
    !gert:::global_user_is_configured()) {
  gert:::configure_global_user()
}
```

gert is a simple git client based on 'libgit2' ([libgit2.org](https://libgit2.org)):

> libgit2 is a portable, pure C implementation of the Git core methods provided as a re-entrant linkable library with a solid API, allowing you to write native speed custom Git applications in any language which supports C bindings.

What this means for R users is that we can work with local and remote Git repositories from the comfort of R!

User-friendly authentication is a high priority for gert, which supports both SSH and HTTPS remotes on all platforms. User credentials are shared with command line Git through the git-credential store and ssh keys stored on disk or ssh-agent.

Let's attach gert.

```{r setup}
library(gert)
```

## Introduce yourself to Git

Before you can do anything with Git, you must first configure your user name and email. When you attach gert, it actually reveals whether you've already done this and, above, you can see that we have.

But what if you have not already configured your user name and email? Do this with `git_config_global_set()`:

```{r eval = FALSE}
git_config_global_set("user.name", "Jerry Johnson")
git_config_global_set("user.email", "jerry@gmail.com")
```

We can verify our success (and see all global options) with `git_config_global()`.

```{r}
git_config_global()
```

This is equivalent to these commands in command line Git:

```
git config --global user.name 'Jerry Johnson'
git config --global user.email 'jerry@gmail.com'
git config --global --list
```

To inspect and change local Git config, i.e. options specific to one repository, use `git_config()` and `git_config_set()`.

## Local repository basics

`gert::git_init()` is essentially `git init`; it's how we create a new local repository. You provide the path to the repository you want to create.

```{r}
(path <- file.path(tempdir(), "aaa", "bbb", "repo_ccc"))
dir.exists(path)

(r <- git_init(path))
dir.exists(path)
```

Note that all non-existing parts of the path are created: `aaa`, `bbb`, and `repo_ccc` (the actual git repository).

`git_find()` finds a git repository at or above the path you provide and errors otherwise.

```{r}
git_find(r)

dir.create(file.path(r, "child_dir"))
git_find(file.path(r, "child_dir"))

git_find(file.path(tempdir(), "aaa", "bbb"))
```

`git_init()` can also create a repository in a pre-existing directory, as long as it is empty.

```{r}
r2 <- file.path(tempdir(), "repo_ddd")
dir.create(r2)

git_init(r2)
```

Cleanup.

```{r}
unlink(r, recursive = TRUE)
unlink(r2, recursive = TRUE)
```

---
title: "gert"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{gert}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  error = TRUE
)
```

gert is a simple git client based on 'libgit2' ([libgit2.org](https://libgit2.org)):

> libgit2 is a portable, pure C implementation of the Git core methods provided as a re-entrant linkable library with a solid API, allowing you to write native speed custom Git applications in any language which supports C bindings.

What this means for R users is that we can work with local and remote Git repositories from the comfort of R!

User-friendly authentication is a high priority for gert, which supports both SSH and HTTPS remotes on all platforms. User credentials are shared with command line Git through the git-credential store and ssh keys stored on disk or ssh-agent.

Let's attach gert.

```{r setup}
library(gert)
```

## Introduce yourself to Git

Before you can do anything with Git, you must first configure your user name and email. When you attach gert, it actually reveals whether you've already done this and, above, you can see that we have.

But what if you have not already configured your user name and email? Do this with `git_config_global_set()`:

```{r eval = FALSE}
git_config_global_set("user.name", "Jerry Johnson")
git_config_global_set("user.email", "jerry@gmail.com")
```

We can verify our success (and see all global options) with `git_config_global()`.

```{r}
git_config_global()
```

This is equivalent to these commands in command line Git:

```
git config --global user.name 'Jerry Johnson'
git config --global user.email 'jerry@gmail.com'
git config --global --list
```

To inspect and change local Git config, i.e. options specific to one repository, use `git_config()` and `git_config_set()`.

## Local repository basics

`gert::git_init()` is essentially `git init`; it's how we create a new local repository. You provide the path to the repository you want to create.

```{r}
(path <- file.path(tempdir(), "aaa", "bbb", "repo_ccc"))
dir.exists(path)

(r <- git_init(path))
dir.exists(path)
```

Note that all non-existing parts of the path are created: `aaa`, `bbb`, and `repo_ccc` (the actual git repository).

`git_find()` finds a git repository at or above the path you provide and errors otherwise.

```{r}
git_find(r)

dir.create(file.path(r, "child_dir"))
git_find(file.path(r, "child_dir"))

git_find(file.path(tempdir(), "aaa", "bbb"))
```

`git_init()` can also create a repository in a pre-existing directory, as long as it is empty.

```{r}
r2 <- file.path(tempdir(), "repo_ddd")
dir.create(r2)

git_init(r2)
```

Cleanup.

```{r}
unlink(r, recursive = TRUE)
unlink(r2, recursive = TRUE)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/worktree.R
\name{git_worktree}
\alias{git_worktree}
\alias{git_worktree_list}
\alias{git_worktree_exists}
\alias{git_worktree_path}
\alias{git_worktree_add}
\alias{git_worktree_remove}
\alias{git_worktree_prune}
\alias{git_worktree_is_prunable}
\alias{git_worktree_lock}
\alias{git_worktree_unlock}
\alias{git_worktree_is_locked}
\alias{git_worktree_is_valid}
\title{Git Worktrees}
\usage{
git_worktree_list(repo = ".")

git_worktree_exists(name, repo = ".")

git_worktree_path(name, repo = ".")

git_worktree_add(name, path, branch, lock = FALSE, local = TRUE, repo = ".")

git_worktree_remove(name, repo = ".")

git_worktree_prune(
  name,
  prune_valid = FALSE,
  prune_locked = FALSE,
  prune_working_tree = FALSE,
  repo = "."
)

git_worktree_is_prunable(
  name,
  prune_valid = FALSE,
  prune_locked = FALSE,
  repo = "."
)

git_worktree_lock(name, repo = ".")

git_worktree_unlock(name, repo = ".")

git_worktree_is_locked(name, repo = ".")

git_worktree_is_valid(name, repo = ".")
}
\arguments{
\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{name}{The name of the worktree.}

\item{path}{The path to checkout \code{branch} into. Importantly, the path up to
the folder name must exist, but the folder name itself must not exist yet
and will be created.}

\item{branch}{The branch to checkout into \code{path}.}

\item{lock}{Whether or not to lock the worktree on creation.}

\item{local}{set TRUE to only check for local branches, FALSE to check for remote
branches. Use NULL to return all branches.}

\item{prune_valid}{Whether or not to forcibly prune a \emph{valid} worktree.}

\item{prune_locked}{Whether or not to forcibly prune a \emph{locked} worktree.}

\item{prune_working_tree}{Whether or not to also remove the folder that the
worktree was using, i.e. the \code{path} supplied to \code{git_worktree_add()}.}
}
\description{
Worktrees represent an alternative location to checkout a branch into. Rather
than checking out a branch in your main working tree (which changes the
branch you are currently on and forces you to stash any existing work), you
can instead check that branch out into a separate \emph{linked worktree} with its
own working tree. Practically, a worktree is just a separate folder that a
branch is checked out into, with some extra git metadata that links it back
to the main working tree.

\code{git_worktree_list()} returns a data frame of information about the worktrees
linked to the main working tree.

\code{git_worktree_exists()} lets you check whether or not a worktree by the name
of \code{name} exists for this \code{repo}.

\code{git_worktree_path()} returns the file path to the worktree.

\code{git_worktree_add()} creates a new worktree called \code{name} in the folder
pointed to by \code{path}, and checks \code{branch} out into it.

\code{git_worktree_remove()} removes a worktree. It does so by deleting the folder
provided as the \code{path} to \code{git_worktree_add()}, and then cleaning up some git
metadata in the main working tree that linked the main working tree to the
removed worktree. The \code{branch} checked out by the worktree is not deleted.
Note that this is just a wrapper around \code{git_worktree_prune()} that sets some
desirable defaults for aggressive removal.

\code{git_worktree_prune()} is more cautious than \code{git_worktree_remove()}. It
refuses to prune \emph{valid} or \emph{locked} worktrees by default, and also refuses
the delete the working tree of the worktree by default (i.e. the folder at
\code{path}). It is automatically run by git itself on periodic intervals to prune
outdated worktrees. For interactive usage, you typically want
\code{git_worktree_remove()} instead. \code{git_worktree_is_prunable()} lets you check
if a worktree is prunable with the given options.

\code{git_worktree_lock()}, \code{git_worktree_unlock()}, and
\code{git_worktree_is_locked()} help you manage whether or not a worktree is
\emph{locked}. When a worktree is locked, it is not automatically cleaned up by
\code{git_worktree_prune()} (and git itself) on periodic intervals, even when it
looks \emph{invalid}. This is typically only useful when your worktree is on a
hard drive that isn't always connected (which can make it look \emph{invalid} when
disconnected, typically making it a candidate for automatic pruning).

\code{git_worktree_is_valid()} checks whether a worktree is valid or not. A
\emph{valid} worktree requires both the git data structures inside the main
working tree and this worktree to be present.
}
\examples{
repo <- git_init(tempfile("gert-examples-repo"))

writeLines("hello", file.path(repo, 'hello.txt'))
git_add('hello.txt', repo = repo)
git_commit("First commit", author = "jeroen <jeroen@blabla.nl>", repo = repo)

# Create a branch that is going to be used for the worktree,
# but don't check it out!
git_branch_create(branch = "branch", checkout = FALSE, repo = repo)

path <- tempfile("gert-examples-worktree")

# Add a worktree for this branch
git_worktree_add(
  name = "worktree",
  path = path,
  branch = "branch",
  repo = repo
)

# Worktree info
git_worktree_list(repo = repo)

# Note how the files are checked out here
dir(path, all.files = TRUE)

# And the branch that we are on at `path` is `"branch"`
git_branch(repo = path)

# Cleanup worktree, and the folder at `path`
git_worktree_remove("worktree", repo = repo)

# Cleanup repo
unlink(repo, recursive = TRUE)
}
\seealso{
Other git: 
\code{\link{git_archive}},
\code{\link{git_branch}()},
\code{\link{git_commit}()},
\code{\link{git_config}()},
\code{\link{git_diff}()},
\code{\link{git_fetch}()},
\code{\link{git_ignore}},
\code{\link{git_merge}()},
\code{\link{git_rebase}()},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link{git_reset}()},
\code{\link{git_signature}()},
\code{\link{git_stash}},
\code{\link{git_tag}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/worktree.R
\name{git_worktree}
\alias{git_worktree}
\alias{git_worktree_list}
\alias{git_worktree_exists}
\alias{git_worktree_path}
\alias{git_worktree_add}
\alias{git_worktree_remove}
\alias{git_worktree_prune}
\alias{git_worktree_is_prunable}
\alias{git_worktree_lock}
\alias{git_worktree_unlock}
\alias{git_worktree_is_locked}
\alias{git_worktree_is_valid}
\title{Git Worktrees}
\usage{
git_worktree_list(repo = ".")

git_worktree_exists(name, repo = ".")

git_worktree_path(name, repo = ".")

git_worktree_add(name, path, branch, lock = FALSE, local = TRUE, repo = ".")

git_worktree_remove(name, repo = ".")

git_worktree_prune(
  name,
  prune_valid = FALSE,
  prune_locked = FALSE,
  prune_working_tree = FALSE,
  repo = "."
)

git_worktree_is_prunable(
  name,
  prune_valid = FALSE,
  prune_locked = FALSE,
  repo = "."
)

git_worktree_lock(name, repo = ".")

git_worktree_unlock(name, repo = ".")

git_worktree_is_locked(name, repo = ".")

git_worktree_is_valid(name, repo = ".")
}
\arguments{
\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \code{\link[=git_find]{git_find()}}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{name}{The name of the worktree.}

\item{path}{The path to checkout \code{branch} into. Importantly, the path up to
the folder name must exist, but the folder name itself must not exist yet
and will be created.}

\item{branch}{The branch to checkout into \code{path}.}

\item{lock}{Whether or not to lock the worktree on creation.}

\item{local}{set TRUE to only check for local branches, FALSE to check for remote
branches. Use NULL to return all branches.}

\item{prune_valid}{Whether or not to forcibly prune a \emph{valid} worktree.}

\item{prune_locked}{Whether or not to forcibly prune a \emph{locked} worktree.}

\item{prune_working_tree}{Whether or not to also remove the folder that the
worktree was using, i.e. the \code{path} supplied to \code{git_worktree_add()}.}
}
\description{
Worktrees represent an alternative location to checkout a branch into. Rather
than checking out a branch in your main working tree (which changes the
branch you are currently on and forces you to stash any existing work), you
can instead check that branch out into a separate \emph{linked worktree} with its
own working tree. Practically, a worktree is just a separate folder that a
branch is checked out into, with some extra git metadata that links it back
to the main working tree.

\code{git_worktree_list()} returns a data frame of information about the worktrees
linked to the main working tree.

\code{git_worktree_exists()} lets you check whether or not a worktree by the name
of \code{name} exists for this \code{repo}.

\code{git_worktree_path()} returns the file path to the worktree.

\code{git_worktree_add()} creates a new worktree called \code{name} in the folder
pointed to by \code{path}, and checks \code{branch} out into it.

\code{git_worktree_remove()} removes a worktree. It does so by deleting the folder
provided as the \code{path} to \code{git_worktree_add()}, and then cleaning up some git
metadata in the main working tree that linked the main working tree to the
removed worktree. The \code{branch} checked out by the worktree is not deleted.
Note that this is just a wrapper around \code{git_worktree_prune()} that sets some
desirable defaults for aggressive removal.

\code{git_worktree_prune()} is more cautious than \code{git_worktree_remove()}. It
refuses to prune \emph{valid} or \emph{locked} worktrees by default, and also refuses
the delete the working tree of the worktree by default (i.e. the folder at
\code{path}). It is automatically run by git itself on periodic intervals to prune
outdated worktrees. For interactive usage, you typically want
\code{git_worktree_remove()} instead. \code{git_worktree_is_prunable()} lets you check
if a worktree is prunable with the given options.

\code{git_worktree_lock()}, \code{git_worktree_unlock()}, and
\code{git_worktree_is_locked()} help you manage whether or not a worktree is
\emph{locked}. When a worktree is locked, it is not automatically cleaned up by
\code{git_worktree_prune()} (and git itself) on periodic intervals, even when it
looks \emph{invalid}. This is typically only useful when your worktree is on a
hard drive that isn't always connected (which can make it look \emph{invalid} when
disconnected, typically making it a candidate for automatic pruning).

\code{git_worktree_is_valid()} checks whether a worktree is valid or not. A
\emph{valid} worktree requires both the git data structures inside the main
working tree and this worktree to be present.
}
\examples{
repo <- git_init(tempfile("gert-examples-repo"))

writeLines("hello", file.path(repo, 'hello.txt'))
git_add('hello.txt', repo = repo)
git_commit("First commit", author = "jeroen <jeroen@blabla.nl>", repo = repo)

# Create a branch that is going to be used for the worktree,
# but don't check it out!
git_branch_create(branch = "branch", checkout = FALSE, repo = repo)

path <- tempfile("gert-examples-worktree")

# Add a worktree for this branch
git_worktree_add(
  name = "worktree",
  path = path,
  branch = "branch",
  repo = repo
)

# Worktree info
git_worktree_list(repo = repo)

# Note how the files are checked out here
dir(path, all.files = TRUE)

# And the branch that we are on at `path` is `"branch"`
git_branch(repo = path)

# Cleanup worktree, and the folder at `path`
git_worktree_remove("worktree", repo = repo)

# Cleanup repo
unlink(repo, recursive = TRUE)
}
\seealso{
Other git:
\code{\link{git_archive}},
\code{\link[=git_branch]{git_branch()}},
\code{\link[=git_commit]{git_commit()}},
\code{\link[=git_config]{git_config()}},
\code{\link[=git_diff]{git_diff()}},
\code{\link[=git_fetch]{git_fetch()}},
\code{\link{git_history}},
\code{\link{git_ignore}},
\code{\link[=git_merge]{git_merge()}},
\code{\link[=git_rebase]{git_rebase()}},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link[=git_reset]{git_reset()}},
\code{\link[=git_restore]{git_restore()}},
\code{\link[=git_revert]{git_revert()}},
\code{\link[=git_signature]{git_signature()}},
\code{\link{git_stash}},
\code{\link{git_tag}}
}
\concept{git}
\section{Related libgit2 documentation}{\href{https://libgit2.org/docs/reference/main/worktree/index.html}{\code{worktree}}.}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gert-package.R
\docType{package}
\name{gert-package}
\alias{gert}
\alias{gert-package}
\title{gert: Simple Git Client for R}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Simple git client for R based on 'libgit2' \url{https://libgit2.org} with support for SSH and HTTPS remotes. All functions in 'gert' use basic R data types (such as vectors and data-frames) for their arguments and return values. User credentials are shared with command line 'git' through the git-credential store and ssh keys stored on disk or ssh-agent.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://docs.ropensci.org/gert/}
  \item \url{https://github.com/r-lib/gert}
  \item Report bugs at \url{https://github.com/r-lib/gert/issues}
}

}
\author{
\strong{Maintainer}: Jeroen Ooms \email{jeroen@berkeley.edu} (\href{https://orcid.org/0000-0002-4035-0289}{ORCID})

Other contributors:
\itemize{
  \item Jennifer Bryan \email{jenny@rstudio.com} (\href{https://orcid.org/0000-0002-6983-2759}{ORCID}) [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rebase.R
\name{git_reset}
\alias{git_reset}
\alias{git_reset_hard}
\alias{git_reset_soft}
\alias{git_reset_mixed}
\title{Reset your repo to a previous state}
\usage{
git_reset_hard(ref = "HEAD", repo = ".")

git_reset_soft(ref = "HEAD", repo = ".")

git_reset_mixed(ref = "HEAD", repo = ".")
}
\arguments{
\item{ref}{string with a branch/tag/commit}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}
}
\description{
\itemize{
\item \code{git_reset_hard()} resets the index and working tree
\item \code{git_reset_soft()} does not touch the index file or the working tree
\item \code{git_reset_mixed()} resets the index but not the working tree.
}
}
\seealso{
Other git: 
\code{\link{git_archive}},
\code{\link{git_branch}()},
\code{\link{git_commit}()},
\code{\link{git_config}()},
\code{\link{git_diff}()},
\code{\link{git_fetch}()},
\code{\link{git_ignore}},
\code{\link{git_merge}()},
\code{\link{git_rebase}()},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link{git_signature}()},
\code{\link{git_stash}},
\code{\link{git_tag}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rebase.R
\name{git_rebase}
\alias{git_rebase}
\alias{git_rebase_list}
\alias{git_rebase_commit}
\alias{git_reset_hard}
\alias{git_reset_soft}
\alias{git_reset_mixed}
\alias{git_cherry_pick}
\alias{git_ahead_behind}
\title{Cherry-Pick and Rebase}
\usage{
git_rebase_list(upstream = NULL, repo = ".")

git_rebase_commit(upstream = NULL, repo = ".")

git_reset_hard(ref = "HEAD", repo = ".")

git_reset_soft(ref = "HEAD", repo = ".")

git_reset_mixed(ref = "HEAD", repo = ".")

git_cherry_pick(commit, repo = ".")

git_ahead_behind(upstream = NULL, ref = "HEAD", repo = ".")
}
\arguments{
\item{upstream}{branch to which you want to rewind and re-apply your
local commits. The default uses the remote upstream branch with the
current state on the git server, simulating \link{git_pull}.}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{ref}{string with a branch/tag/commit}

\item{commit}{id of the commit to cherry pick}
}
\description{
A cherry-pick applies the changes from a given commit (from another branch)
onto the current branch. A rebase resets the branch to the state of another
branch (upstream) and then re-applies your local changes by cherry-picking
each of your local commits onto the upstream commit history.
}
\details{
\code{git_rebase_list} shows your local commits that are missing from the \code{upstream}
history, and if they conflict with upstream changes. It does so by performing
a rebase dry-run, without committing anything. If there are no conflicts, you
can use \code{git_rebase_commit} to rewind and rebase your branch onto \code{upstream}.
Gert only support a clean rebase; it never leaves the repository in unfinished
"rebasing" state. If conflicts arise, \code{git_rebase_commit} will raise an error
without making changes.
}
\seealso{
Other git: 
\code{\link{git_archive}},
\code{\link{git_branch}()},
\code{\link{git_commit}()},
\code{\link{git_config}()},
\code{\link{git_diff}()},
\code{\link{git_fetch}()},
\code{\link{git_merge}()},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link{git_signature}()},
\code{\link{git_stash}},
\code{\link{git_tag}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pr.R
\name{git_checkout_pull_request}
\alias{git_checkout_pull_request}
\alias{git_fetch_pull_requests}
\title{GitHub Wrappers}
\usage{
git_checkout_pull_request(pr = 1, remote = NULL, repo = ".")

git_fetch_pull_requests(pr = "*", remote = NULL, repo = ".")
}
\arguments{
\item{pr}{number with PR to fetch or check out. Use \code{"*"} to fetch all
pull requests.}

\item{remote}{Optional. Name of a remote listed in \code{\link[=git_remote_list]{git_remote_list()}}. If
unspecified and the current branch is already tracking branch a remote
branch, that remote is honored. Otherwise, defaults to \code{origin}.}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}
}
\description{
Fetch and checkout pull requests.
}
\details{
By default \code{git_fetch_pull_requests} will download all PR branches. To
remove these again simply use \code{git_fetch(prune = TRUE)}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/restore.R
\name{git_restore}
\alias{git_restore}
\title{Restore working tree files}
\usage{
git_restore(path, ref = "HEAD", repo = ".")
}
\arguments{
\item{path}{character vector with file paths to restore, relative to the
repository root. Use \code{"."} to restore all tracked files.}

\item{ref}{revision string with a branch/tag/commit to restore from.
Defaults to \code{"HEAD"}.}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \code{\link[=git_find]{git_find()}}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}
}
\value{
Invisibly, the \code{\link[=git_status]{git_status()}} after restoring.
}
\description{
Restores specified paths in the working tree from a given ref, equivalent
to \verb{git restore --source=<ref> <path>} (or the older
\verb{git checkout <ref> -- <path>}). The ref must be reachable from the
current HEAD. By default restores from HEAD, discarding any local
modifications.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
repo <- file.path(tempdir(), "myrepo")
git_init(repo)

# Set a user if no default
if (!user_is_configured()) {
  git_config_set("user.name", "Jerry")
  git_config_set("user.email", "jerry@gmail.com")
}

writeLines("hello", file.path(repo, "hello.txt"))
git_add("hello.txt", repo = repo)
git_commit("First commit", repo = repo)

# Modify the file, then restore it from HEAD
writeLines("oops", file.path(repo, "hello.txt"))
git_restore("hello.txt", repo = repo)
readLines(file.path(repo, "hello.txt"))  # "hello"

unlink(repo, recursive = TRUE)
\dontshow{\}) # examplesIf}
}
\seealso{
Other git:
\code{\link{git_archive}},
\code{\link[=git_branch]{git_branch()}},
\code{\link[=git_commit]{git_commit()}},
\code{\link[=git_config]{git_config()}},
\code{\link[=git_diff]{git_diff()}},
\code{\link[=git_fetch]{git_fetch()}},
\code{\link{git_history}},
\code{\link{git_ignore}},
\code{\link[=git_merge]{git_merge()}},
\code{\link[=git_rebase]{git_rebase()}},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link[=git_reset]{git_reset()}},
\code{\link[=git_revert]{git_revert()}},
\code{\link[=git_signature]{git_signature()}},
\code{\link{git_stash}},
\code{\link{git_tag}},
\code{\link{git_worktree}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rebase.R
\name{git_rebase}
\alias{git_rebase}
\alias{git_rebase_list}
\alias{git_rebase_commit}
\alias{git_reset_hard}
\alias{git_reset_soft}
\alias{git_reset_mixed}
\alias{git_cherry_pick}
\alias{git_ahead_behind}
\title{Cherry-Pick and Rebase}
\usage{
git_rebase_list(upstream = NULL, repo = ".")

git_rebase_commit(upstream = NULL, repo = ".")

git_reset_hard(ref = "HEAD", repo = ".")

git_reset_soft(ref = "HEAD", repo = ".")

git_reset_mixed(ref = "HEAD", repo = ".")

git_cherry_pick(commit, repo = ".")

git_ahead_behind(upstream = NULL, ref = "HEAD", repo = ".")
}
\arguments{
\item{upstream}{branch to which you want to rewind and re-apply your
local commits. The default uses the remote upstream branch with the
current state on the git server, simulating \link{git_pull}.}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{ref}{string with a branch/tag/commit}

\item{commit}{id of the commit to cherry pick}
}
\description{
A cherry-pick applies the changes from a given commit (from another branch)
onto the current branch. A rebase resets the branch to the state of another
branch (upstream) and then re-applies your local changes by cherry-picking
each of your local commits onto the upstream commit history.
}
\details{
\code{git_rebase_list} shows your local commits that are missing from the \code{upstream}
history, and if they conflict with upstream changes. It does so by performing
a rebase dry-run, without committing anything. If there are no conflicts, you
can use \code{git_rebase_commit} to rewind and rebase your branch onto \code{upstream}.
Gert only support a clean rebase; it never leaves the repository in unfinished
"rebasing" state. If conflicts arise, \code{git_rebase_commit} will raise an error
without making changes.
}
\seealso{
Other git: 
\code{\link{git_archive}},
\code{\link{git_branch}()},
\code{\link{git_commit}()},
\code{\link{git_config}()},
\code{\link{git_diff}()},
\code{\link{git_fetch}()},
\code{\link{git_ignore}},
\code{\link{git_merge}()},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link{git_signature}()},
\code{\link{git_stash}},
\code{\link{git_tag}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gert-package.R
\docType{package}
\name{gert-package}
\alias{gert}
\alias{gert-package}
\title{gert: Simple Git Client for R}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Simple git client for R based on 'libgit2' \url{https://libgit2.org} with support for SSH and HTTPS remotes. All functions in 'gert' use basic R data types (such as vectors and data-frames) for their arguments and return values. User credentials are shared with command line 'git' through the git-credential store and ssh keys stored on disk or ssh-agent.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://docs.ropensci.org/gert/}
  \item \url{https://ropensci.r-universe.dev/gert}
  \item Report bugs at \url{https://github.com/r-lib/gert/issues}
}

}
\author{
\strong{Maintainer}: Jeroen Ooms \email{jeroenooms@gmail.com} (\href{https://orcid.org/0000-0002-4035-0289}{ORCID})

Other contributors:
\itemize{
  \item Jennifer Bryan \email{jenny@posit.co} (\href{https://orcid.org/0000-0002-6983-2759}{ORCID}) [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/commit.R
\name{git_commit}
\alias{git_commit}
\alias{git_commit_all}
\alias{git_commit_info}
\alias{git_commit_id}
\alias{git_commit_descendant_of}
\alias{git_add}
\alias{git_rm}
\alias{git_status}
\alias{git_conflicts}
\alias{git_ls}
\alias{git_log}
\alias{git_stat_files}
\title{Stage and commit changes}
\usage{
git_commit(message, author = NULL, committer = NULL, repo = ".")

git_commit_all(message, author = NULL, committer = NULL, repo = ".")

git_commit_info(ref = "HEAD", repo = ".")

git_commit_id(ref = "HEAD", repo = ".")

git_commit_descendant_of(ancestor, ref = "HEAD", repo = ".")

git_add(files, force = FALSE, repo = ".")

git_rm(files, repo = ".")

git_status(staged = NULL, repo = ".")

git_conflicts(repo = ".")

git_ls(repo = ".")

git_log(ref = "HEAD", max = 100, repo = ".")

git_stat_files(files, ref = "HEAD", repo = ".")
}
\arguments{
\item{message}{a commit message}

\item{author}{A \link{git_signature} value, default is \code{\link[=git_signature_default]{git_signature_default()}}.}

\item{committer}{A \link{git_signature} value, default is same as \code{author}}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{ref}{revision string with a branch/tag/commit value}

\item{ancestor}{a reference to a potential ancestor commit}

\item{files}{vector of paths relative to the git root directory.
Use \code{"."} to stage all changed files.}

\item{force}{add files even if in gitignore}

\item{staged}{return only staged (TRUE) or unstaged files (FALSE).
Use \code{NULL} or \code{NA} to show both (default).}

\item{max}{lookup at most latest n parent commits}
}
\value{
\itemize{
\item \code{git_status()}, \code{git_ls()}: A data frame with one row per file
\item \code{git_log()}: A data frame with one row per commit
\item \code{git_commit()}, \code{git_commit_all()}: A SHA
}
}
\description{
To commit changes, start by \emph{staging} the files to be included in the commit
using \code{git_add()} or \code{git_rm()}. Use \code{git_status()} to see an overview of
staged and unstaged changes, and finally \code{git_commit()} creates a new commit
with currently staged files.

\code{git_commit_all()} is a convenience function that automatically stages and
commits all modified files. Note that \code{git_commit_all()} does \strong{not} add
new, untracked files to the repository. You need to make an explicit call to
\code{git_add()} to start tracking new files.

\code{git_log()} shows the most recent commits and \code{git_ls()} lists all the files
that are being tracked in the repository. \code{git_stat_files()}
}
\examples{
oldwd <- getwd()
repo <- file.path(tempdir(), "myrepo")
git_init(repo)
setwd(repo)

# Set a user if no default
if(!user_is_configured()){
  git_config_set("user.name", "Jerry")
  git_config_set("user.email", "jerry@gmail.com")
}

writeLines(letters[1:6], "alphabet.txt")
git_status()

git_add("alphabet.txt")
git_status()

git_commit("Start alphabet file")
git_status()

git_ls()

git_log()

cat(letters[7:9], file = "alphabet.txt", sep = "\n", append = TRUE)
git_status()

git_commit_all("Add more letters")

# cleanup
setwd(oldwd)
unlink(repo, recursive = TRUE)
}
\seealso{
Other git: 
\code{\link{git_archive}},
\code{\link{git_branch}()},
\code{\link{git_config}()},
\code{\link{git_diff}()},
\code{\link{git_fetch}()},
\code{\link{git_merge}()},
\code{\link{git_rebase}()},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link{git_signature}()},
\code{\link{git_stash}},
\code{\link{git_tag}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rebase.R
\name{git_reset}
\alias{git_reset}
\alias{git_reset_hard}
\alias{git_reset_soft}
\alias{git_reset_mixed}
\title{Reset your repo to a previous state}
\usage{
git_reset_hard(ref = "HEAD", repo = ".")

git_reset_soft(ref = "HEAD", repo = ".")

git_reset_mixed(ref = "HEAD", repo = ".")
}
\arguments{
\item{ref}{string with a branch/tag/commit}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \code{\link[=git_find]{git_find()}}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}
}
\description{
\itemize{
\item \code{git_reset_hard()} resets the index and working tree
\item \code{git_reset_soft()} does not touch the index file or the working tree
\item \code{git_reset_mixed()} resets the index but not the working tree.
}
}
\seealso{
Other git:
\code{\link{git_archive}},
\code{\link[=git_branch]{git_branch()}},
\code{\link[=git_commit]{git_commit()}},
\code{\link[=git_config]{git_config()}},
\code{\link[=git_diff]{git_diff()}},
\code{\link[=git_fetch]{git_fetch()}},
\code{\link{git_history}},
\code{\link{git_ignore}},
\code{\link[=git_merge]{git_merge()}},
\code{\link[=git_rebase]{git_rebase()}},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link[=git_restore]{git_restore()}},
\code{\link[=git_revert]{git_revert()}},
\code{\link[=git_signature]{git_signature()}},
\code{\link{git_stash}},
\code{\link{git_tag}},
\code{\link{git_worktree}}
}
\concept{git}
\section{Related libgit2 documentation}{\href{https://libgit2.org/docs/reference/main/reset/index.html}{\code{reset}}.}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/commit.R
\name{git_commit}
\alias{git_commit}
\alias{git_commit_all}
\alias{git_commit_info}
\alias{git_commit_id}
\alias{git_commit_stats}
\alias{git_commit_descendant_of}
\alias{git_add}
\alias{git_rm}
\alias{git_status}
\alias{git_conflicts}
\alias{git_ls}
\alias{git_log}
\alias{git_stat_files}
\title{Stage and commit changes}
\usage{
git_commit(message, author = NULL, committer = NULL, repo = ".")

git_commit_all(message, author = NULL, committer = NULL, repo = ".")

git_commit_info(ref = "HEAD", repo = ".")

git_commit_id(ref = "HEAD", repo = ".")

git_commit_stats(ref = "HEAD", repo = ".")

git_commit_descendant_of(ancestor, ref = "HEAD", repo = ".")

git_add(files, force = FALSE, repo = ".")

git_rm(files, repo = ".")

git_status(staged = NULL, pathspec = NULL, repo = ".")

git_conflicts(repo = ".")

git_ls(repo = ".", ref = NULL)

git_log(ref = "HEAD", max = 100, after = NULL, repo = ".")

git_stat_files(files, ref = "HEAD", repo = ".")
}
\arguments{
\item{message}{a commit message}

\item{author}{A \link{git_signature} value, default is \code{\link[=git_signature_default]{git_signature_default()}}.}

\item{committer}{A \link{git_signature} value, default is same as \code{author}}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{ref}{revision string with a branch/tag/commit value}

\item{ancestor}{a reference to a potential ancestor commit}

\item{files}{vector of paths relative to the git root directory.
Use \code{"."} to stage all changed files.}

\item{force}{add files even if in gitignore}

\item{staged}{return only staged (TRUE) or unstaged files (FALSE).
Use \code{NULL} or \code{NA} to show both (default).}

\item{pathspec}{character vector with paths to match}

\item{max}{lookup at most latest n parent commits}

\item{after}{date or timestamp: only include commits starting this date}
}
\value{
\itemize{
\item \code{git_status()}, \code{git_ls()}: A data frame with one row per file
\item \code{git_log()}: A data frame with one row per commit
\item \code{git_commit()}, \code{git_commit_all()}: A SHA
}
}
\description{
To commit changes, start by \emph{staging} the files to be included in the commit
using \code{git_add()} or \code{git_rm()}. Use \code{git_status()} to see an overview of
staged and unstaged changes, and finally \code{git_commit()} creates a new commit
with currently staged files.

\code{git_commit_all()} is a convenience function that automatically stages and
commits all modified files. Note that \code{git_commit_all()} does \strong{not} add
new, untracked files to the repository. You need to make an explicit call to
\code{git_add()} to start tracking new files.

\code{git_log()} shows the most recent commits and \code{git_ls()} lists all the files
that are being tracked in the repository. \code{git_stat_files()}
}
\examples{
oldwd <- getwd()
repo <- file.path(tempdir(), "myrepo")
git_init(repo)
setwd(repo)

# Set a user if no default
if(!user_is_configured()){
  git_config_set("user.name", "Jerry")
  git_config_set("user.email", "jerry@gmail.com")
}

writeLines(letters[1:6], "alphabet.txt")
git_status()

git_add("alphabet.txt")
git_status()

git_commit("Start alphabet file")
git_status()

git_ls()

git_log()

cat(letters[7:9], file = "alphabet.txt", sep = "\n", append = TRUE)
git_status()

git_commit_all("Add more letters")

# cleanup
setwd(oldwd)
unlink(repo, recursive = TRUE)
}
\seealso{
Other git: 
\code{\link{git_archive}},
\code{\link{git_branch}()},
\code{\link{git_config}()},
\code{\link{git_diff}()},
\code{\link{git_fetch}()},
\code{\link{git_ignore}},
\code{\link{git_merge}()},
\code{\link{git_rebase}()},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link{git_signature}()},
\code{\link{git_stash}},
\code{\link{git_tag}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repository.R
\name{git_repo}
\alias{git_repo}
\alias{git_init}
\alias{git_find}
\alias{git_info}
\title{Create or discover a local Git repository}
\usage{
git_init(path = ".", bare = FALSE)

git_find(path = ".")

git_info(repo = ".")
}
\arguments{
\item{path}{the location of the git repository, see details.}

\item{bare}{if true, a Git repository without a working directory is created}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}
}
\value{
The path to the Git repository.
}
\description{
Use \code{git_init()} to create a new repository or \code{git_find()} to discover an
existing local repository. \code{git_info()} shows basic information about a
repository, such as the SHA and branch of the current HEAD.
}
\details{
For \code{git_init()} the \code{path} parameter sets the directory of the git repository
to create. If this directory already exists, it must be empty. If it does
not exist, it is created, along with any intermediate directories that don't
yet exist. For \code{git_find()} the \code{path} arguments specifies the directory at
which to start the search for a git repository. If it is not a git repository
itself, then its parent directory is consulted, then the parent's parent, and
so on.
}
\examples{
# directory does not yet exist
r <- tempfile(pattern = "gert")
git_init(r)
git_find(r)

# create a child directory, then a grandchild, then search
r_grandchild_dir <- file.path(r, "aaa", "bbb")
dir.create(r_grandchild_dir, recursive = TRUE)
git_find(r_grandchild_dir)

# cleanup
unlink(r, recursive = TRUE)

# directory exists but is empty
r <- tempfile(pattern = "gert")
dir.create(r)
git_init(r)
git_find(r)

# cleanup
unlink(r, recursive = TRUE)
}
\seealso{
Other git: 
\code{\link{git_archive}},
\code{\link{git_branch}()},
\code{\link{git_commit}()},
\code{\link{git_config}()},
\code{\link{git_diff}()},
\code{\link{git_fetch}()},
\code{\link{git_merge}()},
\code{\link{git_rebase}()},
\code{\link{git_remote}},
\code{\link{git_signature}()},
\code{\link{git_stash}},
\code{\link{git_tag}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/commit.R
\name{git_revert}
\alias{git_revert}
\title{Revert a commit}
\usage{
git_revert(ref, commit = TRUE, ..., repo = ".")
}
\arguments{
\item{ref}{revision string with a branch/tag/commit value}

\item{commit}{if \code{FALSE}, stage the reverted changes without creating a
commit. Default is \code{TRUE}, that is to say, by default a commit is made.}

\item{...}{parameters passed to \code{git_commit} such as \code{message} or \code{author}}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \code{\link[=git_find]{git_find()}}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}
}
\value{
The SHA of the new revert commit (invisibly), or \code{NULL} when
\code{commit = FALSE}.
}
\description{
Applies the inverse of the changes introduced by a given commit, equivalent
to \verb{git revert <commit>}. The commit must be reachable from the current HEAD.
}
\details{
By default, a new revert commit is created immediately. Set \code{commit = FALSE}
to only stage the reverted changes without committing, leaving you free to
amend or combine them before calling \code{\link[=git_commit]{git_commit()}}.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
repo <- file.path(tempdir(), "myrepo")
git_init(repo)

# Set a user if no default
if (!user_is_configured()) {
  git_config_set("user.name", "Jerry")
  git_config_set("user.email", "jerry@gmail.com")
}

writeLines("hello", file.path(repo, "hello.txt"))
git_add("hello.txt", repo = repo)
git_commit("First commit", repo = repo)

writeLines("world", file.path(repo, "hello.txt"))
git_add("hello.txt", repo = repo)
bad_commit <- git_commit("Second commit", repo = repo)

# Default: revert and commit with an auto-generated message
git_revert(bad_commit, repo = repo)
git_log(repo = repo)

# Revert with a custom message
writeLines("oops", file.path(repo, "hello.txt"))
git_add("hello.txt", repo = repo)
bad_commit2 <- git_commit("Third commit", repo = repo)
git_revert(bad_commit2, message = "Undo third commit\n", repo = repo)
git_log(repo = repo)

# Stage the revert without committing
writeLines("again", file.path(repo, "hello.txt"))
git_add("hello.txt", repo = repo)
bad_commit3 <- git_commit("Fourth commit", repo = repo)
git_revert(bad_commit3, commit = FALSE, repo = repo)
git_status(repo = repo)

unlink(repo, recursive = TRUE)
\dontshow{\}) # examplesIf}
}
\seealso{
Other git:
\code{\link{git_archive}},
\code{\link[=git_branch]{git_branch()}},
\code{\link[=git_commit]{git_commit()}},
\code{\link[=git_config]{git_config()}},
\code{\link[=git_diff]{git_diff()}},
\code{\link[=git_fetch]{git_fetch()}},
\code{\link{git_history}},
\code{\link{git_ignore}},
\code{\link[=git_merge]{git_merge()}},
\code{\link[=git_rebase]{git_rebase()}},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link[=git_reset]{git_reset()}},
\code{\link[=git_restore]{git_restore()}},
\code{\link[=git_signature]{git_signature()}},
\code{\link{git_stash}},
\code{\link{git_tag}},
\code{\link{git_worktree}}
}
\concept{git}
\section{Related libgit2 documentation}{\href{https://libgit2.org/docs/reference/main/revert/index.html}{\code{revert}}.}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rebase.R
\name{git_rebase}
\alias{git_rebase}
\alias{git_rebase_list}
\alias{git_rebase_commit}
\alias{git_cherry_pick}
\alias{git_ahead_behind}
\title{Cherry-Pick and Rebase}
\usage{
git_rebase_list(upstream = NULL, repo = ".")

git_rebase_commit(upstream = NULL, repo = ".")

git_cherry_pick(commit, repo = ".")

git_ahead_behind(upstream = NULL, ref = "HEAD", repo = ".")
}
\arguments{
\item{upstream}{branch to which you want to rewind and re-apply your
local commits. The default uses the remote upstream branch with the
current state on the git server, simulating \code{\link[=git_pull]{git_pull()}}.}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \code{\link[=git_find]{git_find()}}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{commit}{id of the commit to cherry pick}

\item{ref}{string with a branch/tag/commit}
}
\description{
\itemize{
\item \code{git_cherry_pick()} applies the changes from a given commit (from another branch)
onto the current branch.
*\code{git_rebase_commit()} resets the branch to the state of another branch (upstream)
and then re-applies your local changes by cherry-picking each of your local
commits onto the upstream commit history.
*\code{git_rebase_list()} shows your local commits that are missing from the \code{upstream}
history, and if they conflict with upstream changes.
}
}
\details{
To find if your local commits are missing from \code{upstream},
\code{git_rebase_list()} first performs a rebase dry-run, without committing
anything. If there are no conflicts, you can use \code{git_rebase_commit()}
to rewind and rebase your branch onto \code{upstream}.

Gert only support a clean rebase; it never leaves the repository in unfinished
"rebasing" state. If conflicts arise, \code{git_rebase_commit()} will raise an error
without making changes.
}
\seealso{
Other git:
\code{\link{git_archive}},
\code{\link[=git_branch]{git_branch()}},
\code{\link[=git_commit]{git_commit()}},
\code{\link[=git_config]{git_config()}},
\code{\link[=git_diff]{git_diff()}},
\code{\link[=git_fetch]{git_fetch()}},
\code{\link{git_history}},
\code{\link{git_ignore}},
\code{\link[=git_merge]{git_merge()}},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link[=git_reset]{git_reset()}},
\code{\link[=git_restore]{git_restore()}},
\code{\link[=git_revert]{git_revert()}},
\code{\link[=git_signature]{git_signature()}},
\code{\link{git_stash}},
\code{\link{git_tag}},
\code{\link{git_worktree}}
}
\concept{git}
\section{Related libgit2 documentation}{\href{https://libgit2.org/docs/reference/main/rebase/index.html}{\code{rebase}}, \href{https://libgit2.org/docs/reference/main/cherrypick/index.html}{\code{cherrypick}}, \href{https://libgit2.org/docs/reference/main/graph/index.html}{\code{graph}}.}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/signature.R
\name{git_signature}
\alias{git_signature}
\alias{git_signature_default}
\alias{git_signature_parse}
\title{Author Signature}
\usage{
git_signature_default(repo = ".")

git_signature(name, email, time = NULL)

git_signature_parse(sig)
}
\arguments{
\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{name}{Real name of the committer}

\item{email}{Email address of the committer}

\item{time}{timestamp of class POSIXt or NULL}

\item{sig}{string in proper \code{"First Last <your@email.com>"} format, see details.}
}
\description{
A signature contains the author and timestamp of a commit. Each commit
includes a signature of the author and committer (which can be identical).
}
\details{
A signature string has format \code{"Real Name <email> timestamp tzoffset"}. The
\verb{timestamp tzoffset} piece can be omitted in which case the current local
time is used. If not omitted, \code{timestamp} must contain the number
of seconds since the Unix epoch and \code{tzoffset} is the timezone offset in
\code{hhmm} format (note the lack of a colon separator)
}
\examples{
# Your default user
try(git_signature_default())

# Specify explicit name and email
git_signature("Some committer", "sarah@gmail.com")

# Create signature for an hour ago
(sig <- git_signature("Han", "han@company.com", Sys.time() - 3600))

# Parse a signature
git_signature_parse(sig)
git_signature_parse("Emma <emma@mu.edu>")
}
\seealso{
Other git: 
\code{\link{git_archive}},
\code{\link{git_branch}()},
\code{\link{git_commit}()},
\code{\link{git_config}()},
\code{\link{git_diff}()},
\code{\link{git_fetch}()},
\code{\link{git_merge}()},
\code{\link{git_rebase}()},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link{git_stash}},
\code{\link{git_tag}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repository.R
\name{git_repo}
\alias{git_repo}
\alias{git_init}
\alias{git_find}
\alias{git_info}
\title{Create or discover a local Git repository}
\usage{
git_init(path = ".", bare = FALSE)

git_find(path = ".")

git_info(repo = ".")
}
\arguments{
\item{path}{the location of the git repository, see details.}

\item{bare}{if true, a Git repository without a working directory is created}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}
}
\value{
The path to the Git repository.
}
\description{
Use \code{git_init()} to create a new repository or \code{git_find()} to discover an
existing local repository. \code{git_info()} shows basic information about a
repository, such as the SHA and branch of the current HEAD.
}
\details{
For \code{git_init()} the \code{path} parameter sets the directory of the git repository
to create. If this directory already exists, it must be empty. If it does
not exist, it is created, along with any intermediate directories that don't
yet exist. For \code{git_find()} the \code{path} arguments specifies the directory at
which to start the search for a git repository. If it is not a git repository
itself, then its parent directory is consulted, then the parent's parent, and
so on.
}
\examples{
# directory does not yet exist
r <- tempfile(pattern = "gert")
git_init(r)
git_find(r)

# create a child directory, then a grandchild, then search
r_grandchild_dir <- file.path(r, "aaa", "bbb")
dir.create(r_grandchild_dir, recursive = TRUE)
git_find(r_grandchild_dir)

# cleanup
unlink(r, recursive = TRUE)

# directory exists but is empty
r <- tempfile(pattern = "gert")
dir.create(r)
git_init(r)
git_find(r)

# cleanup
unlink(r, recursive = TRUE)
}
\seealso{
Other git: 
\code{\link{git_archive}},
\code{\link{git_branch}()},
\code{\link{git_commit}()},
\code{\link{git_config}()},
\code{\link{git_diff}()},
\code{\link{git_fetch}()},
\code{\link{git_ignore}},
\code{\link{git_merge}()},
\code{\link{git_rebase}()},
\code{\link{git_remote}},
\code{\link{git_signature}()},
\code{\link{git_stash}},
\code{\link{git_tag}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pr.R
\name{git_checkout_pull_request}
\alias{git_checkout_pull_request}
\alias{git_fetch_pull_requests}
\title{GitHub Wrappers}
\usage{
git_checkout_pull_request(pr = 1, remote = NULL, repo = ".")

git_fetch_pull_requests(pr = "*", remote = NULL, repo = ".")
}
\arguments{
\item{pr}{number with PR to fetch or check out. Use \code{"*"} to fetch all
pull requests.}

\item{remote}{Optional. Name of a remote listed in \code{\link[=git_remote_list]{git_remote_list()}}. If
unspecified and the current branch is already tracking branch a remote
branch, that remote is honored. Otherwise, defaults to \code{origin}.}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \code{\link[=git_find]{git_find()}}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}
}
\description{
Fetch and checkout pull requests.
}
\details{
By default \code{git_fetch_pull_requests()} will download all PR branches. To
remove these again simply use \link[=git_fetch]{git_fetch(prune = TRUE)}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/signature.R
\name{git_signature}
\alias{git_signature}
\alias{git_signature_default}
\alias{git_signature_parse}
\title{Author Signature}
\usage{
git_signature_default(repo = ".")

git_signature(name, email, time = NULL)

git_signature_parse(sig)
}
\arguments{
\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{name}{Real name of the committer}

\item{email}{Email address of the committer}

\item{time}{timestamp of class POSIXt or NULL}

\item{sig}{string in proper \code{"First Last <your@email.com>"} format, see details.}
}
\description{
A signature contains the author and timestamp of a commit. Each commit
includes a signature of the author and committer (which can be identical).
}
\details{
A signature string has format \code{"Real Name <email> timestamp tzoffset"}. The
\verb{timestamp tzoffset} piece can be omitted in which case the current local
time is used. If not omitted, \code{timestamp} must contain the number
of seconds since the Unix epoch and \code{tzoffset} is the timezone offset in
\code{hhmm} format (note the lack of a colon separator)
}
\examples{
# Your default user
try(git_signature_default())

# Specify explicit name and email
git_signature("Some committer", "sarah@gmail.com")

# Create signature for an hour ago
(sig <- git_signature("Han", "han@company.com", Sys.time() - 3600))

# Parse a signature
git_signature_parse(sig)
git_signature_parse("Emma <emma@mu.edu>")
}
\seealso{
Other git: 
\code{\link{git_archive}},
\code{\link{git_branch}()},
\code{\link{git_commit}()},
\code{\link{git_config}()},
\code{\link{git_diff}()},
\code{\link{git_fetch}()},
\code{\link{git_ignore}},
\code{\link{git_merge}()},
\code{\link{git_rebase}()},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link{git_stash}},
\code{\link{git_tag}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/branch.R
\name{git_branch}
\alias{git_branch}
\alias{git_branch_list}
\alias{git_branch_checkout}
\alias{git_branch_create}
\alias{git_branch_delete}
\alias{git_branch_move}
\alias{git_branch_fast_forward}
\alias{git_branch_set_upstream}
\alias{git_branch_exists}
\title{Git Branch}
\usage{
git_branch(repo = ".")

git_branch_list(local = NULL, repo = ".")

git_branch_checkout(branch, force = FALSE, orphan = FALSE, repo = ".")

git_branch_create(branch, ref = "HEAD", checkout = TRUE, repo = ".")

git_branch_delete(branch, repo = ".")

git_branch_move(branch, new_branch, force = FALSE, repo = ".")

git_branch_fast_forward(ref, repo = ".")

git_branch_set_upstream(upstream, branch = git_branch(repo), repo = ".")

git_branch_exists(branch, local = TRUE, repo = ".")
}
\arguments{
\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{local}{set TRUE to only check for local branches, FALSE to check for remote
branches. Use NULL to return all branches.}

\item{branch}{name of branch to check out}

\item{force}{ignore conflicts and overwrite modified files}

\item{orphan}{if branch does not exist, checkout unborn branch}

\item{ref}{string with a branch/tag/commit}

\item{checkout}{move HEAD to the newly created branch}

\item{new_branch}{target name of the branch once the move is performed; this name is validated for consistency.}

\item{upstream}{remote branch from \link{git_branch_list}, for example \code{"origin/master"}}
}
\description{
Create, list, and checkout branches.
}
\seealso{
Other git: 
\code{\link{git_archive}},
\code{\link{git_commit}()},
\code{\link{git_config}()},
\code{\link{git_diff}()},
\code{\link{git_fetch}()},
\code{\link{git_merge}()},
\code{\link{git_rebase}()},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link{git_signature}()},
\code{\link{git_stash}},
\code{\link{git_tag}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tag.R
\name{git_tag}
\alias{git_tag}
\alias{git_tag_list}
\alias{git_tag_create}
\alias{git_tag_delete}
\alias{git_tag_push}
\title{Git Tag}
\usage{
git_tag_list(match = "*", repo = ".")

git_tag_create(name, message, ref = "HEAD", repo = ".")

git_tag_delete(name, repo = ".")

git_tag_push(name, ..., repo = ".")
}
\arguments{
\item{match}{pattern to filter tags (use \code{*} for wildcard)}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{name}{tag name}

\item{message}{tag message}

\item{ref}{target reference to tag}

\item{...}{other arguments passed to \link{git_push}}
}
\description{
Create and list tags.
}
\seealso{
Other git: 
\code{\link{git_archive}},
\code{\link{git_branch}()},
\code{\link{git_commit}()},
\code{\link{git_config}()},
\code{\link{git_diff}()},
\code{\link{git_fetch}()},
\code{\link{git_merge}()},
\code{\link{git_rebase}()},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link{git_signature}()},
\code{\link{git_stash}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ignore.R
\name{git_ignore}
\alias{git_ignore}
\alias{git_ignore_path_is_ignored}
\title{Git Ignore}
\usage{
git_ignore_path_is_ignored(path, repo = ".")
}
\arguments{
\item{path}{A character vector of paths to test within the repo}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}
}
\value{
A logical vector the same length as \code{path}, indicating if the
paths would be ignored.
}
\description{
Test if files would be ignored by \code{.gitignore} rules
}
\seealso{
Other git: 
\code{\link{git_archive}},
\code{\link{git_branch}()},
\code{\link{git_commit}()},
\code{\link{git_config}()},
\code{\link{git_diff}()},
\code{\link{git_fetch}()},
\code{\link{git_merge}()},
\code{\link{git_rebase}()},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link{git_signature}()},
\code{\link{git_stash}},
\code{\link{git_tag}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diff.R
\name{git_diff}
\alias{git_diff}
\alias{git_diff_patch}
\title{Git Diff}
\usage{
git_diff(ref = NULL, repo = ".")

git_diff_patch(ref = NULL, repo = ".")
}
\arguments{
\item{ref}{a reference such as \code{"HEAD"}, or a commit id, or \code{NULL}
to the diff the working directory against the repository index.}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}
}
\description{
View changes in a commit or in the current working directory.
}
\seealso{
Other git: 
\code{\link{git_archive}},
\code{\link{git_branch}()},
\code{\link{git_commit}()},
\code{\link{git_config}()},
\code{\link{git_fetch}()},
\code{\link{git_merge}()},
\code{\link{git_rebase}()},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link{git_signature}()},
\code{\link{git_stash}},
\code{\link{git_tag}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ignore.R
\name{git_ignore}
\alias{git_ignore}
\alias{git_ignore_path_is_ignored}
\title{Git Ignore}
\usage{
git_ignore_path_is_ignored(path, repo = ".")
}
\arguments{
\item{path}{A character vector of paths to test within the repo}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}
}
\value{
A logical vector the same length as \code{path}, indicating if the
paths would be ignored.
}
\description{
Test if files would be ignored by \code{.gitignore} rules
}
\seealso{
Other git: 
\code{\link{git_archive}},
\code{\link{git_branch}()},
\code{\link{git_commit}()},
\code{\link{git_config}()},
\code{\link{git_diff}()},
\code{\link{git_fetch}()},
\code{\link{git_merge}()},
\code{\link{git_rebase}()},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link{git_reset}()},
\code{\link{git_signature}()},
\code{\link{git_stash}},
\code{\link{git_tag}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/branch.R
\name{git_branch}
\alias{git_branch}
\alias{git_branch_list}
\alias{git_branch_checkout}
\alias{git_branch_create}
\alias{git_branch_delete}
\alias{git_branch_move}
\alias{git_branch_fast_forward}
\alias{git_branch_set_upstream}
\alias{git_branch_exists}
\title{Git Branch}
\usage{
git_branch(repo = ".")

git_branch_list(local = NULL, repo = ".")

git_branch_checkout(branch, force = FALSE, orphan = FALSE, repo = ".")

git_branch_create(branch, ref = "HEAD", checkout = TRUE, repo = ".")

git_branch_delete(branch, repo = ".")

git_branch_move(branch, new_branch, force = FALSE, repo = ".")

git_branch_fast_forward(ref, repo = ".")

git_branch_set_upstream(upstream, branch = git_branch(repo), repo = ".")

git_branch_exists(branch, local = TRUE, repo = ".")
}
\arguments{
\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{local}{set TRUE to only check for local branches, FALSE to check for remote
branches. Use NULL to return all branches.}

\item{branch}{name of branch to check out}

\item{force}{ignore conflicts and overwrite modified files}

\item{orphan}{if branch does not exist, checkout unborn branch}

\item{ref}{string with a branch/tag/commit}

\item{checkout}{move HEAD to the newly created branch}

\item{new_branch}{target name of the branch once the move is performed; this name is validated for consistency.}

\item{upstream}{remote branch from \link{git_branch_list}, for example \code{"origin/master"}}
}
\description{
Create, list, and checkout branches.
}
\seealso{
Other git: 
\code{\link{git_archive}},
\code{\link{git_commit}()},
\code{\link{git_config}()},
\code{\link{git_diff}()},
\code{\link{git_fetch}()},
\code{\link{git_ignore}},
\code{\link{git_merge}()},
\code{\link{git_rebase}()},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link{git_signature}()},
\code{\link{git_stash}},
\code{\link{git_tag}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tag.R
\name{git_tag}
\alias{git_tag}
\alias{git_tag_list}
\alias{git_tag_create}
\alias{git_tag_delete}
\alias{git_tag_push}
\title{Git Tag}
\usage{
git_tag_list(match = "*", repo = ".")

git_tag_create(name, message, ref = "HEAD", repo = ".")

git_tag_delete(name, repo = ".")

git_tag_push(name, ..., repo = ".")
}
\arguments{
\item{match}{pattern to filter tags (use \code{*} for wildcard)}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{name}{tag name}

\item{message}{tag message}

\item{ref}{target reference to tag}

\item{...}{other arguments passed to \link{git_push}}
}
\description{
Create and list tags.
}
\seealso{
Other git: 
\code{\link{git_archive}},
\code{\link{git_branch}()},
\code{\link{git_commit}()},
\code{\link{git_config}()},
\code{\link{git_diff}()},
\code{\link{git_fetch}()},
\code{\link{git_ignore}},
\code{\link{git_merge}()},
\code{\link{git_rebase}()},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link{git_signature}()},
\code{\link{git_stash}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/archive.R
\name{git_archive}
\alias{git_archive}
\alias{git_archive_zip}
\title{Git Archive}
\usage{
git_archive_zip(file = NULL, repo = ".")
}
\arguments{
\item{file}{name of the output zip file. Default is returned
by the function}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}
}
\value{
path to the zip file that was created
}
\description{
Exports the files in your repository to a zip file that
is returned by the function.
}
\seealso{
Other git: 
\code{\link{git_branch}()},
\code{\link{git_commit}()},
\code{\link{git_config}()},
\code{\link{git_diff}()},
\code{\link{git_fetch}()},
\code{\link{git_merge}()},
\code{\link{git_rebase}()},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link{git_signature}()},
\code{\link{git_stash}},
\code{\link{git_tag}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diff.R
\name{git_diff}
\alias{git_diff}
\alias{git_diff_patch}
\title{Git Diff}
\usage{
git_diff(ref = NULL, repo = ".")

git_diff_patch(ref = NULL, repo = ".")
}
\arguments{
\item{ref}{a reference such as \code{"HEAD"}, or a commit id, or \code{NULL}
to the diff the working directory against the repository index.}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}
}
\description{
View changes in a commit or in the current working directory.
}
\seealso{
Other git: 
\code{\link{git_archive}},
\code{\link{git_branch}()},
\code{\link{git_commit}()},
\code{\link{git_config}()},
\code{\link{git_fetch}()},
\code{\link{git_ignore}},
\code{\link{git_merge}()},
\code{\link{git_rebase}()},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link{git_signature}()},
\code{\link{git_stash}},
\code{\link{git_tag}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/signature.R
\name{git_signature}
\alias{git_signature}
\alias{git_signature_default}
\alias{git_signature_parse}
\title{Author Signature}
\usage{
git_signature_default(repo = ".")

git_signature(name, email, time = NULL)

git_signature_parse(sig)
}
\arguments{
\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \code{\link[=git_find]{git_find()}}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{name}{Real name of the committer}

\item{email}{Email address of the committer}

\item{time}{timestamp of class POSIXt or NULL}

\item{sig}{string in proper \code{"First Last <your@email.com>"} format, see details.}
}
\description{
A signature contains the author and timestamp of a commit. Each commit
includes a signature of the author and committer (which can be identical).
}
\details{
A signature string has format \code{"Real Name <email> timestamp tzoffset"}. The
\verb{timestamp tzoffset} piece can be omitted in which case the current local
time is used. If not omitted, \code{timestamp} must contain the number
of seconds since the Unix epoch and \code{tzoffset} is the timezone offset in
\code{hhmm} format (note the lack of a colon separator)
}
\examples{
# Your default user
try(git_signature_default())

# Specify explicit name and email
git_signature("Some committer", "sarah@gmail.com")

# Create signature for an hour ago
(sig <- git_signature("Han", "han@company.com", Sys.time() - 3600))

# Parse a signature
git_signature_parse(sig)
git_signature_parse("Emma <emma@mu.edu>")
}
\seealso{
Other git:
\code{\link{git_archive}},
\code{\link[=git_branch]{git_branch()}},
\code{\link[=git_commit]{git_commit()}},
\code{\link[=git_config]{git_config()}},
\code{\link[=git_diff]{git_diff()}},
\code{\link[=git_fetch]{git_fetch()}},
\code{\link{git_history}},
\code{\link{git_ignore}},
\code{\link[=git_merge]{git_merge()}},
\code{\link[=git_rebase]{git_rebase()}},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link[=git_reset]{git_reset()}},
\code{\link[=git_restore]{git_restore()}},
\code{\link[=git_revert]{git_revert()}},
\code{\link{git_stash}},
\code{\link{git_tag}},
\code{\link{git_worktree}}
}
\concept{git}
\section{Related libgit2 documentation}{\href{https://libgit2.org/docs/reference/main/signature/index.html}{\code{signature}}.}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.R
\name{user_is_configured}
\alias{user_is_configured}
\title{Test if a Git user is configured}
\usage{
user_is_configured(repo = ".")
}
\arguments{
\item{repo}{An optional \code{repo}, in the sense of \code{\link[=git_open]{git_open()}}.}
}
\value{
\code{TRUE} if \code{user.name} and \code{user.email} are set locally or globally,
\code{FALSE} otherwise.
}
\description{
This function exists mostly to guard examples that rely on having a user
configured, in order to make commits. \code{user_is_configured()} makes no
distinction between local or global user config.
}
\examples{
user_is_configured()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/archive.R
\name{git_archive}
\alias{git_archive}
\alias{git_archive_zip}
\title{Git Archive}
\usage{
git_archive_zip(file = NULL, repo = ".")
}
\arguments{
\item{file}{name of the output zip file. Default is returned
by the function}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}
}
\value{
path to the zip file that was created
}
\description{
Exports the files in your repository to a zip file that
is returned by the function.
}
\seealso{
Other git: 
\code{\link{git_branch}()},
\code{\link{git_commit}()},
\code{\link{git_config}()},
\code{\link{git_diff}()},
\code{\link{git_fetch}()},
\code{\link{git_ignore}},
\code{\link{git_merge}()},
\code{\link{git_rebase}()},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link{git_signature}()},
\code{\link{git_stash}},
\code{\link{git_tag}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fetch.R
\name{git_fetch}
\alias{git_fetch}
\alias{git_remote_ls}
\alias{git_push}
\alias{git_clone}
\alias{git_pull}
\title{Push and pull}
\usage{
git_fetch(
  remote = NULL,
  refspec = NULL,
  password = askpass,
  ssh_key = NULL,
  prune = FALSE,
  verbose = interactive(),
  repo = "."
)

git_remote_ls(
  remote = NULL,
  password = askpass,
  ssh_key = NULL,
  verbose = interactive(),
  repo = "."
)

git_push(
  remote = NULL,
  refspec = NULL,
  set_upstream = NULL,
  password = askpass,
  ssh_key = NULL,
  mirror = FALSE,
  force = FALSE,
  verbose = interactive(),
  repo = "."
)

git_clone(
  url,
  path = NULL,
  branch = NULL,
  password = askpass,
  ssh_key = NULL,
  bare = FALSE,
  mirror = FALSE,
  verbose = interactive()
)

git_pull(remote = NULL, rebase = FALSE, ..., repo = ".")
}
\arguments{
\item{remote}{Optional. Name of a remote listed in \code{\link[=git_remote_list]{git_remote_list()}}. If
unspecified and the current branch is already tracking branch a remote
branch, that remote is honored. Otherwise, defaults to \code{origin}.}

\item{refspec}{string with mapping between remote and local refs. Default
uses the default refspec from the remote, which usually fetches all branches.}

\item{password}{a string or a callback function to get passwords for authentication
or password protected ssh keys. Defaults to \link[askpass:askpass]{askpass} which
checks \code{getOption('askpass')}.}

\item{ssh_key}{path or object containing your ssh private key. By default we
look for keys in \code{ssh-agent} and \link[credentials:ssh_credentials]{credentials::ssh_key_info}.}

\item{prune}{delete tracking branches that no longer exist on the remote, or
are not in the refspec (such as pull requests).}

\item{verbose}{display some progress info while downloading}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{set_upstream}{change the branch default upstream to \code{remote}.
If \code{NULL}, this will set the branch upstream only if the push was
successful and if the branch does not have an upstream set yet.}

\item{mirror}{use the \code{--mirror} flag}

\item{force}{use the \code{--force} flag}

\item{url}{remote url. Typically starts with \verb{https://github.com/} for public
repositories, and \verb{https://yourname@github.com/} or \verb{git@github.com/} for
private repos. You will be prompted for a password or pat when needed.}

\item{path}{Directory of the Git repository to create.}

\item{branch}{name of branch to check out locally}

\item{bare}{use the \code{--bare} flag}

\item{rebase}{if TRUE we try to rebase instead of merge local changes. This
is not possible in case of conflicts (you will get an error).}

\item{...}{arguments passed to \link{git_fetch}}
}
\description{
Functions to connect with a git server (remote) to fetch or push changes.
The 'credentials' package is used to handle authentication, the
\href{https://docs.ropensci.org/credentials/articles/intro.html}{credentials vignette}
explains the various authentication methods for SSH and HTTPS remotes.
}
\details{
Use \code{\link[=git_fetch]{git_fetch()}} and \code{\link[=git_push]{git_push()}} to sync a local branch with a remote
branch. Here \code{\link[=git_pull]{git_pull()}} is a wrapper for \code{\link[=git_fetch]{git_fetch()}} which then tries to
\link[=git_branch_fast_forward]{fast-forward} the local branch after fetching.
}
\examples{
{# Clone a small repository
git_dir <- file.path(tempdir(), 'antiword')
git_clone('https://github.com/ropensci/antiword', git_dir)

# Change into the repo directory
olddir <- getwd()
setwd(git_dir)

# Show some stuff
git_log()
git_branch_list()
git_remote_list()

# Add a file
write.csv(iris, 'iris.csv')
git_add('iris.csv')

# Commit the change
jerry <- git_signature("Jerry", "jerry@hotmail.com")
git_commit('added the iris file', author = jerry)

# Now in the log:
git_log()

# Cleanup
setwd(olddir)
unlink(git_dir, recursive = TRUE)
}
}
\seealso{
Other git: 
\code{\link{git_archive}},
\code{\link{git_branch}()},
\code{\link{git_commit}()},
\code{\link{git_config}()},
\code{\link{git_diff}()},
\code{\link{git_merge}()},
\code{\link{git_rebase}()},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link{git_signature}()},
\code{\link{git_stash}},
\code{\link{git_tag}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/branch.R
\name{git_branch}
\alias{git_branch}
\alias{git_branch_list}
\alias{git_branch_checkout}
\alias{git_branch_switch}
\alias{git_branch_create}
\alias{git_branch_delete}
\alias{git_branch_move}
\alias{git_branch_fast_forward}
\alias{git_branch_set_upstream}
\alias{git_branch_exists}
\title{Git Branch}
\usage{
git_branch(repo = ".")

git_branch_list(local = NULL, repo = ".")

git_branch_checkout(branch, force = FALSE, orphan = FALSE, repo = ".")

git_branch_switch(branch, force = FALSE, orphan = FALSE, repo = ".")

git_branch_create(
  branch,
  ref = "HEAD",
  checkout = TRUE,
  force = FALSE,
  repo = "."
)

git_branch_delete(branch, repo = ".")

git_branch_move(branch, new_branch, force = FALSE, repo = ".")

git_branch_fast_forward(ref, repo = ".")

git_branch_set_upstream(upstream, branch = git_branch(repo), repo = ".")

git_branch_exists(branch, local = TRUE, repo = ".")
}
\arguments{
\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \code{\link[=git_find]{git_find()}}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{local}{set TRUE to only check for local branches, FALSE to check for remote
branches. Use NULL to return all branches.}

\item{branch}{name of branch to check out}

\item{force}{overwrite existing branch}

\item{orphan}{if branch does not exist, checkout unborn branch}

\item{ref}{string with a branch/tag/commit}

\item{checkout}{move HEAD to the newly created branch}

\item{new_branch}{target name of the branch once the move is performed; this name is validated for consistency.}

\item{upstream}{remote branch from \link{git_branch_list}, for example \code{"origin/master"}}
}
\description{
Create, list, and checkout branches.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
# Creating a branch
repo <- file.path(tempdir(), "myrepo")
git_init(repo)
writeLines("hello", file.path(repo, "hello.txt"))
git_add("hello.txt", repo = repo)
first_commit <- git_commit("First commit", repo = repo)

git_branch_create("new-feat", repo = repo)
writeLines("world", file.path(repo, "hello.txt"))
git_add("hello.txt", repo = repo)
second_commit <- git_commit("Second commit", repo = repo)

# Changing branches
git_branch(repo = repo)
git_branch_switch("main", repo = repo)
git_branch(repo = repo)

git_branch_exists("new-feat", repo = repo)

# Listing branches
git_branch_list(repo = repo)

# Renaming a branch
git_branch_move("new-feat", "better-name", repo = repo)
git_branch_exists("new-feat", repo = repo)
git_branch_list(repo = repo)

# Deleting a branch
git_branch_delete("better-name", repo = repo)
git_branch_exists("better-name", repo = repo)

# clean up
unlink(repo, recursive = TRUE)
# ------------------------
# Creating an orphan branch
repo <- file.path(tempdir(), "myrepo")
git_init(repo)

# Set a user if no default
if (!user_is_configured()) {
  git_config_set("user.name", "Jerry")
  git_config_set("user.email", "jerry@gmail.com")
}
writeLines("hello", file.path(repo, "hello.txt"))
git_add("hello.txt", repo = repo)
first_commit <- git_commit("First commit", repo = repo)

writeLines("world", file.path(repo, "hello.txt"))
git_add("hello.txt", repo = repo)
second_commit <- git_commit("Second commit", repo = repo)

# Check out first commit
git_branch_checkout(first_commit, orphan = TRUE, repo = repo)
git_status(repo)
# clean up
unlink(repo, recursive = TRUE)
\dontshow{\}) # examplesIf}
}
\seealso{
Other git:
\code{\link{git_archive}},
\code{\link[=git_commit]{git_commit()}},
\code{\link[=git_config]{git_config()}},
\code{\link[=git_diff]{git_diff()}},
\code{\link[=git_fetch]{git_fetch()}},
\code{\link{git_history}},
\code{\link{git_ignore}},
\code{\link[=git_merge]{git_merge()}},
\code{\link[=git_rebase]{git_rebase()}},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link[=git_reset]{git_reset()}},
\code{\link[=git_restore]{git_restore()}},
\code{\link[=git_revert]{git_revert()}},
\code{\link[=git_signature]{git_signature()}},
\code{\link{git_stash}},
\code{\link{git_tag}},
\code{\link{git_worktree}}
}
\concept{git}
\section{Related libgit2 documentation}{\href{https://libgit2.org/docs/reference/main/branch/index.html}{\code{branch}}, \href{https://libgit2.org/docs/reference/main/checkout/index.html}{\code{checkout}}.}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stash.R
\name{git_stash}
\alias{git_stash}
\alias{git_stash_save}
\alias{git_stash_pop}
\alias{git_stash_drop}
\alias{git_stash_list}
\title{Stashing changes}
\usage{
git_stash_save(
  message = "",
  keep_index = FALSE,
  include_untracked = FALSE,
  include_ignored = FALSE,
  repo = "."
)

git_stash_pop(index = 0, repo = ".")

git_stash_drop(index = 0, repo = ".")

git_stash_list(repo = ".")
}
\arguments{
\item{message}{optional message to store the stash}

\item{keep_index}{changes already added to the index are left intact in
the working directory}

\item{include_untracked}{untracked files are also stashed and then
cleaned up from the working directory}

\item{include_ignored}{ignored files are also stashed and then cleaned
up from the working directory}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{index}{The position within the stash list. 0 points to the
most recent stashed state.}
}
\description{
Temporary stash away changed from the working directory.
}
\seealso{
Other git: 
\code{\link{git_archive}},
\code{\link{git_branch}()},
\code{\link{git_commit}()},
\code{\link{git_config}()},
\code{\link{git_diff}()},
\code{\link{git_fetch}()},
\code{\link{git_merge}()},
\code{\link{git_rebase}()},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link{git_signature}()},
\code{\link{git_tag}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tag.R
\name{git_tag}
\alias{git_tag}
\alias{git_tag_list}
\alias{git_tag_create}
\alias{git_tag_delete}
\alias{git_tag_push}
\title{Git Tag}
\usage{
git_tag_list(match = "*", repo = ".")

git_tag_create(name, message, ref = "HEAD", repo = ".")

git_tag_delete(name, repo = ".")

git_tag_push(name, ..., repo = ".")
}
\arguments{
\item{match}{pattern to filter tags (use \code{*} for wildcard)}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \code{\link[=git_find]{git_find()}}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{name}{tag name}

\item{message}{tag message}

\item{ref}{target reference to tag}

\item{...}{other arguments passed to \code{\link[=git_push]{git_push()}}}
}
\description{
Create and list tags.
}
\seealso{
Other git:
\code{\link{git_archive}},
\code{\link[=git_branch]{git_branch()}},
\code{\link[=git_commit]{git_commit()}},
\code{\link[=git_config]{git_config()}},
\code{\link[=git_diff]{git_diff()}},
\code{\link[=git_fetch]{git_fetch()}},
\code{\link{git_history}},
\code{\link{git_ignore}},
\code{\link[=git_merge]{git_merge()}},
\code{\link[=git_rebase]{git_rebase()}},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link[=git_reset]{git_reset()}},
\code{\link[=git_restore]{git_restore()}},
\code{\link[=git_revert]{git_revert()}},
\code{\link[=git_signature]{git_signature()}},
\code{\link{git_stash}},
\code{\link{git_worktree}}
}
\concept{git}
\section{Related libgit2 documentation}{\href{https://libgit2.org/docs/reference/main/tag/index.html}{\code{tag}}.}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/remotes.R
\name{git_remote}
\alias{git_remote}
\alias{git_remote_list}
\alias{git_remote_add}
\alias{git_remote_remove}
\alias{git_remote_info}
\alias{git_remote_set_url}
\alias{git_remote_set_pushurl}
\alias{git_remote_refspecs}
\title{Git Remotes}
\usage{
git_remote_list(repo = ".")

git_remote_add(url, name = "origin", refspec = NULL, repo = ".")

git_remote_remove(remote, repo = ".")

git_remote_info(remote = NULL, repo = ".")

git_remote_set_url(url, remote = NULL, repo = ".")

git_remote_set_pushurl(url, remote = NULL, repo = ".")

git_remote_refspecs(remote = NULL, repo = ".")
}
\arguments{
\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{url}{server url (https or ssh)}

\item{name}{unique name for the new remote}

\item{refspec}{optional string with the remote fetch value}

\item{remote}{name of an existing remote. Default \code{NULL} means the remote
from the upstream of the current branch.}
}
\description{
List, add, configure, or remove remotes.
}
\seealso{
Other git: 
\code{\link{git_archive}},
\code{\link{git_branch}()},
\code{\link{git_commit}()},
\code{\link{git_config}()},
\code{\link{git_diff}()},
\code{\link{git_fetch}()},
\code{\link{git_merge}()},
\code{\link{git_rebase}()},
\code{\link{git_repo}},
\code{\link{git_signature}()},
\code{\link{git_stash}},
\code{\link{git_tag}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.R
\name{git_config}
\alias{git_config}
\alias{git_config_global}
\alias{git_config_set}
\alias{git_config_global_set}
\title{Get or set Git configuration}
\usage{
git_config(repo = ".")

git_config_global()

git_config_set(name, value, repo = ".")

git_config_global_set(name, value)
}
\arguments{
\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{name}{Name of the option to set}

\item{value}{Value to set. Must be a string, logical, number or \code{NULL} (to
unset).}
}
\value{
\itemize{
\item \code{git_config()}: a \code{data.frame} of the Git options "in force" in the context
of \code{repo}, one row per option. The \code{level} column reveals whether the
option is determined from global or local config.
\item \code{git_config_global()}: a \code{data.frame}, as for \code{git_config()}, except only
for global Git options.
\item \code{git_config_set()}, \code{git_config_global_set()}: The previous value of
\code{name} in local or global config, respectively. If this option was
previously unset, returns \code{NULL}. Returns invisibly.
}
}
\description{
Get or set Git options, as \verb{git config} does on the command line. \strong{Global}
settings affect all of a user's Git operations (\verb{git config --global}),
whereas \strong{local} settings are scoped to a specific repository (\verb{git config --local}). When both exist, local options always win. Four functions address
the four possible combinations of getting vs setting and global vs. local.\tabular{lll}{
    \tab \strong{local} \tab \strong{global} \cr
   get \tab \code{git_config()} \tab \code{git_config_global()} \cr
   set \tab \code{git_config_set()} \tab \code{git_config_global_set()} \cr
}
}
\note{
All entries in the \code{name} column are automatically normalised to
lowercase (see
\url{https://libgit2.org/libgit2/#HEAD/type/git_config_entry} for details).
}
\examples{
# Set and inspect a local, custom Git option
r <- file.path(tempdir(), "gert-demo")
git_init(r)

previous <- git_config_set("aaa.bbb", "ccc", repo = r)
previous
cfg <- git_config(repo = r)
subset(cfg, level == "local")
cfg$value[cfg$name == "aaa.bbb"]

previous <- git_config_set("aaa.bbb", NULL, repo = r)
previous
cfg <- git_config(repo = r)
subset(cfg, level == "local")
cfg$value[cfg$name == "aaa.bbb"]

unlink(r, recursive = TRUE)

\dontrun{
# Set global Git options
git_config_global_set("user.name", "Your Name")
git_config_global_set("user.email", "your@email.com")
git_config_global()
}
}
\seealso{
Other git: 
\code{\link{git_archive}},
\code{\link{git_branch}()},
\code{\link{git_commit}()},
\code{\link{git_diff}()},
\code{\link{git_fetch}()},
\code{\link{git_merge}()},
\code{\link{git_rebase}()},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link{git_signature}()},
\code{\link{git_stash}},
\code{\link{git_tag}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stash.R
\name{git_stash}
\alias{git_stash}
\alias{git_stash_save}
\alias{git_stash_pop}
\alias{git_stash_drop}
\alias{git_stash_list}
\title{Stashing changes}
\usage{
git_stash_save(
  message = "",
  keep_index = FALSE,
  include_untracked = FALSE,
  include_ignored = FALSE,
  repo = "."
)

git_stash_pop(index = 0, repo = ".")

git_stash_drop(index = 0, repo = ".")

git_stash_list(repo = ".")
}
\arguments{
\item{message}{optional message to store the stash}

\item{keep_index}{changes already added to the index are left intact in
the working directory}

\item{include_untracked}{untracked files are also stashed and then
cleaned up from the working directory}

\item{include_ignored}{ignored files are also stashed and then cleaned
up from the working directory}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{index}{The position within the stash list. 0 points to the
most recent stashed state.}
}
\description{
Temporary stash away changed from the working directory.
}
\seealso{
Other git: 
\code{\link{git_archive}},
\code{\link{git_branch}()},
\code{\link{git_commit}()},
\code{\link{git_config}()},
\code{\link{git_diff}()},
\code{\link{git_fetch}()},
\code{\link{git_ignore}},
\code{\link{git_merge}()},
\code{\link{git_rebase}()},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link{git_signature}()},
\code{\link{git_tag}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/remotes.R
\name{git_remote}
\alias{git_remote}
\alias{git_remote_list}
\alias{git_remote_add}
\alias{git_remote_remove}
\alias{git_remote_info}
\alias{git_remote_set_url}
\alias{git_remote_set_pushurl}
\alias{git_remote_refspecs}
\title{Git Remotes}
\usage{
git_remote_list(repo = ".")

git_remote_add(url, name = "origin", refspec = NULL, repo = ".")

git_remote_remove(remote, repo = ".")

git_remote_info(remote = NULL, repo = ".")

git_remote_set_url(url, remote = NULL, repo = ".")

git_remote_set_pushurl(url, remote = NULL, repo = ".")

git_remote_refspecs(remote = NULL, repo = ".")
}
\arguments{
\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{url}{server url (https or ssh)}

\item{name}{unique name for the new remote}

\item{refspec}{optional string with the remote fetch value}

\item{remote}{name of an existing remote. Default \code{NULL} means the remote
from the upstream of the current branch.}
}
\description{
List, add, configure, or remove remotes.
}
\seealso{
Other git: 
\code{\link{git_archive}},
\code{\link{git_branch}()},
\code{\link{git_commit}()},
\code{\link{git_config}()},
\code{\link{git_diff}()},
\code{\link{git_fetch}()},
\code{\link{git_ignore}},
\code{\link{git_merge}()},
\code{\link{git_rebase}()},
\code{\link{git_repo}},
\code{\link{git_signature}()},
\code{\link{git_stash}},
\code{\link{git_tag}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/open.R
\name{git_open}
\alias{git_open}
\title{Open local repository}
\usage{
git_open(repo = ".")
}
\arguments{
\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}
}
\value{
an pointer to the libgit2 repository
}
\description{
Returns a pointer to a libgit2 repository object.This function is mainly
for internal use; users should simply reference a repository in gert by
by the path to the directory.
}
\examples{
r <- tempfile(pattern = "gert")
git_init(r)
r_ptr <- git_open(r)
r_ptr
git_open(r_ptr)
git_info(r)

# cleanup
unlink(r, recursive = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.R
\name{user_is_configured}
\alias{user_is_configured}
\title{Test if a Git user is configured}
\usage{
user_is_configured(repo = ".")
}
\arguments{
\item{repo}{An optional \code{repo}, in the sense of \code{\link[=git_open]{git_open()}}.}
}
\value{
\code{TRUE} if \code{user.name} and \code{user.email} are set locally or globally,
\code{FALSE} otherwise.
}
\description{
This function exists mostly to guard examples that rely on having a user
configured, in order to make commits. \code{user_is_configured()} makes no
distinction between local or global user config.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
user_is_configured()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fetch.R
\name{git_fetch}
\alias{git_fetch}
\alias{git_remote_ls}
\alias{git_push}
\alias{git_clone}
\alias{git_pull}
\title{Push and pull}
\usage{
git_fetch(
  remote = NULL,
  refspec = NULL,
  password = askpass,
  ssh_key = NULL,
  prune = FALSE,
  verbose = interactive(),
  repo = "."
)

git_remote_ls(
  remote = NULL,
  password = askpass,
  ssh_key = NULL,
  verbose = interactive(),
  repo = "."
)

git_push(
  remote = NULL,
  refspec = NULL,
  set_upstream = NULL,
  password = askpass,
  ssh_key = NULL,
  mirror = FALSE,
  force = FALSE,
  verbose = interactive(),
  repo = "."
)

git_clone(
  url,
  path = NULL,
  branch = NULL,
  password = askpass,
  ssh_key = NULL,
  bare = FALSE,
  mirror = FALSE,
  depth = 0,
  verbose = interactive()
)

git_pull(remote = NULL, rebase = FALSE, ..., repo = ".")
}
\arguments{
\item{remote}{Optional. Name of a remote listed in \code{\link[=git_remote_list]{git_remote_list()}}. If
unspecified and the current branch is already tracking branch a remote
branch, that remote is honored. Otherwise, defaults to \code{origin}.}

\item{refspec}{string with mapping between remote and local refs. Default
uses the default refspec from the remote, which usually fetches all branches.}

\item{password}{a string or a callback function to get passwords for authentication
or password protected ssh keys. Defaults to \link[askpass:askpass]{askpass} which
checks \code{getOption('askpass')}.}

\item{ssh_key}{path or object containing your ssh private key. By default we
look for keys in \code{ssh-agent} and \code{\link[credentials:ssh_key_info]{credentials::ssh_key_info()}}.}

\item{prune}{delete tracking branches that no longer exist on the remote, or
are not in the refspec (such as pull requests).}

\item{verbose}{display some progress info while downloading}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \code{\link[=git_find]{git_find()}}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{set_upstream}{change the branch default upstream to \code{remote}.
If \code{NULL}, this will set the branch upstream only if the push was
successful and if the branch does not have an upstream set yet.}

\item{mirror}{use the \code{--mirror} flag}

\item{force}{use the \code{--force} flag}

\item{url}{remote url. Typically starts with \verb{https://github.com/} for public
repositories, and \verb{https://yourname@github.com/} or \verb{git@github.com/} for
private repos. You will be prompted for a password or pat when needed.}

\item{path}{Directory of the Git repository to create.
By default, the "humanish" part of the URL.
For instance, "git@github.com:someone/myrepo.git" will be cloned to
\verb{myrepo/}.}

\item{branch}{name of branch to check out locally}

\item{bare}{use the \code{--bare} flag}

\item{depth}{number of commits to fetch when creating a shallow clone. The
default \code{0} clones the entire history. A positive value truncates the history
to the given number of commits from the tip of each branch. Requires
libgit2 >= 1.7.0.}

\item{rebase}{if TRUE we try to rebase instead of merge local changes. This
is not possible in case of conflicts (you will get an error).}

\item{...}{arguments passed to \code{git_fetch()}}
}
\description{
Functions to connect with a git server (remote) to fetch or push changes.
The 'credentials' package is used to handle authentication, the
\href{https://docs.ropensci.org/credentials/articles/intro.html}{credentials vignette}
explains the various authentication methods for SSH and HTTPS remotes.
}
\details{
Use \code{\link[=git_fetch]{git_fetch()}} and \code{\link[=git_push]{git_push()}} to sync a local branch with a remote
branch. Here \code{\link[=git_pull]{git_pull()}} is a wrapper for \code{\link[=git_fetch]{git_fetch()}} which then tries to
\link[=git_branch_fast_forward]{fast-forward} the local branch after fetching.
}
\examples{
{# Clone a small repository
git_dir <- file.path(tempdir(), 'antiword')
git_clone('https://github.com/ropensci/antiword', git_dir)

# Change into the repo directory
olddir <- getwd()
setwd(git_dir)

# Show some stuff
git_log()
git_branch_list()
git_remote_list()

# Add a file
write.csv(iris, 'iris.csv')
git_add('iris.csv')

# Commit the change
jerry <- git_signature("Jerry", "jerry@hotmail.com")
git_commit('added the iris file', author = jerry)

# Now in the log:
git_log()

# Cleanup
setwd(olddir)
unlink(git_dir, recursive = TRUE)
}
}
\seealso{
Other git:
\code{\link{git_archive}},
\code{\link[=git_branch]{git_branch()}},
\code{\link[=git_commit]{git_commit()}},
\code{\link[=git_config]{git_config()}},
\code{\link[=git_diff]{git_diff()}},
\code{\link{git_history}},
\code{\link{git_ignore}},
\code{\link[=git_merge]{git_merge()}},
\code{\link[=git_rebase]{git_rebase()}},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link[=git_reset]{git_reset()}},
\code{\link[=git_restore]{git_restore()}},
\code{\link[=git_revert]{git_revert()}},
\code{\link[=git_signature]{git_signature()}},
\code{\link{git_stash}},
\code{\link{git_tag}},
\code{\link{git_worktree}}
}
\concept{git}
\section{Related libgit2 documentation}{\href{https://libgit2.org/docs/reference/main/remote/index.html}{\code{remote}}, \href{https://libgit2.org/docs/reference/main/clone/index.html}{\code{clone}}.}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.R
\name{git_config}
\alias{git_config}
\alias{git_config_global}
\alias{git_config_set}
\alias{git_config_global_set}
\title{Get or set Git configuration}
\usage{
git_config(repo = ".")

git_config_global()

git_config_set(name, value, repo = ".")

git_config_global_set(name, value)
}
\arguments{
\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{name}{Name of the option to set}

\item{value}{Value to set. Must be a string, logical, number or \code{NULL} (to
unset).}
}
\value{
\itemize{
\item \code{git_config()}: a \code{data.frame} of the Git options "in force" in the context
of \code{repo}, one row per option. The \code{level} column reveals whether the
option is determined from global or local config.
\item \code{git_config_global()}: a \code{data.frame}, as for \code{git_config()}, except only
for global Git options.
\item \code{git_config_set()}, \code{git_config_global_set()}: The previous value of
\code{name} in local or global config, respectively. If this option was
previously unset, returns \code{NULL}. Returns invisibly.
}
}
\description{
Get or set Git options, as \verb{git config} does on the command line. \strong{Global}
settings affect all of a user's Git operations (\verb{git config --global}),
whereas \strong{local} settings are scoped to a specific repository (\verb{git config --local}). When both exist, local options always win. Four functions address
the four possible combinations of getting vs setting and global vs. local.\tabular{lll}{
    \tab \strong{local} \tab \strong{global} \cr
   get \tab \code{git_config()} \tab \code{git_config_global()} \cr
   set \tab \code{git_config_set()} \tab \code{git_config_global_set()} \cr
}
}
\note{
All entries in the \code{name} column are automatically normalised to
lowercase (see
\url{https://libgit2.org/libgit2/#HEAD/type/git_config_entry} for details).
}
\examples{
# Set and inspect a local, custom Git option
r <- file.path(tempdir(), "gert-demo")
git_init(r)

previous <- git_config_set("aaa.bbb", "ccc", repo = r)
previous
cfg <- git_config(repo = r)
subset(cfg, level == "local")
cfg$value[cfg$name == "aaa.bbb"]

previous <- git_config_set("aaa.bbb", NULL, repo = r)
previous
cfg <- git_config(repo = r)
subset(cfg, level == "local")
cfg$value[cfg$name == "aaa.bbb"]

unlink(r, recursive = TRUE)

\dontrun{
# Set global Git options
git_config_global_set("user.name", "Your Name")
git_config_global_set("user.email", "your@email.com")
git_config_global()
}
}
\seealso{
Other git: 
\code{\link{git_archive}},
\code{\link{git_branch}()},
\code{\link{git_commit}()},
\code{\link{git_diff}()},
\code{\link{git_fetch}()},
\code{\link{git_ignore}},
\code{\link{git_merge}()},
\code{\link{git_rebase}()},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link{git_signature}()},
\code{\link{git_stash}},
\code{\link{git_tag}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.R
\name{libgit2_config}
\alias{libgit2_config}
\title{Show libgit2 version and capabilities}
\usage{
libgit2_config()
}
\description{
\code{libgit2_config()} reveals which version of libgit2 gert is using and which
features are supported, such whether you are able to use ssh remotes.
}
\examples{
libgit2_config()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge.R
\name{git_merge}
\alias{git_merge}
\alias{git_merge_stage_only}
\alias{git_merge_find_base}
\alias{git_merge_analysis}
\alias{git_merge_abort}
\title{Merging tools}
\usage{
git_merge(ref, commit = TRUE, squash = FALSE, repo = ".")

git_merge_stage_only(ref, squash = FALSE, repo = ".")

git_merge_find_base(ref, target = "HEAD", repo = ".")

git_merge_analysis(ref, repo = ".")

git_merge_abort(repo = ".")
}
\arguments{
\item{ref}{branch or commit that you want to merge}

\item{commit}{automatically create a merge commit if the merge succeeds without
conflicts. Set this to \code{FALSE} if you want to customize your commit message/author.}

\item{squash}{omits the second parent from the commit, which make the merge a regular
single-parent commit.}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{target}{the branch where you want to merge into. Defaults to current \code{HEAD}.}
}
\description{
Use \code{git_merge} to merge a branch into the current head. Based on how the branches
have diverged, the function will select a fast-forward or merge-commit strategy.
}
\details{
By default \code{git_merge} automatically commits the merge commit upon success.
However if the merge fails with merge-conflicts, or if \code{commit} is set to
\code{FALSE}, the changes are staged and the repository is put in merging state,
and you have to manually run \code{git_commit} or \code{git_merge_abort} to proceed.

Other functions are more low-level tools that are used by \code{git_merge}.
\code{git_merge_find_base} looks up the commit where two branches have diverged
(i.e. the youngest common ancestor). The \code{git_merge_analysis} is used to
test if a merge can simply be fast forwarded or not.

The \code{git_merge_stage_only} function applies and stages changes, without
committing or fast-forwarding.
}
\seealso{
Other git: 
\code{\link{git_archive}},
\code{\link{git_branch}()},
\code{\link{git_commit}()},
\code{\link{git_config}()},
\code{\link{git_diff}()},
\code{\link{git_fetch}()},
\code{\link{git_rebase}()},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link{git_signature}()},
\code{\link{git_stash}},
\code{\link{git_tag}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/remotes.R
\name{git_remote}
\alias{git_remote}
\alias{git_remote_list}
\alias{git_remote_add}
\alias{git_remote_remove}
\alias{git_remote_info}
\alias{git_remote_set_url}
\alias{git_remote_set_pushurl}
\alias{git_remote_refspecs}
\title{Git Remotes}
\usage{
git_remote_list(repo = ".")

git_remote_add(url, name = "origin", refspec = NULL, repo = ".")

git_remote_remove(remote, repo = ".")

git_remote_info(remote = NULL, repo = ".")

git_remote_set_url(url, remote = NULL, repo = ".")

git_remote_set_pushurl(url, remote = NULL, add = FALSE, repo = ".")

git_remote_refspecs(remote = NULL, repo = ".")
}
\arguments{
\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \code{\link[=git_find]{git_find()}}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{url}{server url (https or ssh)}

\item{name}{unique name for the new remote}

\item{refspec}{optional string with the remote fetch value}

\item{remote}{name of an existing remote. Default \code{NULL} means the remote
from the upstream of the current branch.}

\item{add}{if \code{TRUE}, append the push URL instead of replacing it.
Equivalent to \verb{git remote set-url --push --add}.}
}
\description{
List, add, configure, or remove remotes.
}
\seealso{
Other git:
\code{\link{git_archive}},
\code{\link[=git_branch]{git_branch()}},
\code{\link[=git_commit]{git_commit()}},
\code{\link[=git_config]{git_config()}},
\code{\link[=git_diff]{git_diff()}},
\code{\link[=git_fetch]{git_fetch()}},
\code{\link{git_history}},
\code{\link{git_ignore}},
\code{\link[=git_merge]{git_merge()}},
\code{\link[=git_rebase]{git_rebase()}},
\code{\link{git_repo}},
\code{\link[=git_reset]{git_reset()}},
\code{\link[=git_restore]{git_restore()}},
\code{\link[=git_revert]{git_revert()}},
\code{\link[=git_signature]{git_signature()}},
\code{\link{git_stash}},
\code{\link{git_tag}},
\code{\link{git_worktree}}
}
\concept{git}
\section{Related libgit2 documentation}{\href{https://libgit2.org/docs/reference/main/remote/index.html}{\code{remote}}.}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge.R
\name{git_merge}
\alias{git_merge}
\alias{git_merge_stage_only}
\alias{git_merge_find_base}
\alias{git_merge_analysis}
\alias{git_merge_abort}
\title{Merging tools}
\usage{
git_merge(ref, commit = TRUE, squash = FALSE, repo = ".")

git_merge_stage_only(ref, squash = FALSE, repo = ".")

git_merge_find_base(ref, target = "HEAD", repo = ".")

git_merge_analysis(ref, repo = ".")

git_merge_abort(repo = ".")
}
\arguments{
\item{ref}{branch or commit that you want to merge}

\item{commit}{automatically create a merge commit if the merge succeeds without
conflicts. Set this to \code{FALSE} if you want to customize your commit message/author.}

\item{squash}{omits the second parent from the commit, which make the merge a regular
single-parent commit.}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{target}{the branch where you want to merge into. Defaults to current \code{HEAD}.}
}
\description{
Use \code{git_merge} to merge a branch into the current head. Based on how the branches
have diverged, the function will select a fast-forward or merge-commit strategy.
}
\details{
By default \code{git_merge} automatically commits the merge commit upon success.
However if the merge fails with merge-conflicts, or if \code{commit} is set to
\code{FALSE}, the changes are staged and the repository is put in merging state,
and you have to manually run \code{git_commit} or \code{git_merge_abort} to proceed.

Other functions are more low-level tools that are used by \code{git_merge}.
\code{git_merge_find_base} looks up the commit where two branches have diverged
(i.e. the youngest common ancestor). The \code{git_merge_analysis} is used to
test if a merge can simply be fast forwarded or not.

The \code{git_merge_stage_only} function applies and stages changes, without
committing or fast-forwarding.
}
\seealso{
Other git: 
\code{\link{git_archive}},
\code{\link{git_branch}()},
\code{\link{git_commit}()},
\code{\link{git_config}()},
\code{\link{git_diff}()},
\code{\link{git_fetch}()},
\code{\link{git_ignore}},
\code{\link{git_rebase}()},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link{git_signature}()},
\code{\link{git_stash}},
\code{\link{git_tag}}
}
\concept{git}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/submodules.R
\name{git_submodule_list}
\alias{git_submodule_list}
\alias{git_submodule_info}
\alias{git_submodule_init}
\alias{git_submodule_set_to}
\alias{git_submodule_add}
\alias{git_submodule_fetch}
\title{Submodules}
\usage{
git_submodule_list(repo = ".")

git_submodule_info(submodule, repo = ".")

git_submodule_init(submodule, overwrite = FALSE, repo = ".")

git_submodule_set_to(submodule, ref, checkout = TRUE, repo = ".")

git_submodule_add(url, path = basename(url), ref = "HEAD", ..., repo = ".")

git_submodule_fetch(submodule, ..., repo = ".")
}
\arguments{
\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \link{git_find}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{submodule}{name of the submodule}

\item{overwrite}{overwrite existing entries}

\item{ref}{a branch or tag or hash with}

\item{checkout}{actually switch the contents of the directory to this commit}

\item{url}{full git url of the submodule}

\item{path}{relative of the submodule}

\item{...}{extra arguments for \link{git_fetch} for authentication things}
}
\description{
Interact with submodules in the repository.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/commit.R
\name{git_history}
\alias{git_history}
\alias{git_commit_info}
\alias{git_commit_id}
\alias{git_commit_stats}
\alias{git_log}
\alias{git_stat_files}
\title{View commit history}
\usage{
git_commit_info(ref = "HEAD", repo = ".")

git_commit_id(ref = "HEAD", repo = ".")

git_commit_stats(ref = "HEAD", repo = ".")

git_log(ref = "HEAD", max = 100, after = NULL, path = NULL, repo = ".")

git_stat_files(files, ref = "HEAD", max = NULL, repo = ".")
}
\arguments{
\item{ref}{revision string with a branch/tag/commit value}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \code{\link[=git_find]{git_find()}}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{max}{lookup at most latest n parent commits}

\item{after}{date or timestamp: only include commits starting this date}

\item{path}{character vector with paths to filter on; only commits that
touch these paths are included}

\item{files}{vector of paths relative to the git root directory.
Use \code{"."} to stage all changed files.}
}
\value{
\itemize{
\item \code{git_commit_info()} and \code{git_commit_stats()} return a list.
}
}
\description{
\itemize{
\item \code{git_commit_stats()} returns information about commit insertion and deletion
\item \code{git_commit_info()} a list of commit info
\item \code{git_commit_id()} is a shortcut for \code{git_commit_info()$id}
\item \code{git_log()} shows the most recent commits
\item \code{git_ls()} lists all the files that are being tracked in the repository.
\item \code{git_stat_files()} shows information of when \code{files} was last modified.
}
}
\seealso{
Other git:
\code{\link{git_archive}},
\code{\link[=git_branch]{git_branch()}},
\code{\link[=git_commit]{git_commit()}},
\code{\link[=git_config]{git_config()}},
\code{\link[=git_diff]{git_diff()}},
\code{\link[=git_fetch]{git_fetch()}},
\code{\link{git_ignore}},
\code{\link[=git_merge]{git_merge()}},
\code{\link[=git_rebase]{git_rebase()}},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link[=git_reset]{git_reset()}},
\code{\link[=git_restore]{git_restore()}},
\code{\link[=git_revert]{git_revert()}},
\code{\link[=git_signature]{git_signature()}},
\code{\link{git_stash}},
\code{\link{git_tag}},
\code{\link{git_worktree}}
}
\concept{git}
\section{Related libgit2 documentation}{\href{https://libgit2.org/docs/reference/main/commit/index.html}{\code{commit}}.}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.R
\name{git_config}
\alias{git_config}
\alias{git_config_global}
\alias{git_config_get}
\alias{git_config_local_get}
\alias{git_config_global_get}
\alias{git_config_set}
\alias{git_config_global_set}
\alias{git_config_unset}
\alias{git_config_global_unset}
\title{Get or set Git configuration}
\usage{
git_config(repo = ".")

git_config_global()

git_config_get(name, repo = ".")

git_config_local_get(name, repo = ".")

git_config_global_get(name)

git_config_set(name, value, add = FALSE, repo = ".")

git_config_global_set(name, value, add = FALSE)

git_config_unset(name, pattern = NULL, repo = ".")

git_config_global_unset(name, pattern = NULL)
}
\arguments{
\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \code{\link[=git_find]{git_find()}}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{name}{Name of the option to get or set}

\item{value}{Value to set. Must be a string, logical, number or \code{NULL} (to
unset).}

\item{add}{if \code{TRUE}, append a new entry for \code{name} instead of replacing
existing one(s). Equivalent to \verb{git config --add}. Only supported for
string values.}

\item{pattern}{optional regular expression, for matching values to unset in
case of multiple values}
}
\value{
\itemize{
\item \code{git_config()}: a \code{data.frame} of the Git options "in force" in the context
of \code{repo}, one row per option. The \code{level} column reveals whether the
option is determined from global or local config.
\item \code{git_config_global()}: a \code{data.frame}, as for \code{git_config()}, except only
for global Git options.
\item \code{git_config_get()}: the value of the named option considering both local and
global config (local wins), or \code{NULL} if unset.
\item \code{git_config_local_get()}: as for \code{git_config_get()}, but restricted to
local (repository-level) config only.
\item \code{git_config_global_get()}: as for \code{git_config_get()}, but for global config only.
\item \code{git_config_set()}, \code{git_config_global_set()}: The previous value(s) of
\code{name} in local or global config, respectively. If this option was
previously unset, returns \code{NULL}. Returns invisibly.
\item \code{git_config_unset()}, \code{git_config_global_unset()}: The previous value(s) of
\code{name} that were unset. Returns invisibly.
}
}
\description{
Get, set, or unset Git options, as \verb{git config} does on the command line.
\strong{Global} settings affect all of a user's Git operations
(\verb{git config --global}), whereas \strong{local} settings are scoped to a specific
repository (\verb{git config --local}). When both exist, local options always win.\tabular{lll}{
    \tab \strong{local} \tab \strong{global} \cr
   get all \tab \code{git_config()} \tab \code{git_config_global()} \cr
   get one (local+global) \tab \code{git_config_get()} \tab \code{git_config_get()} \cr
   get one (local or global only) \tab \code{git_config_local_get()} \tab \code{git_config_global_get()} \cr
   set \tab \code{git_config_set()} \tab \code{git_config_global_set()} \cr
   unset \tab \code{git_config_unset()} \tab \code{git_config_global_unset()} \cr
}
}
\note{
All entries in the \code{name} column are automatically normalised to
lowercase (see
\url{https://libgit2.org/libgit2/#HEAD/type/git_config_entry} for details).
}
\examples{
# Set and inspect a local, custom Git option
r <- file.path(tempdir(), "gert-demo")
git_init(r)

previous <- git_config_set("aaa.bbb", "ccc", repo = r)
previous
git_config_local_get("aaa.bbb", repo = r)

previous <- git_config_set("aaa.bbb", NULL, repo = r)
previous
git_config_local_get("aaa.bbb", repo = r)

# Get a single named option (returns NULL if unset)
git_config_get("aaa.bbb", repo = r)
git_config_set("aaa.bbb", "ccc", repo = r)
git_config_get("aaa.bbb", repo = r)

unlink(r, recursive = TRUE)

\dontrun{
# Set global Git options
git_config_global_set("user.name", "Your Name")
git_config_global_set("user.email", "your@email.com")
git_config_global()

# Get a single global option (returns NULL if unset)
git_config_global_get("user.name")
git_config_global_get("gert.nonexistent")
}
}
\seealso{
Other git:
\code{\link{git_archive}},
\code{\link[=git_branch]{git_branch()}},
\code{\link[=git_commit]{git_commit()}},
\code{\link[=git_diff]{git_diff()}},
\code{\link[=git_fetch]{git_fetch()}},
\code{\link{git_history}},
\code{\link{git_ignore}},
\code{\link[=git_merge]{git_merge()}},
\code{\link[=git_rebase]{git_rebase()}},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link[=git_reset]{git_reset()}},
\code{\link[=git_restore]{git_restore()}},
\code{\link[=git_revert]{git_revert()}},
\code{\link[=git_signature]{git_signature()}},
\code{\link{git_stash}},
\code{\link{git_tag}},
\code{\link{git_worktree}}
}
\concept{git}
\section{Related libgit2 documentation}{\href{https://libgit2.org/docs/reference/main/config/index.html}{\code{config}}.}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/open.R
\name{git_open}
\alias{git_open}
\title{Open local repository}
\usage{
git_open(repo = ".")
}
\arguments{
\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \code{\link[=git_find]{git_find()}}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}
}
\value{
an pointer to the libgit2 repository
}
\description{
Returns a pointer to a libgit2 repository object. This function is mainly
for internal use; users should simply reference a repository in gert by
by the path to the directory.
}
\examples{
r <- tempfile(pattern = "gert")
git_init(r)
r_ptr <- git_open(r)
r_ptr
git_open(r_ptr)
git_info(r)

# cleanup
unlink(r, recursive = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge.R, R/commit.R
\name{git_merge}
\alias{git_merge}
\alias{git_merge_stage_only}
\alias{git_merge_find_base}
\alias{git_merge_analysis}
\alias{git_merge_abort}
\alias{git_commit_descendant_of}
\alias{git_conflicts}
\title{Git merge}
\usage{
git_merge(ref, commit = TRUE, squash = FALSE, repo = ".")

git_merge_stage_only(ref, squash = FALSE, repo = ".")

git_merge_find_base(ref, target = "HEAD", repo = ".")

git_merge_analysis(ref, repo = ".")

git_merge_abort(repo = ".")

git_commit_descendant_of(ancestor, ref = "HEAD", repo = ".")

git_conflicts(repo = ".")
}
\arguments{
\item{ref}{branch or commit that you want to merge}

\item{commit}{automatically create a merge commit if the merge succeeds without
conflicts. Set this to \code{FALSE} if you want to customize your commit message/author.}

\item{squash}{omits the second parent from the commit, which make the merge a regular
single-parent commit.}

\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \code{\link[=git_find]{git_find()}}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{target}{the branch where you want to merge into. Defaults to current \code{HEAD}.}

\item{ancestor}{a reference to a potential ancestor commit}
}
\description{
Use \code{git_merge()} to merge a branch into the current head. Based on how the branches
have diverged, the function will select a fast-forward or merge-commit strategy.

Other functions are more low-level tools that are used by \code{git_merge()}:
\itemize{
\item \code{git_merge_find_base()} looks up the commit where two branches have diverged
(i.e. the youngest common ancestor).
\item \code{git_merge_analysis()} is used to
test if a merge can simply be fast forwarded or not.
\item \code{git_merge_stage_only()} applies and stages changes, without committing or fast-forwarding.
\item \code{git_conflicts()} lists merge conflicts.
}
}
\details{
By default \code{git_merge()} automatically commits the merge commit upon success.
However if the merge fails with merge-conflicts, or if \code{commit} is set to
\code{FALSE}, the changes are staged and the repository is put in merging state,
and you have to manually run \code{\link[=git_commit]{git_commit()}} or \code{git_merge_abort()} to proceed.
}
\seealso{
Other git:
\code{\link{git_archive}},
\code{\link[=git_branch]{git_branch()}},
\code{\link[=git_commit]{git_commit()}},
\code{\link[=git_config]{git_config()}},
\code{\link[=git_diff]{git_diff()}},
\code{\link[=git_fetch]{git_fetch()}},
\code{\link{git_history}},
\code{\link{git_ignore}},
\code{\link[=git_rebase]{git_rebase()}},
\code{\link{git_remote}},
\code{\link{git_repo}},
\code{\link[=git_reset]{git_reset()}},
\code{\link[=git_restore]{git_restore()}},
\code{\link[=git_revert]{git_revert()}},
\code{\link[=git_signature]{git_signature()}},
\code{\link{git_stash}},
\code{\link{git_tag}},
\code{\link{git_worktree}}
}
\concept{git}
\section{Related libgit2 documentation}{\href{https://libgit2.org/docs/reference/main/merge/index.html}{\code{merge}}.}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/submodules.R
\name{git_submodule_list}
\alias{git_submodule_list}
\alias{git_submodule_info}
\alias{git_submodule_init}
\alias{git_submodule_set_to}
\alias{git_submodule_add}
\alias{git_submodule_fetch}
\title{Submodules}
\usage{
git_submodule_list(repo = ".")

git_submodule_info(submodule, repo = ".")

git_submodule_init(submodule, overwrite = FALSE, repo = ".")

git_submodule_set_to(submodule, ref, checkout = TRUE, repo = ".")

git_submodule_add(url, path = basename(url), ref = "HEAD", ..., repo = ".")

git_submodule_fetch(submodule, ..., repo = ".")
}
\arguments{
\item{repo}{The path to the git repository. If the directory is not a
repository, parent directories are considered (see \code{\link[=git_find]{git_find()}}). To disable
this search, provide the filepath protected with \code{\link[=I]{I()}}. When using this
parameter, always explicitly call by name (i.e. \verb{repo = }) because future
versions of gert may have additional parameters.}

\item{submodule}{name of the submodule}

\item{overwrite}{overwrite existing entries}

\item{ref}{a branch or tag or hash with}

\item{checkout}{actually switch the contents of the directory to this commit}

\item{url}{full git url of the submodule}

\item{path}{relative of the submodule}

\item{...}{extra arguments for \code{\link[=git_fetch]{git_fetch()}} for authentication things}
}
\description{
Interact with \href{https://git-scm.com/book/en/v2/Git-Tools-Submodules}{submodules}
in the repository.
}
\section{Related libgit2 documentation}{\href{https://libgit2.org/docs/reference/main/submodule/index.html}{\code{submodule}}.}
